/** \file
 * Executable which runs hellocomm::helloapp.
 * hello-comm-world is a very simple executable to run the hellocomm::helloapp
 * example skeleton mpi application within the SST/Macro simulator.
 */

#include <sstmac/sstmac.h>
#include <unistd.h> // getopt
#include "helloapp.h"


using namespace sstmac;
using namespace sstmac::hw;
using namespace sstmac::sw;
using namespace sstmac::native;

int main () {

  // Hardcode for 2 nodes, 1 processor per node
  const int nodes(2);
  const int ppn(1);

  //--------------------------------------
  // Section 1. Set up simulation manager
  //--------------------------------------

  // Construct the simulation manager.
  manager::ptr mgr = manager::construct();

  //--------------------------------------
  // Section 2. Set up interconnect model
  //--------------------------------------

  // Construct a network topology.
  // Use a line network (only 2 nodes after all) 
  // more interesting networks are available.
  topology::ptr net = line::construct(nodes);

  // Construct an interconnect model (circuit).
  const timestamp latency(1e-9); // latency in seconds
  const int bandwidth(125000000); // bandwidth in bytes/second
  interconnect::ptr ic =
    circuit::construct(latency,bandwidth,net);
  
  mgr->set_interconnect(ic);

  //----------------------
  // Section 3. Add nodes
  //----------------------

  // Add a node to the manager, manager will clone it "nodes" times
  // and assign those nodes to network slots.
  const double memory_bandwidth(1e10); // bandwidth in bytes/second
  const timestamp memory_latency(1e-9); // latency in seconds
  simplenode::ptr node =
    simplenode::construct(memory_latency, memory_bandwidth); // bandwidth in bytes/second
  simplenic::ptr nic = simplenic::construct(node, ic);
  mgr->add_nodes(node, nic, nodes);

  //-------------------------------------
  // Section 4. Add application launcher 
  //-------------------------------------

  // Construct an application object
  sstmac::sw::appid aid(0);
  sstmac::sw::taskid tid(0);
  sstmac::sw::softwareid sid(aid, tid);
  app::ptr application(hellocomm::helloapp::construct(sid));
  std::list<std::string> params;
  params.push_back("Why, hello world!");

  // Construct a launcher
  launchinfo::ptr l1 = launchinfo::construct(application, appid(0), params, timestamp(1), nodes);
  std::list<launchinfo::ptr> apps;
  apps.push_back(l1);

  allocationstrategy::ptr as = firstavailableallocation::construct(nodes);
  indexstrategy::ptr is = blockindexing::construct();
  instantlaunch::ptr job = instantlaunch::construct(apps, as, is);


  //-------------------------------
  // Section 5. Run the simulation
  //-------------------------------

  mgr->run(job);

}
