/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/driver_util.h>
#include <unistd.h> // getopt
#include <math.h> // pow

#include <sstmac/hardware/network/congestion/linearlatency.h>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

 using namespace sstmac::hw;
  using namespace sstmac;
  using namespace sstmac::native;
  using namespace sstmac::sw;


void
print_help(const char *name)
{
  std::cerr << name << " [-h] [-d debug_string] [-n nodes] [-p ppn] "
      << " [-c congestion_model] [-f flower name] [-a appname] [-r path prefix] \n";
}

int
main(int argc, char **argv)
{
  bool helpflag = false;
  time_t start = time(NULL);
  int opt;
  int nodes = 16, ppn = 1;
  int argc0 = argc;
  std::string dbgstr = "<debug> app";
  std::string appname = "minimd";
  std::string prefix = "";
  std::string congestion = "circuit";
  std::string flower = "ordered";


  while (!helpflag && (opt = getopt(argc, argv, "hd:r:n:a:p:c:f:")) != -1)
    {
      switch (opt)
        {
      case 'h':
        helpflag = true;
        break;

      case 'd':
        dbgstr = optarg;
        break;
      case 'n':
        nodes = atoi(optarg);
        break;
      case 'p':
        ppn = atoi(optarg);
        break;
      case 'a':
        appname = optarg;
        break;
      case 'r':
        prefix = std::string(optarg) + "/";
        break;
      case 'c':
        congestion = optarg;
        break;
      case 'f':
        flower = optarg;
        break;

      default:
        std::cerr << "Unknown argument -" << opt << "\n";
        helpflag = true;
        }
    }
  if (helpflag)
    {
      print_help(argv[0]);
      return 0;
    }

  bool errflag = false;
  if (nodes <= 0)
    {
      std::cerr << "Invalid node count \"" << nodes << "\"\n";
      errflag = true;
    }
  if (ppn <= 0)
    {
      std::cerr << "Invalid ppn count \"" << ppn << "\"\n";
      errflag = true;
    }


  if (errflag)
    {
      return 1;
    }
  debug::set_user_param(dbgstr);

  sstmac::sw::skeletonfactory::init_skeletons();


  int x = 1, y = 1, z = 1;
  int iter = 0;
  while(x*y*z < nodes){
      if (iter == 0) x++;
      if (iter == 1) y++;
      if (iter == 2) z++;

      iter = (iter + 1) % 3;
  }


   std::list<std::string> params;

   if(appname == "minimd"){
       params.push_back(prefix + "src/sstmac/software/skeletons/mini/minimd/lj.in");
   }else if (appname == "mpi3d"){
       std::stringstream ss1;
       ss1 << x;
       params.push_back(ss1.str());

       std::stringstream ss2;
       ss2 << x;
       params.push_back(ss2.str());

       std::stringstream ss3;
       ss3 << x;
       params.push_back(ss3.str());
   }else if(appname == "actormatmul"){
       params.push_back("1");
       params.push_back("1000");
       params.push_back("1");
   }


   bool errorflag = false;

   topology::ptr top = torus::construct_3D(x, y, z, 10);
   interconnect::ptr fab;

   // ------------ get congestion model ---------- //
   circuittype cir = parse_circuit(congestion.c_str(), errorflag);

   if(errorflag){
       std::cout << "ERROR: parsing congestion model named " << congestion << std::endl;
       return 1;
   }

   latencymodel::ptr lat = linearlatency::construct(4, timestamp(40e-9));
   bandwidth bw(8e9);
   if(cir == SHARED){
   // ------------ get flower model --------- //
     netsharing fl = parse_flower(flower.c_str(), errorflag);
       if(errorflag){
           std::cout << "ERROR: parsing flower model named " << flower << std::endl;
           return 1;
       }

       flower::ptr flow;

       if(fl == OVERCOMMIT){
           flow = overcommit_flower::construct(top, 1e9, 1e9);
       }else if(fl == THROTTLE_OVERCOMMITS){
           flow = throttle_flower::construct(top, 1e9, 1e9);
       }else if(fl == ORDERED){
           flow = ordered_flower::construct(top, 1e9, 1e9);
       }


       fab = sharedcircuit::construct(lat, bw, top, flow, power_config::ptr());
   }else if(cir == EXCLUSIVE){
       fab = circuit::construct(lat, bw, top, power_config::ptr());
   }else if(cir == BUFFER){
       fab = sharedbuffer::construct(1e9, 1e9, timestamp(0), timestamp(100e-9), timestamp(20e-9), 8192, 2, top);
   }



   run_simple_sim(nodes, ppn, appname, params, fab, randomindexing::construct(), 500e6, timestamp(1e-9));

   time_t end = time(NULL);

   std::cout << "wall time: " << (end - start) << std::endl;

  return 0;
}
