/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BIN_UNDUMPI_H
#define SSTMAC_BIN_UNDUMPI_H

#include <sstmac/software/skeletons/undumpi/dumpimeta.h>
//#include <sstmac/software/skeletons/undumpi/rsqindexing.h>
//#include <sstmac/software/skeletons/undumpi/allocation.h>
#include <sstmac/software/skeletons/undumpi/parsedumpi.h>
//#include <sstmac/backends/native/nodeid.h>
#include <dumpi/libundumpi/libundumpi.h>
#include <sstmac/sstmac.h>

#include <iostream>
#include <iterator>

using namespace sstmac;
using namespace sstmac::sw;
using namespace sstmac::hw;
using namespace sstmac::native;


std::string dumpi_file_name(int rank, std::string fileroot_) {
   static std::string here("dumpi_file_name: ");
   // Figure out our input file name.
   // Try fileroot-"%01d", fileroot-"%02d", ... up to some reasonable length
   // (in the short term that we will probably not have over 10 billion peers)
   FILE *fp = NULL;
   char fname[256]; fname[255] = '\0';
   const char *fnamefmt = "%%s-%%0%dd.bin";

   char fmt[30]; fmt[29] = '\0';
   for(int i = 1; i < 10; ++i) {
     snprintf(fmt, 29, fnamefmt, i);
     snprintf(fname, 255, fmt, fileroot_.c_str(), rank);
     //std::cerr << here << "looking for file " << fname << std::endl;
     //SSTMAC_DEBUG << "dumpi_file_name: checking file name: " << fname << "\n";
     fp = fopen(fname, "r");
     if(fp) {
       fclose(fp);
       return std::string(fname);
     }
   }
   // If we get here, then no file was found.
   throw ioerror(here + "failed to find a file starting from the fileroot \""
                 + fileroot_ + "\": " + strerror(errno));
 }

 int getnumprocs(dumpimeta::ptr dmeta_) {
   static std::string here("parsedumpi::getnumprocs: ");
   if (dmeta_) {
     return dmeta_->getnumprocs();
   }
   int nrank = 0;
   try {
     while (true) {
       std::string fname = dumpi_file_name(nrank, dmeta_->dirplusfileprefix_);
       nrank++;
     }
   } catch (ioerror&) {
     // ioerror is thrown when no more trace files can be found
   }
 // SSTMAC_DEBUG << here << "nrank = " << nrank << "\n";
   return nrank;
 }


/**
 * This function reads in a map file for mapping host names to network coordinates.
 * @param here string used for debugging / exception throwing
 * @param mapfile name of the map file
 * @param hostmap map in which host mappings are placed
 */
void read_map_file(const std::string &here,
    const std::string &mapfile,
    std::map<std::string,std::vector<int> >&hostmap)
{
  hostmap.clear();
  std::ifstream in(mapfile.c_str());
  if (in.bad()) {
      throw valueerror(here+"could not find node map file");
  }
  int nnode = -1;
  in >> nnode;
  if (nnode <= 0) throw valueerror(here+"bad nnode in node map file");
  int ncoor = -1;
  in >> ncoor;
  if (ncoor <= 0) throw valueerror(here+"bad ncoor in node map file");
  for (int i=0; i<nnode; i++) {

      std::string hostname;
      in >> hostname;

      if (hostname.size()==0) throw valueerror(here+"bad hostname in map file");
      std::vector<int> coor(ncoor);
      for (int j=0; j<ncoor; j++) {
          coor[j] = -1;
          in >> coor[j];
          if (coor[j] < 0) throw valueerror(here+"bad coordinate in map file");
      }

      std::cout << "inserting " << hostname << " into map at coor";
      for (int j=0; j<ncoor; j++)
        std::cout << " " << coor[j];
      std::cout << std::endl;

      hostmap[hostname] = coor;
  }
}


/**
 * This function reads in a given map file (see read_map_file()), and
 * returns the allocationstrategy and indexstrategy necessary to impement
 * the given mapping, given the topology.
 * @param mapfile name of the map file
 * @param parse the parsedumpi skeleton app
 * @param net the topology pointer
 * @param alloc reference variable where the allocationstrategy is returned
 * @param index indexstrategy variable where the indexstrategy is returned
 */
void
layout(const std::string &mapfile, const dumpimeta::ptr &meta,
    const sstmac::hw::topology::ptr &net,
    sstmac::sw::allocationstrategy::ptr &alloc,
    sstmac::sw::indexstrategy::ptr &index)
{
  static std::string here("parsedumpi::layout: ");

  std::map<std::string, std::vector<int> > hostmap;
  read_map_file(here, mapfile, hostmap);

  std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> nodeidset;
  int nrank = getnumprocs(meta);
  std::vector<nodeaddress::ptr> process_to_nodeid(nrank);
  std::cout << "nrank: " << nrank << std::endl;
  for (int i=0; i<nrank; i++) {
      std::string fname = dumpi_file_name(i, meta->dirplusfileprefix_);
      dumpi_profile *profile = undumpi_open(fname.c_str());
      if (profile == NULL) {
          throw ioerror(here + "unable to open \"" + fname + "\"");
      }
      dumpi_header *header = dumpi_build_header();
      if (header == NULL) {
          throw ioerror(here + " header is null");
      }
      dumpi_read_header(profile, header);
      if (header->hostname == NULL || *header->hostname == 0) {
          throw ioerror(here + " hostname is null or empty");
      }
      std::string hostname(header->hostname);
      dumpi_free_header(header);
      std::cout << here << " rank " << i << " is on " << hostname
          << " with coordinate ";
      std::map<std::string, std::vector<int> >::iterator maploc
      = hostmap.find(hostname);
      if (maploc==hostmap.end()) throw valueerror(here+hostname+" not in map");
      std::vector<int> coor = maploc->second;
      std::copy(coor.begin(),coor.end(),
          std::ostream_iterator<int>(std::cout," "));

      // find node index for this vertex
      int64_t nodeindex;
      vertex hostvert(coor);
      try {
          nodeindex = net->node_to_index(hostvert);
      }
      catch(invalidkeyerror e) {
          std::cout << e.what() << "\n" << here
              << "possible mismatch between network geometry and hostmap\n";
          throw;
      }
      catch(std::exception const& e) {
          std::cout << "can't find index for vertex: " << hostvert << "\n";
          throw;
      }
      std::cout << " and is on simulator node " << nodeindex;
      std::cout << "\n";

      nodeaddress::ptr aptr = nodeaddress::convert_from_uniqueid(nodeindex);
     // nodeaddress::ptr aptr = nodeid::construct(nodeindex);
      process_to_nodeid[i] = aptr;
      nodeidset.insert(aptr);
  }
 // std::vector<nodeaddress::ptr> nodeidvec;
 // std::copy(nodeidset.begin(), nodeidset.end(), std::back_inserter(nodeidvec));
  alloc = fixedallocation::construct(nodeidset);
  index = fixedindexing::construct(process_to_nodeid);
}


#endif


