/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/backends/native/components/node/simplenode.h>
//#include <sstmac/processor/simpleinstruction.h>
#include <sstmac/backends/native/components/nic/networkinterface.h>
//#include <sstmac/simulator/eventmanager.h>
//#include <sstmac/processor/instructions.h>
//#include <sstmac/simulator/networkdata.h>
//#include <sstmac/simulator/nodeinterface.h>
//#include <sstmac/simulator/taskid.h>
//#include <sstmac/process/server.h>
//#include <sstmac/processor/procscheduler.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstream>

namespace sstmac { namespace native
  {




  simplenode::simplenode(const timestamp &memory_latency,
      double memory_bandwidth, int n_core) :
      memory_latency_(memory_latency), memory_bandwidth_(memory_bandwidth), cores_(
          n_core)
  {
    os_->setAMI(sstmac::sw::AMI::AMI_COMM, this);
    os_->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
    os_->setAMI(sstmac::sw::AMI::AMI_QUERY, this);

  }

  /// Construct a new node with the given memory bandwidth.
  simplenode::ptr
  simplenode::construct(const timestamp &memory_latency,
      double memory_bandwidth, int n_core)
  {
    simplenode::ptr retval(
        new simplenode(memory_latency, memory_bandwidth, n_core));


    return retval;
  }

  node::ptr
  simplenode::clone() const
  {
    node::ptr ret = construct(memory_latency_, memory_bandwidth_,
        cores_);

    return ret;
  }

  void
  simplenode::set_eventmanager(const eventmanager::ptr &m)
  {
    eventmodule::set_eventmanager(m);

    compute_sched_->set_eventmanager(m);

    os_->start_apps();
  }

  void
  simplenode::initialize(nodeid::ptr my_id, const sstmac::sw::service::ptr &the_launcher)
  {
    node::initialize(my_id, the_launcher);

    double pfreq = 1e9;

    double tcycle = 1.0 / (pfreq);

    timestamp tflop(tcycle * 10); // 10 clock cycles for a floating point
    timestamp tmemop(memory_latency_.sec() + 1.0 / memory_bandwidth_);


    sstmac::hw::simpleprocessor_config::ptr cfg =
        sstmac::hw::simpleprocessor_config::construct(cores_, pfreq, 10, 30);

    compute_sched_ = sstmac::hw::computescheduler::construct(cfg->n_cores_);

    proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_, compute_sched_);


    sstmac::hw::instructionprocessor_config::ptr pcfg =
        sstmac::hw::instructionprocessor_config::construct(cores_,
            INST_FLOP_LABEL, tflop, INST_MEMOP_LABEL, tmemop, INST_MEMOP_LOCAL_LABEL,
            tmemop, INST_MEMOP_RAND_LABEL, tmemop);

    proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg, msg_handler_,
        compute_sched_);

  }

  void
  simplenode::set_nic(const networkinterface::ptr &n)
  {
    nic_ = n;
  }

  networkinterface::ptr
  simplenode::get_nic() const
  {
    return nic_;
  }

  // --- sstmac::sw::AMI pure virtuals ---- //
  void
  simplenode::execute_kernel(int func, sst_message::ptr data)
  {
    switch (func)
    {
    case sstmac::sw::AMI::AMI_COMM_SEND:
      {

        SSTMAC_DEBUG << "[t=" << now() << "] -- Node(" << nodeid_->to_string()
                        << "): sending to " << data->toaddr_->to_string() << "\n";

        nodeid::ptr nid = boost::dynamic_pointer_cast<nodeid>(
            data->toaddr_);
        nic_->send(nid, data, msg_handler_);
        break;
      }

    case sstmac::sw::AMI::AMI_COMP_TIME:
      {

        proc_model_->compute(data);

        break;
      }

    case sstmac::sw::AMI::AMI_COMP_INSTR:
      {

        proc_inst_->compute(data);

        break;
      }

    case sstmac::sw::AMI::AMI_COMP_SLEEP:
      schedule(now() + data->get_time(), ptr(this), data);
      break;

    default:
      std::cout << "Kernel number: " << func << std::endl;
      throw ssterror("simplenode: cannot process kernel");

    }

  }


  ptr_type::ptr
  simplenode::execute_kernel(int func){
    switch(func){
    case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:{
      sstmac::ptr_type::ptr top = nic_->get_topology();
      return top;
      break;
    }
    case sstmac::sw::AMI::AMI_QUERY_ADDRESS:{
      return nodeid_;
      break;
    }

    case sstmac::sw::AMI::AMI_QUERY_MACHINE:{
      return sstmac::sw::machine_descriptor::construct(nic_->get_topology()->n(), nic_->get_topology()->name());
      break;
    }



    std::cout << "Kernel number: " << func << std::endl;
    throw ssterror("simplenode: cannot process kernel");
    }
  }

  timestamp
  simplenode::now()
  {
    return eventmodule::now();
  }

  bool
  simplenode::kernel_supported(int func)
  {
    switch (func)
    {
    case sstmac::sw::AMI::AMI_COMM_SEND:
      return true;

    case sstmac::sw::AMI::AMI_COMM_NIC_SEND_ACK:
      return true;

    case sstmac::sw::AMI::AMI_COMP_TIME:
      return true;

    case sstmac::sw::AMI::AMI_COMP_SLEEP:
      return true;

    case sstmac::sw::AMI::AMI_COMP_INSTR:
      return true;

    case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:
      return true;

    case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
      return true;

    default:
      return false;

    }
  }

  } } // end of namespace sstmac.
