/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_MANAGER_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_MANAGER_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/backends/native/nodeid.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/ptr_type.h>

#include <sstmac/backends/native/nodelist.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/software/services/launch/indexing/indexstrategy.h>
#include <sstmac/hardware/network/congestion/interconnect.h>
#include <sstmac/common/eventmanager.h>
#include <sstmac/software/services/launch/launcher.h>
#include <sstmac/backends/native/components/node/node.h>
#include <vector>
#include <set>
//#include <backends/sstmacro-native/statscollector.h>

namespace sstmac  { namespace native
  {

  // Forward declaration of relevant types.

  class networkinterface;

  /**
   * The owner of a discrete-event network simulation.
   *
   * All time progression is handled by this object
   * and messages between nodes are managed here as well.
   */
  class manager: virtual public nocopy, public ptr_type
  {
  public:
    virtual std::string to_string() const{
      return "manager";
    }

    typedef boost::intrusive_ptr<manager> ptr;
  private:
    /// The nested specialization of the environment type.
    /// One of these gets handed to each launcher started under this manager.
    // class env;




    /// The event manager.
    eventmanager::ptr eventmanager_;

    /// The network simulator.
    sstmac::hw::interconnect::ptr interconnect_;

    /// The node objects.
    nodelist::ptr nodes_;

    /// Keep track of parallel processes.
    typedef std::set<sstmac::sw::launcher::ptr > launchers_t;
    launchers_t launchers_;

    /// The set of nodes that are available.
   // std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> available_;

    /// The set of nodes that are allocated.
    //std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> allocated_;

    /// Monitor whether the simulator is currently running.
    bool running_;

    /// Next parallel process id.
    sstmac::sw::appid next_ppid_;

    /// Default allocation strategy.
   // sstmac::sw::allocationstrategy::ptr default_allocation_;

    /// Default index strategy.
  //  sstmac::sw::indexstrategy::ptr default_indexing_;

    debug dbg_;

    /// Power manager, collects power stats.
    //boost::shared_ptr<statscollector> stats;

    /// The processes to be fired up at startup.
    typedef eventhandler::ptr event_t;
    typedef std::vector<std::pair<timestamp, event_t> > startup_t;
    startup_t startup_;

    /// Private method to signal that a launcher has called exit.
    void
    launcher_exiting(const sstmac::sw::launcher::ptr &,
        const std::vector<nodeid::ptr> &released_nodes);

  protected:
    /// Default constructor.
    /// By default the manager has no nodes or network; it can only
    /// be run after one or more nodes have been added.
    manager();



  public:
    /// Construct a new manager.
    static manager::ptr
    construct();

    /// Goodbye.
    virtual
    ~manager() throw ();



    /// Define a interconnect for this manager.
    void
    set_interconnect(const sstmac::hw::interconnect::ptr &net);

    /// Access the simulator used by this manager.
    sstmac::hw::interconnect::const_ptr
    get_interconnect() const;

    /// Access the number of nodes added to date.
    /// This can be any value <= get_simulator()->size().
    int64_t
    nodes() const;

    /// Get the node corresponding to nodeid.
    /// \param id is the nodeid of the desired node.
    /// \return the node object correspondint to nodeid.
    node::ptr get_node(nodeid::ptr id) const;

    /// Find out how many slots in the network are free (after subtracting
    /// nodes added to date).  Return -1 if an unlimited
    /// number of slots are free.
    int64_t
    open_slots() const;

    /// Add one or more copies of the given node type to the manager.
    /// sequentially numbered starting at this->nodes().
    /// \param type The type of nodes added (replicated using node::clone()).
    /// \param count The number of instances of this type to add.
    ///        If count < 0, fill all available network connections with
    ///        this node type.
    /// \return the number of unfilled connections on the network.
    /// \throw nullerror if simulator has not been defined.
    /// \throw nullerror if type is a NULL pointer.
    /// \throw valueerror if this->nodes() is less than count.
    int64_t
    add_nodes(const node::ptr &type,
        const boost::intrusive_ptr<networkinterface> &nictype, int64_t count =
            -1);

    /// Schedule the launch of a parallel job on the given number of nodes
    /// at or after the given start time.
    /// The launcher will be given a list of nodes to run on (via the
    /// launcher::start method).  The launcher is free to set up shared
    /// resources between tasks running on a shared physical computer,
    /// but should not assume that all virtual nodes are running on the
    /// same physical computer.
    ///
    /// At startup, the launcher will be given a manager::delegate object
    /// that permits calls to add_launcher, schedule_trigger, schedule_send,
    /// and complete_launcher (these methods are not publically accessible).
    ///
    /// Please try to resist the urge to perform out-of-band communications
    /// directly between nodes, since this will harm future attempts to
    /// parallelize the simulator.
    /// \param start_time  the time at which the job is to be started (>= 0).
    /// \param launcher the object that gets run at the given time.
    ///        Note that this object is not cloned -- at the end of the run
    ///        it will reflect any changes performed during the execution.
    /// \param nodes the number of nodes for the simulation.
    /// \param ppn the number of launchers per node.
    /// \param is the strategy for assigning node indices.  Defaults to
    ///        blockindexing if no index strategy is specified.
    /// \param as the strategy for allocating nodes.  Defaults to
    ///        firstavailableallocationstrategy.
    /// \param nproc optional parameter giving the total number of processes. If nproc is not given or -1, then nodes*ppn processes are allocated.
    /// \throw timeerror if start_time is less than zero or out of range.
    /// \throw valueerror if nodes <= 0 or nodes > this->nodes().
    /// \throw valueerror if ppn <= 0.
    void
    add_launcher(
        const timestamp &start_time,
        const sstmac::sw::launcher::ptr &the_launcher,
        int nodes,
        int ppn,
        const sstmac::sw::indexstrategy::ptr &is =
            sstmac::sw::indexstrategy::ptr(),
            const sstmac::sw::allocationstrategy::ptr &as =
                sstmac::sw::allocationstrategy::ptr(), int nproc = -1);

    // NOTES:
    //     What should we do about launchers added after the simulation
    //     has started with a start-time in the future (when more or fewer
    //     nodes may be available)?  How much of a "queueing system"
    //     do we want to work in here?

    // TODO:
    // 1)  Add a way to specify which nodes get used for a given job
    //     (to facilitate simulating non-uniform node/network topologies).
    // 2)  Make it easier to name (or label) nodes, either individually by
    //     class (same reasoning as above).
    // 3)  Decide on a more final strategy for handling tiered topologies
    //     such as systems with GPGPUs or accelerators (e.g. RoadRunner).

    /// Setup the simulation.
    /// Initializes eventmanager, interconnect, nodes, and scheduler.
    void setup(const sstmac::sw::launcher::ptr &the_launcher);

    /// Run the simulation.
    /// \return the time at the end of the simulation.
    /// \throw illformederror if the simulation cannot be run
    /// (e.g. if no nodes have been added).
    timestamp run(const sstmac::sw::launcher::ptr &the_launcher);

    /// End the simulation.
    /// Does error checking and prints statistics.
    void done();

  };
  } } // end of namespace sstmac.

#endif
