/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "cycleaccurateswitch.h"

#include "network_m.h"

namespace sstmac { namespace omnetpp {


  using namespace std;

  Define_Module(cycleaccurateswitch)
  ;

  void
  cycleaccurateswitch::initialize()
  {

    //latencyPerHop = getParam("latencyPerHop");

    int myrackid = par("rackid");
    int npr = par("nodesPerRack");
    int nrc = par("numRackConnections");

    omnetpp_address::ptr me = omnetpp_address::construct(-1, -1, -1, myrackid);

    int topP1 = par("topParam1");
    int topP2 = par("topParam2");
    int topP3 = par("topParam3");

    string top = par("networkTopology");

    init(me, npr, nrc, top, topP1, topP2, topP3);

    numVC_ = par("numVC");
    inBuffSize_ = par("inBuffSize");
    outBuffSize_ = par("outBuffSize");

    arbLatency_ = par("arbitrationLatency");
    switchLatency_ = par("xBarLatency");
    switchBandwidth_ = par("xBarBandwidth");
    switchBandwidth_ *= 1e9;

    clockFreq_ = par("clockFreq");
    clockFreq_ *= 1e6;
    clockper_ = 1.0 / clockFreq_;

    for (int i = 0; i < numRackConnections; i++)
      {
        inports[gate("ports$i", i)] = i;
        outports[i] = gate("ports$o", i);

        for (int j = 0; j < numVC_; j++)
          {
            inbuffs_[i * numVC_ + j] = new cPacketQueue();

            credits_[i * numVC_ + j] = inBuffSize_;
          }

        outbuffs_[i] = new cPacketQueue();

        inports_free_[i] = 0;
        outports_free_[i] = 0;

        pending_creds_up_[i] = new std::queue<cMessage*>();
        arb_vc_[i] = 0;
        num_in_buff_[i] = 0;

        outBuffMsgs_[i] = new cMessage("outbuffmsg");
        revOutBuffMsgs_[outBuffMsgs_[i]] = i;
      }

    clockmsg_ = new cMessage("clockmsg");

    arb_inport_ = 0;
  }

  void
  cycleaccurateswitch::finish()
  {

    for (int i = 0; i < numRackConnections; i++)
      {

        for (int j = 0; j < numVC_; j++)
          {

            int index = i * numVC_ + j;

            if (credits_[index] != inBuffSize_)
              {
                std::cout << "cycleaccurateswitch[" << myAddr->rack()
                      << "]: credits for VC " << j << " on inport " << i << " (" << credits_[index]
                                                    << ") do not equal buff size(" << inBuffSize_ << ")"
                                                    << std::endl;
              }

            if (!inbuffs_[index]->empty())
              {
                std::cout << "cycleaccurateswitch[" << myAddr->rack()
                      << "]: messages still in inbuff " << i << ", vc = " << j
                      << std::endl;
              }
          }

        if (!outbuffs_[i]->empty())
          {
            std::cout << "cycleaccurateswitch[" << myAddr->rack()
                  << "]: messages still in outbuff " << i << std::endl;
          }
      }

    for (int i = 0; i < numRackConnections; i++)
      {

        for (int j = 0; j < numVC_; j++)
          {
            delete inbuffs_[i * numVC_ + j];

          }

        delete outbuffs_[i];
        delete pending_creds_up_[i];
        delete outBuffMsgs_[i];
      }

    delete clockmsg_;
  }

  void
  cycleaccurateswitch::handleMessage(cMessage* cmsg)
  {

    if (cmsg->isSelfMessage())
      {
        if (cmsg == clockmsg_)
          {
            arbitrate();
          }
        else if (revOutBuffMsgs_.find(cmsg) != revOutBuffMsgs_.end())
          {
            int outport = revOutBuffMsgs_[cmsg];

            SSTMAC_DEBUG << "switch[" << myAddr->rack() << "]: sending on outport "
              << outport << "\n";

            ///send out the out buffer
            NetworkAccurateMsg* n =
                (NetworkAccurateMsg*) outbuffs_[outport]->front();
            outbuffs_[outport]->pop();

            send(n, outports[outport]);

            if (!outbuffs_[outport]->empty())
              {
                SSTMAC_DEBUG << "switch[" << myAddr->rack()
                  << "]: scheduling outport selfmessage, queue size is "
                  << outports[outport]->size() << " \n";
                scheduleAt(
                    outports[outport]->getTransmissionChannel()->getTransmissionFinishTime(),
                    cmsg);
              }

          }
        else
          { //it's a data message arriving to the outbuff
            NetworkAccurateMsg* n = check_and_cast<NetworkAccurateMsg*> (cmsg);
            omnetpp_address::ptr addr = n->getToAddr();

            int outport = router_->route(addr);

            SSTMAC_DEBUG << "switch[" << myAddr->rack()
              << "]: data message arriving to outbuff " << outport << "\n";

            outbuffs_[outport]->insert(n);

            if (!outBuffMsgs_[outport]->isScheduled())
              {
                simtime_t
                atTime =
                    max(
                        simTime() + clockper_,
                        outports[outport]->getTransmissionChannel()->getTransmissionFinishTime());
                scheduleAt(atTime, outBuffMsgs_[outport]);

              }

            if (!clockmsg_->isScheduled())
              {
                scheduleAt(simTime() + clockper_, clockmsg_);
              }
          }
      }
    else
      {
        NetworkAccurateMsg* n = check_and_cast<NetworkAccurateMsg*> (cmsg);
        omnetpp_address::ptr addr = n->getToAddr();
        omnetpp_address::ptr from = n->getFromAddr();

        cGate* ingate = cmsg->getArrivalGate();

        int inport = inports[ingate];
        int inbuffindex = inport * numVC_ + n->getVc();
        int msgtype = n->getCamsgtype();

        SSTMAC_DEBUG<< "switch[" << myAddr->rack() << "]: message arrived on inport "
          << inport << " on VC " << n->getVc() << " ...";

        if (msgtype == CAMSG_CREDITUP)
          {
            int creds = n->getCredits();

            credits_[inbuffindex] += creds;

            SSTMAC_DEBUG << "credits up by " << creds << " to " << credits_[inbuffindex]
                                                                     << "\n";

            if (!clockmsg_->isScheduled())
              {
                scheduleAt(simTime() + clockper_, clockmsg_);
              }

            delete cmsg;
            return;
          }
        else
          { //it's data
            SSTMAC_DEBUG << "data \n";

            inbuffs_[inbuffindex]->insert(n);
            num_in_buff_[inport]++;

            if (!clockmsg_->isScheduled())
              {
                scheduleAt(simTime() + clockper_, clockmsg_);
              }

          }
      }

  }

  void
  cycleaccurateswitch::arbitrate()
  {

    //iterate over input ports, starting with the last one that went
    int maxGrants = 2;
    int numGrants = 0;

    int begininport = (arb_inport_ + 1) % numRackConnections;
    bool firstinport = true;
    for (int i = begininport; firstinport || (i != begininport && numGrants
        < maxGrants); i = (i + 1) % numRackConnections)
      {
        SSTMAC_DEBUG << "switch[" << myAddr->rack() << "]: examining inport " << i
          << ", num_in_buff = " << num_in_buff_[i] << ", time = "
          << simTime().dbl() << ", inport_free = " << inports_free_[i].dbl()
          << "\n";
        firstinport = false;
        if (num_in_buff_[i] > 0)
          {
            if (inports_free_[i] <= simTime())
              {

                int beginVC = (arb_vc_[i] + 1) % numVC_;
                bool firstVC = true;
                for (int vc = beginVC; firstVC || vc != beginVC; vc = (vc + 1)
                % numVC_)
                  {

                    firstVC = false;
                    int index = i * numVC_ + vc;

                    SSTMAC_DEBUG << "switch[" << myAddr->rack() << "]: examining vc "
                      << vc << ", queue size = " << inbuffs_[index]->length()
                      << "";

                    if (!inbuffs_[index]->empty())
                      {
                        NetworkAccurateMsg * msg =
                            (NetworkAccurateMsg*) inbuffs_[index]->front();
                        omnetpp_address::ptr addr = msg->getToAddr();
                        omnetpp_address::ptr from = msg->getFromAddr();
                        int outport = router_->route(addr);

                        SSTMAC_DEBUG << ", outports_free = "
                          << outports_free_[outport].dbl()
                          << ", outbuff occupancy = "
                          << outbuffs_[outport]->getByteLength()
                          << ", credits = "
                          << credits_[outport * numVC_ + vc] << "\n";

                        if (outports_free_[outport] <= simTime() && (outBuffSize_
                            - outbuffs_[outport]->getByteLength())
                            >= msg->getByteLength() && credits_[outport * numVC_
                                                                + vc] >= msg->getByteLength())
                          {
                            SSTMAC_DEBUG << "switch[" << myAddr->rack()
                              << "]: arbitration good from " << i << " to "
                              << outport << " \n";

                            inbuffs_[i * numVC_ + vc]->pop();
                            //now we're ok to send through the crossbar
                            //  sendDelayed(msg, arbLatency_, outports[outport]);
                            simtime_t outbuffarrival = simTime() + switchLatency_
                                + msg->getByteLength() * 8 / switchBandwidth_;
                            scheduleAt(outbuffarrival, msg);

                            inports_free_[i] = outbuffarrival;
                            outports_free_[outport] = outbuffarrival;
                            arb_vc_[i] = vc;
                            numGrants++;
                            arb_inport_ = i;
                            num_in_buff_[i]--;

                            int inbuffindex = outport * numVC_ + vc;
                            credits_[inbuffindex] -= msg->getByteLength();

                            //send credit ack back

                            NetworkAccurateMsg* ack = new NetworkAccurateMsg();
                            ack->setByteLength(8);
                            ack->setCredits(msg->getByteLength());
                            ack->setCamsgtype(CAMSG_CREDITUP);
                            ack->setVc(vc);
                            outbuffs_[i]->insert(ack);

                            if (!outBuffMsgs_[i]->isScheduled())
                              {
                                simtime_t
                                atTime =
                                    max(
                                        simTime() + clockper_,
                                        outports[i]->getTransmissionChannel()->getTransmissionFinishTime());
                                scheduleAt(atTime, outBuffMsgs_[i]);
                              }

                            break;
                          }
                      }
                  }
              }
          }

      }

    if (numGrants > 0)
      {
        scheduleAt(simTime() + clockper_, clockmsg_);
      }
  }

  /*

 void cycleaccurateswitch::tryOutQ() {
 bool successful = false;
 simtime_t closestFinish = MAX_SIM_TIME;
 for (int i = 0; i < numRackConnections; i++) {

 if (!outQs[i]->empty()) {
 simtime_t
 txFinishTime =
 outports[i]->getTransmissionChannel()->getTransmissionFinishTime();
 if (txFinishTime <= simTime()) { //channel free
 //std::cout << "NIC: sending message " << endl;
 cPacket* msg = outQs[i]->pop();
 send(msg, outports[i]);

 successful = true;
 } else if (!transFinish->isScheduled()) {
 if (txFinishTime < closestFinish)
 closestFinish = txFinishTime;
 }
 }
 }

 if (successful)
 tryOutQ();
 else if (closestFinish < MAX_SIM_TIME)
 scheduleAt(closestFinish, transFinish);

 }
   */
} }
