/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <networkmodels/flow/flowrackswitch.h>

#include <network_m.h>

namespace sstmac { namespace omnetpp {


  Define_Module(flowrackswitch);

  using namespace std;

  void
  flowrackswitch::initialize()
  {

    //latencyPerHop = getParam("latencyPerHop");

    int myrackid = par("rackid");
    int npr = par("nodesPerRack");
    int nrc = par("numRackConnections");

    omnetpp_address::ptr me = omnetpp_address::construct(-1, -1, -1, myrackid);

    int topP1 = par("topParam1");
    int topP2 = par("topParam2");
    int topP3 = par("topParam3");

    string top = par("networkTopology");

    init(me, npr, nrc, top, topP1, topP2, topP3);

    networkFlowBW = par("networkFlowBandwidth");

    displayConnections = par("updateConnectionDisplays");

    for (int i = 0; i < numRackConnections; i++)
      {
        //inports[i] = gate("ports$i", i);
        outports[i] = gate("ports$o", i);

        // if (outports[i]->findTransmissionChannel() != NULL
        // )
        //   register_channel(outports[i]->getTransmissionChannel(), this);

        //outQs[i] = new cPacketQueue();
        flows[i] = new set<pair<int, int> >();
        requestedBW[i] = 0;
      }

    display_connections_ = false;

  }

  void
  flowrackswitch::finish()
  {
    for (int i = 0; i < numRackConnections; i++)
      {
        delete flows[i];
      }

    //delete transFinish;
  }

  void
  flowrackswitch::handleMessage(cMessage* cmsg)
  {

    NetworkFlowMsg* n = check_and_cast<NetworkFlowMsg*>(cmsg);
    omnetpp_address::ptr addr = n->getToAddr();
    omnetpp_address::ptr from = n->getFromAddr();

    int outport = router_->route(addr);

    pair<int, int> key(from->rack() * nodesPerRack + from->node(), n->getFlowID());

    if (n->isSelfMessage() && !n->getIsHead())
      { //tail scheduled to leave
        requestedBW[outport] -= n->getSourceBW();
        flows[outport]->erase(key);
        flowStart.erase(key);
        flowSrcBW.erase(key);
        flowBW.erase(key);

        lastUpdate[outport] = simTime();

        send(n, outports[outport]);
      }
    else if (n->isSelfMessage() && n->getIsHead())
      { //head scheduled to leave
        send(n, outports[outport]);

      }
    else if (n->getIsHead())
      {
        flowBW[key] = n->getSourceBW();
        flowSrcBW[key] = n->getSourceBW();
        flowStart[key] = simTime();
        requestedBW[outport] += n->getSourceBW();
        flows[outport]->insert(key);

        lastUpdate[outport] = simTime();

        SSTMAC_DEBUG << "Flow " << key << ": head arrived at switch " << myAddr->to_string()
            << " at time " << simTime().dbl() << "\n";

        scheduleAt(simTime() + ROUTER_DELAY, n);
      }
    else
      {
        SSTMAC_DEBUG<< "Flow " << key << ": tail arrived at switch " << myAddr->to_string()
            << " at time " << simTime().dbl() << "\n";

        recalculateBandwidths(outport);

        if(n->getSourceBW() == 0){
            opp_error("flowrackswitch: tail arrival with 0 source BW");
        }

        if(flowBW[key] == 0){
            opp_error("flowrackswitch: tail arrival with assigned bandwidth of 0");
        }

        simtime_t originaltime = n->getEncapsulatedPacket()->getByteLength()
              / n->getSourceBW();
        simtime_t newtime = n->getEncapsulatedPacket()->getByteLength() / flowBW[key];

        SSTMAC_DEBUG << "Delaying by " << (newtime - originaltime).dbl() << "\n";

        scheduleAt(simTime() + ROUTER_DELAY + (newtime - originaltime), n);

      }

  }

  void
  flowrackswitch::recalculateBandwidths(int port)
  {
    set<pair<int, int> >::iterator it;

    double diff = max(0.0, requestedBW[port] - networkFlowBW);

    double totalBW = 0;

    for (it = flows[port]->begin(); it != flows[port]->end(); it++)
      {
        double newBW = flowSrcBW[*it] - diff / flows[port]->size();

        double t1 = (lastUpdate[port] - flowStart[*it]).dbl();
        double t2 = (simTime() - lastUpdate[port]).dbl();
        double t3 = max((simTime() - flowStart[*it]).dbl(), 1E-9);

        if(t1 == 0 && t2 == 0){
            t1 = 1E-9;
        }

        SSTMAC_DEBUG << "old flowBW[] = " << flowBW[*it] << ", newBW=" << newBW << ", t1=" << t1 << ", t2=" << t2 << ", t3=" << t3 << "\n";

        flowBW[*it] = (flowBW[*it] * t1 + newBW * t2) / t3;



        SSTMAC_DEBUG << "Assigning flow " << *it << " effective bandwidth of " << flowBW[*it]
                                                                                         << "\n";
        totalBW += flowBW[*it];
      }

    if(display_connections_){
        double frac = (double) totalBW / (double) (networkFlowBW);

        cDisplayString& connDispStr =
            outports[port]->getTransmissionChannel()->getDisplayString();


        std::stringstream ss;
        int linewidth = (1 + frac * 8);
        ss << "ls=red," << linewidth;
        connDispStr.parse(ss.str().c_str());

        //update_channel_display(outports[port]->getTransmissionChannel(), this, frac);
    }
  }

  /*

 void flowrackswitch::tryOutQ() {
 bool successful = false;
 simtime_t closestFinish = MAX_SIM_TIME;
 for (int i = 0; i < numRackConnections; i++) {

 if (!outQs[i]->empty()) {
 simtime_t
 txFinishTime =
 outports[i]->getTransmissionChannel()->getTransmissionFinishTime();
 if (txFinishTime <= simTime()) { //channel free
 //std::cout << "NIC: sending message " << endl;
 cPacket* msg = outQs[i]->pop();
 send(msg, outports[i]);

 successful = true;
 } else if (!transFinish->isScheduled()) {
 if (txFinishTime < closestFinish)
 closestFinish = txFinishTime;
 }
 }
 }

 if (successful)
 tryOutQ();
 else if (closestFinish < MAX_SIM_TIME)
 scheduleAt(closestFinish, transFinish);

 }
   */
} }
