/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <Node/NIC/mpicollectivessecondarynic.h>

#include <network_m.h>
//#include "mpi_m.h"

namespace sstmac { namespace omnetpp {


  Define_Module(mpicollectivessecondarynic)
    ;

  void
  mpicollectivessecondarynic::initialize()
  {
    fromProc = gate("toProcessor$i", 0);
    toProc = gate("toProcessor$o", 0);
    fromNet = gate("toNetwork$i", 0);
    toNet = gate("toNetwork$o", 0);
    toCollNet = gate("toNetwork$o", 1);
    fromCollNet = gate("toNetwork$i", 1);

    networkFlowBW = par("networkFlowBandwidth");
    memoryFlowBW = par("memoryFlowBandwidth");

    mtu = par("packetMTU");

    effectiveBW = std::min(networkFlowBW, memoryFlowBW);

    currentlyTransmitting = false;
    flowid = 0;

    currentMsg = NULL;

    sent = 0;
    guardMsg = new cMessage("NICguardMsg");
    nextPacketMsg = new cMessage("nextPacketMsg");
  }

  void
  mpicollectivessecondarynic::finish()
  {
    delete guardMsg;
    delete nextPacketMsg;

    if (!netQ.empty())
      std::cout << getFullPath() << ": netQ not empty" << endl;

    if (!collNetQ.empty())
      std::cout << getFullPath() << ": collNetQ not empty" << endl;
  }

  void
  mpicollectivessecondarynic::handleMessage(cMessage *cmsg)
  {
    cGate* arrival = cmsg->getArrivalGate();

    if (cmsg == nextPacketMsg)
      {
        if (currentMsg == NULL)
          tryCollNetQ();
        else
          next_packet();
      }
    else if (arrival == fromProc)
      {
        CoreMsg* c = check_and_cast<CoreMsg*> (cmsg);

        if (c->getMsgType() == APP)
          {
            /*            MPImsg* mmsg = check_and_cast<MPImsg*> (c->getEncapsulatedPacket());

           int type = mmsg->getMsgType();
           int tag = mmsg->getTag();
           if (type == MPI_BCAST || type == MPI_FINALIZE || type
           == MPI_BARRIER || type == MPI_SSEND_ACK) {
           collNetQ.insert(c);
           if (currentMsg == NULL)
           tryCollNetQ();
           } else {
           netQ.insert(c);
           try_net_Q();
           }
             */
          }
        else
          {

            netQ.insert(c);
            try_net_Q();
          }

      }
    else if (arrival == fromNet)
      {
        NetworkFlowMsg* net = check_and_cast<NetworkFlowMsg*> (cmsg);

        if (!net->getIsHead())
          {
            CoreMsg* m = (CoreMsg*) net->decapsulate();
            send(m, toProc);
          }
        delete net;
      }
    else if (arrival == fromCollNet)
      {
        NetworkPacketMsg* net = check_and_cast<NetworkPacketMsg*> (cmsg);

        if (net->getIsLast())
          {
            net->setByteLength(0x7FFFFFFF); //do this so omnet won't complain about decapsulating
            CoreMsg* m = (CoreMsg*) net->decapsulate();
            send(m, toProc);
          }
        delete net;
      }
    else if (cmsg == guardMsg)
      {
        currentlyTransmitting = false;
        try_net_Q();
      }
    else if (cmsg->isSelfMessage())
      { //it's the tail that's scheduled to go out
        send(cmsg, toNet);

        scheduleAt(simTime() + GUARD_TIME, guardMsg);
      }
    else
      {
        opp_error("NIC: unexpected message in handleMessage");
      }
  }

  void
  mpicollectivessecondarynic::tryCollNetQ()
  {
    if (!collNetQ.empty())
      {
        CoreMsg* msg = check_and_cast<CoreMsg*> (collNetQ.pop());
        currentMsg = msg;

        //ack on start of transmission
        if (msg->getEncapsulatedPacket() != NULL)
          {
            ApplicationMsg* appmsg = (ApplicationMsg*) msg->decapsulate();
            send_proc_ack(appmsg->dup(), toProc);
            msg->encapsulate(appmsg);
          }

        next_packet();

      }

  }

  void
  mpicollectivessecondarynic::next_packet()
  {
    if (currentMsg == NULL)
      {
        opp_error(
            "packetnic: in next_packet(), but currentMsg is null. shouldn't be here.");
      }

    if (!nextPacketMsg->isScheduled())
      {

        bool last = (mtu < 0) || ((sent + mtu) > currentMsg->getByteLength());

        NetworkPacketMsg* pack = new NetworkPacketMsg();
        pack->setToAddr(currentMsg->getToAddr());
        pack->setFromAddr(currentMsg->getFromAddr());
        pack->setIsLast(last);
        pack->setKind(currentMsg->getKind());

        if (last)
          {
            int diff = currentMsg->getByteLength() - sent;
            pack->encapsulate(currentMsg);
            pack->setByteLength(diff);
            sent = 0;
          }
        else
          {
            pack->setByteLength(mtu);
            sent += mtu;
          }
        SSTMAC_DEBUG << "NIC: Sending a packet of length "
            << (int) pack->getByteLength() << "\n";

        send(pack, toCollNet);

        if (last)
          {
            currentMsg = NULL;
          }

        simtime_t txFinishTime =
            toCollNet->getTransmissionChannel()->getTransmissionFinishTime();
        scheduleAt(txFinishTime, nextPacketMsg);

      }

  }

  void
  mpicollectivessecondarynic::try_net_Q()
  {
    if (!netQ.empty() && !currentlyTransmitting)
      {

        //std::cout << "NIC: sending message " << endl;
        CoreMsg* msg = check_and_cast<CoreMsg*> (netQ.pop());
        if (msg->getEncapsulatedPacket() != NULL)
          {
            ApplicationMsg* appmsg = (ApplicationMsg*) msg->decapsulate();
            send_proc_ack(appmsg->dup(), toProc);
            msg->encapsulate(appmsg);
          }
        SSTMAC_DEBUG << "NIC: Sending a FLOW of length " << (int) msg->getByteLength()
                << "\n";

        NetworkFlowMsg* head = new NetworkFlowMsg();
        head->setToAddr(msg->getToAddr());
        head->setFromAddr(msg->getFromAddr());
        head->setIsHead(true);
        head->setByteLength(0);
        head->setFlowID(flowid);
        head->setSourceBW(effectiveBW);

        send(head, toNet);

        NetworkFlowMsg* tail = new NetworkFlowMsg();
        tail->setToAddr(msg->getToAddr());
        tail->setFromAddr(msg->getFromAddr());
        tail->setIsHead(false);
        tail->encapsulate(msg);
        tail->setByteLength(0);
        tail->setFlowID(flowid);
        tail->setSourceBW(effectiveBW);

        flowid++;

        currentlyTransmitting = true;

        scheduleAt(simTime() + msg->getByteLength() / effectiveBW, tail);
      }

  }

} }
