/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "abstractnodeprocessor.h"

#include <sstmac/software/skeletons/skeletonfactory.h>

#include <sstmac/software/ami/hardware_descriptor.h>

#include <sstmac/common/logger.h>

#include <Node/Processor/resourceallocator/rackallocator.h>
#include <Node/Processor/resourceallocator/abstractnodeallocator.h>

#include <Node/Processor/Indexing/randomindexer.h>
#include <Node/Processor/Indexing/fixedindexer.h>
#include <Node/Processor/Indexing/fromfileindexer.h>

#include <Node/Processor/space/rackspace.h>
#include <Node/Processor/space/abstractnodespace.h>

#include <sstmac/software/libraries/compute/lib_compute_inst.h>
#include <sstmac/software/libraries/util/lib_display.h>



namespace sstmac { namespace omnetpp {

  Define_Module(sstmac::omnetpp::abstractnodeprocessor);

  using namespace std;

  debug abstractnodeprocessor::dbg_;


  //map<int, Indexer*> Core::appIndexers;
  map<sstmac::sw::appid, sstmac::sw::environment::ptr> abstractnodeprocessor::env_;
  bool abstractnodeprocessor::debug_init = false;


  abstractnodeprocessor::abstractnodeprocessor()
  {
    if(!dbg_.is_real()){
        dbg_ = debug("<debug> core | processor");
        logger::register_logger(&dbg_);
    }

  }

  abstractnodeprocessor::~abstractnodeprocessor()
  {

  }

  void
  abstractnodeprocessor::initialize()
  {

    int c = 0;

    int p = 0;

    int n = par("nodeid");

    int r = par("rackid");

    myid = omnetpp_address::construct(c, p, n, r);

    //app = NULL;

    //dbg.type = debug::DBG_CORE;

    sstmac::sw::skeletonfactory::init_skeletons();

    coresPerProc = par("coresPerProc");
    numOfNodes = par("numOfNodes");
    procsPerNode = par("procsPerNode");
    nodesPerRack = par("nodesPerRack");
    numRacks = par("numRacks");

    omnetpp_address::set_system_params(coresPerProc, procsPerNode, nodesPerRack, numRacks);

    outgate = gate("toNetwork$o");
    ingate = gate("toNetwork$i");

    fillColor = sstmac::sw::GREY;
    borderColor = sstmac::sw::BLACK;

    string temp = par("logger_params");
    bool color = par("debug_coloring");
    logger::set_debug_coloring(color);

    if (!debug_init)
      {
        logger::set_user_param(temp);
        debug_init = true;
      }

    // ------- operating system -------
    sstmac::sw::hardware_descriptor::ptr hw =
        sstmac::sw::hardware_descriptor::construct(coresPerProc);
    opsys = sstmac::sw::operatingsystem::construct(hw);

    opsys->setAMI(sstmac::sw::AMI::AMI_COMM, this);
    opsys->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
    opsys->setAMI(sstmac::sw::AMI::AMI_DISP, this);
    opsys->setAMI(sstmac::sw::AMI::AMI_QUERY, this);

    // ---- get parameters and such

    int numOfApps = par("numOfApps");
    nextApp = 0;

    for (int i = 0; i < numOfApps; i++)
      {
        stringstream ss;

        ss << "app" << i << "Name";
        string temp1 = par(ss.str().c_str());
        appNames[i] = temp1;
        ss.str("");

        if (i == 0)
          appStartTimes[i] = 0;
        else
          {
            ss << "app" << i << "StartTime";
            appStartTimes[1] = par(ss.str().c_str());
            ss.str("");
          }

        ss << "app" << i << "Size";
        appSizes[i] = par(ss.str().c_str());
        ss.str("");

        appIndexers[i] = NULL;

        appStringParams[i] = new list<string>();

        ss << "app" << i << "NumParams";
        int num_params = par(ss.str().c_str());
        ss.str("");

        for (int j = 0; j < num_params; j++)
          {

            ss << "app" << i << "Param" << j;
            string temp2 = par(ss.str().c_str());
            appStringParams[i]->push_back(temp2);
            ss.str("");
          }

      }

    appStartTimes[numOfApps] = -1;

    //----- self messages
    //computeMsg = new cMessage("computeMsg");
    for (int i = 0; i < coresPerProc; i++)
      {
        computeMessages[new cMessage("computemessage")] = sst_message::null();
      }

    appStartMsg = new cMessage("appStartMsg");

    // ---- set up master core
    if (is_master_core())
      {
        //only the master makes new environments. they get sent out at launch for nodes not in this address space
        for (int i = 0; i < numOfApps; i++)
          {
            sstmac::sw::appid aid(i);
            env_[aid] = sstmac::sw::environment::construct();
          }

        newAppMsg = new cMessage("newAppMsg");

        string allocName = par("allocatorName");

        instantiate_allocator(allocName);

        //appspace = fullspace;

        scheduleAt(0, newAppMsg);

      }

    //everyone should set the environment pointers
    /*for (int i = 0; i < numOfApps; i++)
 {
 sstmac::sw::appid aid(i);

 opsys->set_environment(env_[aid], aid);
 }*/

    //--- processor models ----//
    msg_handler_ = msg_handler::construct(this);
    boost::intrusive_ptr<fakeeventmanager> fem = fakeeventmanager::construct(this);

    double pfreq = par("procSpeed");
   // if (myid->rack() % 4 == 1) //degrade every other rack
   //   pfreq = pfreq / 2.0;
    double tcycle = 1.0 / (pfreq * 1e9);

    timestamp tflop(tcycle * 10); // 10 clock cycles for a floating point
    timestamp tmemop(tcycle * 5);

    sstmac::hw::instructionprocessor_config::ptr pcfg =
        sstmac::hw::instructionprocessor_config::construct(coresPerProc,
            INST_FLOP_LABEL, tflop, INST_MEMOP_LABEL, tmemop, INST_MEMOP_LOCAL_LABEL,
            tmemop, INST_MEMOP_RAND_LABEL, tmemop);

    sstmac::hw::simpleprocessor_config::ptr scfg =
        sstmac::hw::simpleprocessor_config::construct(coresPerProc, pfreq, 10, 30);

    compute_sched_ = sstmac::hw::computescheduler::construct(coresPerProc);

    proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg, msg_handler_,
        compute_sched_);
    proc_time_ = sstmac::hw::simpleprocessor::construct(scfg, msg_handler_,
        compute_sched_);

    compute_sched_->set_eventmanager(fem);

  }

  void
  abstractnodeprocessor::finish()
  {

    opsys->apps_kill();

    map<int, list<string>*>::iterator it;
    for (it = appStringParams.begin(); it != appStringParams.end(); it++)
      delete it->second;

    map<cMessage*, sst_message::ptr>::iterator it2;
    for (it2 = computeMessages.begin(); it2 != computeMessages.end(); it2++)
      delete it2->first;

    //delete computeMsg;
    delete appStartMsg;

    if (is_master_core())
      delete newAppMsg;

  }

  void
  abstractnodeprocessor::instantiate_allocator(string allocName)
  {

    if (allocName.compare("rack") == 0)
      {

        allocator = new rackallocator(numOfNodes, coresPerProc, procsPerNode,
            nodesPerRack, numRacks);

      }
    else if (allocName.compare("abstractnode") == 0)
      {
        allocator = new abstractnodeallocator(numOfNodes, coresPerProc, procsPerNode,
            nodesPerRack, numRacks);
      }
    else if (allocName.compare("system") == 0)
      { //the default whole-system allocator
        allocator = new resourceallocator(numOfNodes, coresPerProc, procsPerNode,
            nodesPerRack, numRacks);
      }
    else
      {
        std::cout << "Allocator Name: " << allocName << endl;
        opp_error("Proc: unable to find allocator.");
      }

  }

  bool
  abstractnodeprocessor::is_master_core()
  {
    return ((myid->rack() == 0 || myid->rack() < 0) && myid->node() == 0);
  }

  void
  abstractnodeprocessor::instantiate_indexer(space* appspace, int appid)
  {

    if (appIndexers[appid] == NULL)
      {
        SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc " << myid->to_string()
              << ": instantiating app indexer \n";

        string indexerName = par("appIndexerName");

        if (indexerName.compare("random") == 0)
          {
            appIndexers[appid] = new randomindexer(appspace, appSizes[appid]);
          }
        else if (indexerName.compare("fixed") == 0)
          {
            appIndexers[appid] = new fixedindexer(appspace, appSizes[appid]);
          }
        else if (indexerName.compare("fromFile") == 0)
          {
            string ipar = par("appIndexerParam1");
            appIndexers[appid] = new fromfileindexer(appspace, ipar);
          }
        else
          {
            std::cout << "indexer: " << indexerName << std::endl;
            opp_error("unknown app indexer");
          }

        SSTMAC_DEBUG << "indexer instantiated \n";
      }

  }

  void
  abstractnodeprocessor::instantiate_app(indexer* appind, std::string appname,
      sstmac::sw::softwareid newid)
  {
    SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc " << myid->to_string()
            << ": instantiating app" << "\n";

    if (sstmac::sw::skeletonfactory::apps.find(appname)
        == sstmac::sw::skeletonfactory::apps.end())
      {

        opp_error("Application %s is not in the app factory.", appname.c_str());
      }

    sstmac::sw::app::ptr theapp = sstmac::sw::skeletonfactory::apps[appname]->clone(
        newid);
    // boost::shared_ptr<sstmac::sw::parameter_configurator> params =
    //   theapp->get_param_config();

    //consume parameters
    //list<string>::iterator it;
    // for (it = appStringParams[newid.app_.id]->begin(); it
    //    != appStringParams[newid.app_.id]->end(); it++)
    //  {
    //    params->consume_parameter(*it);
    //  }

    opsys->add_application(theapp, *appStringParams[newid.app_.id]);
  }

  space*
  abstractnodeprocessor::instantiate_space(string spacename, int spaceparam1,
      int spaceparam2)
  {
    space* ret = NULL;

    SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc " << myid->to_string()
            << ": instantiating " << spacename << " space with params " << spaceparam1
            << " and " << spaceparam2 << "\n";

    if (spacename.compare("system") == 0)
      {
        ret = new systemspace(numOfNodes, coresPerProc, procsPerNode, nodesPerRack,
            numRacks);
      }
    else if (spacename.compare("rack") == 0)
      {

        ret = new rackspace(spaceparam1, spaceparam2, numOfNodes, coresPerProc,
            procsPerNode, nodesPerRack, numRacks);
      }
    else if (spacename.compare("abstractnode") == 0)
      {

        ret = new abstractnodespace(spaceparam1, spaceparam2, numOfNodes, coresPerProc,
            procsPerNode, nodesPerRack, numRacks);
      }
    else
      {
        std::cout << "Space name: " << spacename << endl;
        opp_error("Proc: don't know how to instantiate space");

      }

    return ret;
  }

  bool
  abstractnodeprocessor::launch_app(indexer* index, space* spc,
      std::string appname, int aid, int appsize)
  {
    SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc: master: launching " << appsize
        << " app instances" << "\n";

    bool setmyown = false;

    AppLaunchMsg* launch = new AppLaunchMsg();
    launch->setMsgType(OS_LAUNCH);
    launch->setByteLength(APP_SIZE + OS_MSG_HEADER_SIZE);

    launch->setSpaceName(spc->get_name().c_str());
    launch->setSpaceParam1(spc->get_space_param1());
    launch->setSpaceParam2(spc->get_space_param2());

    launch->setAppname(appname.c_str());
    launch->setAppID(aid);
    launch->setAppsize(appsize);
    launch->setKind(0);

    sstmac::sw::appid a(aid);
    environmentpod epod(env_[a]);
    launch->setEnv(epod);

    //int ranks = space->get_num_slots();

    for (int i = 0; i < appsize; i++)
      {
        omnetpp_address::ptr n = index->getLocation(i);

        sstmac::sw::taskid t(i);
        sstmac::sw::softwareid sid(a, t);
        env_[a]->add_item(t, n); // ------ this is MAGIC, to save on space --------- //
        SSTMAC_DEBUG << "[t=" << now() << "] -- Adding to environment map - sid: "
            << sid.to_string() << ", addr: " << n->to_string() << "\n";
        if (!n->equals(myid))
          {
            AppLaunchMsg* msg = launch->dup();
            msg->setInstanceID(i);

            msg->setToAddr(n);
            msg->setFromAddr(myid);

            sendDelayed(msg, SEND_DELAY, outgate);

          }
        else
          {
            opsys->set_environment(env_[a], a);

            instantiate_app(index, appname, sid);

            if (!appStartMsg->isScheduled())
              {
                SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc " << myid->to_string()
            << ": launching app id " << aid << ", instance " << i << "\n";
                scheduleAt(simTime() + APP_START_DELAY, appStartMsg);
              }
            setmyown = true;
          }

      }

    delete launch;
    //appslaunched = ranks;

    return setmyown;
  }

  void
  abstractnodeprocessor::handleMessage(cMessage* cmsg)
  {
    //std::cout << "Core " << myid << ": handle message" << endl;
    /*if (computeMessages.find(cmsg) != computeMessages.end())
 {
 SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc(" << myid->to_string()
 << "): done computing " << "\n";

 sst_message::ptr unblock = computeMessages[cmsg];

 if (!suspendedQ.empty())
 {
 SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc(" << myid->to_string()
 << "): executing suspended kernel " << "\n";

 execute_kernel(sstmac::sw::AMI::AMI_COMP_TIME, suspendedQ.front());
 suspendedQ.pop();
 }

 //opsys->unblock(unblock);
 opsys->handle_message(unblock);
 //app->doneComputing();
 }*/
    if (cmsg->isSelfMessage())
      {
        if (cmsg == appStartMsg)
          {
            opsys->start_apps();

            //set rack display
            cDisplayString& dispStr =
                getParentModule()->getParentModule()->getDisplayString();
            dispStr.parse("b=30,30,rectangle,black,black,4");
          }
        else if (cmsg == newAppMsg)
          {
            SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Master Proc: starting app " << nextApp
                << "\n";
            //get a subset of the space here with a resource allocator
            space *appspace = allocator->allocate(appSizes[nextApp]);

            /* if (appspace->get_num_slots() < appSizes[nextApp])
 {
 std::cout
 << "--- WARNING: allocator returned a space with less nodes than was requested"
 << endl;
 std::cout << "Requested: " << appSizes[nextApp] << endl;
 std::cout << "Allocated: " << appspace->get_num_slots() << endl;
 std::cout << "--- Resizing the app to fit in the space" << endl;

 appSizes[nextApp] = appspace->get_num_slots();
 }*/

            instantiate_indexer(appspace, nextApp);

            //Launch on given hardware space
            if (!launch_app(appIndexers[nextApp], appspace, appNames[nextApp], nextApp,
                appSizes[nextApp]))
              delete appspace;

            delete appIndexers[nextApp]; // get rid of it, so we're not keeping the mapping around twice (already in env_)

            nextApp++;

            if (appStartTimes[nextApp] > 0)
              {
                scheduleAt(appStartTimes[nextApp], newAppMsg);
              }

          }
        else
          {
            EventWrapperMsg* emsg = check_and_cast<EventWrapperMsg*>(cmsg);
            eventhandlerpod epod = emsg->getEvhandler();
            eventhandler::ptr evh = epod.load;
            sst_message::ptr m = emsg->getMsgcontent().load;
            evh->handle(m);

            delete emsg;
          }
      }
    else if (cmsg->getArrivalGate() == ingate)
      { //message arriving in from the NoC
        CoreMsg *m = check_and_cast<CoreMsg*>(cmsg);

        if (!m->getToAddr()->equals(myid))
          {
            std::cout << "My ID: " << myid->to_string() << endl;
            std::cout << "Msg Dest: " << m->getToAddr()->to_string() << endl;
            opp_error("Application: got a message destined for NOT me");
          }

        switch (m->getMsgType())
        {
        case OS_LAUNCH:
          {

            AppLaunchMsg* launch = check_and_cast<AppLaunchMsg*>(m);

            //std::cout << "Core " << myid << ": received an OS_LAUNCH" << endl;
            //first, instantiate the space that describes where
            // this app lives in the physical system
            std::string spacename = launch->getSpaceName();

            space* thespace = instantiate_space(spacename, launch->get_space_param1(),
                launch->get_space_param2());

            instantiate_indexer(thespace, launch->getAppID());

            sstmac::sw::appid a(launch->getAppID());
            sstmac::sw::taskid t(launch->getInstanceID());
            sstmac::sw::softwareid sid(a, t);

            environmentpod epod = launch->getEnv();

            env_[a] = epod.load;

            opsys->set_environment(env_[a], a);

            //now intantiate and start
            instantiate_app(appIndexers[launch->getAppID()], launch->getAppname(), sid);

            SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc " << myid->to_string()
            << ": launching app id " << launch->getAppID() << ", instance "
            << launch->getInstanceID() << "\n";
            if (!appStartMsg->isScheduled())
              {
                scheduleAt(simTime() + APP_START_DELAY, appStartMsg);
              }

            //delete appStringParams[launch->getAppID()];
            break;
          }

        case APP:
          {
            SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc " << myid->to_string()
            << ": app message arrived \n";
            ApplicationMsg* appmsg = check_and_cast<ApplicationMsg*>(m);
            payloadpod ppod = appmsg->getMsgcontent();
            sst_message::ptr bsstm = ppod.load;
            // sst_message sstmsg = appmsg->getMsgcontent();

            opsys->handle_message(bsstm);

            break;
          }

        case NIC_SEND_ACK:
          {
            SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc " << myid->to_string()
            << ": NIC send ack arrived ";
            ApplicationMsg* appmsg = check_and_cast<ApplicationMsg*>(m->decapsulate());

            payloadpod ppod = appmsg->getMsgcontent();
            sst_message::ptr bsstm = ppod.load;

            SSTMAC_DEBUG << bsstm->to_string() << "\n";
            opsys->handle_message(bsstm);

            //  ((CommunicationAPI*) app->getAPI(appmsg->getApiID()))->nicSendAck(
            //     appmsg);
            //((CommunicationAPI*)apis[appmsg->getApiID()])->nicSendAck(appmsg);
            delete appmsg;
            //}
            break;
          }

        case OS_KILL:
          //delete app;
          // app = NULL;

          break;

        default:
          std::cout << "Proc id: " << myid->to_string() << "\n";
          opp_error("Proc: unknown OS message type received.");

        }

        delete cmsg;

      }
    else
      { //it's a scheduled message to send out

        send(cmsg, outgate);

      }
  }

  // --- sstmac::sw::AMI pure virtuals ---- //
  ptr_type::ptr
  abstractnodeprocessor::execute_kernel(int func){
	  switch(func){

	  case sstmac::sw::AMI::AMI_QUERY_ADDRESS:{
		  return myid;
		  break;
	  }

	  }

	  std::cout << "Kernel number: " << func << std::endl;
	  throw ssterror("simplenode: cannot process kernel");
  }

  void
  abstractnodeprocessor::execute_kernel(int func, sst_message::ptr data)
  {
    switch (func)
    {
    case sstmac::sw::AMI::AMI_COMM_SEND:
      {

        payloadpod ppod(data);

        ApplicationMsg* cmsg = new ApplicationMsg();
        cmsg->setMsgType(APP);
        cmsg->setMsgcontent(ppod);
        omnetpp_address::ptr froma = boost::dynamic_pointer_cast < omnetpp_address
            > (data->fromaddr_);
        omnetpp_address::ptr toa = boost::dynamic_pointer_cast < omnetpp_address
            > (data->toaddr_);
        cmsg->setFromAddr(froma);
        cmsg->setToAddr(toa);
        cmsg->setByteLength(data->get_byte_length());
        cmsg->setKind(data->get_kind() + 1); //add 1, because 0 is reserved for Processor/management type stuff like launch

        SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc(" << myid->to_string()
            << "): sending to " << toa->to_string() << "\n";

        send(cmsg, outgate);
        break;
      }

    case sstmac::sw::AMI::AMI_COMP_TIME:
      {

        proc_time_->compute(data);

        break;
      }

    case sstmac::sw::AMI::AMI_COMP_SLEEP:
      {
        EventWrapperMsg* emsg = new EventWrapperMsg();
        eventhandlerpod epod;
        epod.load = msg_handler_;

        emsg->setEvhandler(epod);
        emsg->setMsgcontent(data);

        scheduleAt(simTime() + data->get_time().sec(), emsg);

        break;
      }

    case sstmac::sw::AMI::AMI_COMP_INSTR:
      proc_inst_->compute(data);
      break;

    case sstmac::sw::AMI::AMI_DISP_FILL:
      {
        SSTMAC_DEBUG << "Proc: changing color to code " << data->get_kind() << "\n";
        fillColor = data->get_kind();
        cDisplayString& dispStr =
            getParentModule()->getParentModule()->getDisplayString();
        std::stringstream ss;
        ss << "b=30,30,rectangle," << get_color_string(fillColor) << ","
            << get_color_string(borderColor) << ",4";

        dispStr.parse(ss.str().c_str());
        break;
      }

    case sstmac::sw::AMI::AMI_DISP_BORDER:
      {
        SSTMAC_DEBUG << "Proc: changing color to code " << data->get_kind() << "\n";
        borderColor = data->get_kind();
        cDisplayString& dispStr =
            getParentModule()->getParentModule()->getDisplayString();
        std::stringstream ss;
        ss << "b=30,30,rectangle," << get_color_string(fillColor) << ","
            << get_color_string(borderColor) << ",4";

        dispStr.parse(ss.str().c_str());
        break;
      }

    default:
      std::cout << "Kernel number: " << func << endl;
      opp_error("Proc: received a kernel which I don't know how to execute");

    }

  }

  std::string
  abstractnodeprocessor::get_color_string(int c)
  {
    switch (c)
    {
    case sstmac::sw::BLACK:
      return "black";
      break;
    case sstmac::sw::RED:
      return "red";
      break;
    case sstmac::sw::GREEN:
      return "green";
      break;
    case sstmac::sw::YELLOW:
      return "yellow";
      break;
    case sstmac::sw::ORANGE:
      return "orange";
      break;
    case sstmac::sw::PURPLE:
      return "purple";
      break;
    case sstmac::sw::PINK:
      return "pink";
      break;

    case sstmac::sw::GREY:
      return "gray";
      break;

    default:
      return "white";
    }

  }

  timestamp
  abstractnodeprocessor::now()
  {
    timestamp t(simTime().dbl());
    return t;
  }

  bool
  abstractnodeprocessor::kernel_supported(int func)
  {
    switch (func)
    {
    case sstmac::sw::AMI::AMI_COMM_SEND:
      return true;

    case sstmac::sw::AMI::AMI_COMM_NIC_SEND_ACK:
      return true;

    case sstmac::sw::AMI::AMI_COMP_TIME:
      return true;

    case sstmac::sw::AMI::AMI_COMP_SLEEP:
      return true;

    case sstmac::sw::AMI::AMI_COMP_INSTR:
      return true;

    case sstmac::sw::AMI::AMI_DISP_BORDER:
      return true;

    case sstmac::sw::AMI::AMI_DISP_FILL:
      return true;

    default:
      return false;

    }
  }
  // ------------------------------------ //

} }
