/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_EVENTMANAGER_H_INCLUDED
#define SSTMAC_COMMON_EVENTMANAGER_H_INCLUDED

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/stats/stat_logger.h>


namespace sstmac {

/**
 * Base type for implementations of an engine that
 * is able to schedule events and advance simulation time
 * in the right order.
 */
class eventmanager : public ptr_type
{


protected:
  timestamp now_;

  std::list<stat_logger::ptr> stats_;

  void
  set_now(const timestamp &ts)
  {
    now_ = ts;
  }

  eventmanager()
  {
	  if(!dbg_.is_real()){
		  dbg_ = debug("<debug> eventmanager");
		  logger::register_logger(&dbg_);
	  }

	  if(!updates_.is_real()){
		  updates_ = debug("<debug> eventupdates");
		  logger::register_logger(&updates_);
	  }
  }

  static debug dbg_;
  static debug updates_;

  time_t sim_start_;
  time_t last_update_;
  time_t update_interval_;

public:
  virtual std::string to_string() const{
              return "eventmanager";
            }

  typedef boost::intrusive_ptr<eventmanager> ptr;

  /// Goodbye.
  virtual
  ~eventmanager()
  {
  }

  /// Clear all events and set time back to a zero of your choice.
  /// This call shall not be permitted while the event manager is running.
  virtual void
  clear(const timestamp &zero_time = timestamp(0)) = 0;

  /// Run the eventmanager.
  /// The eventmanager shall return control when no more messages remain.
  virtual void
  run() = 0;

  /// Get the current eventmanager time.
  /// This call must work regardless of whether the eventmanager is currently
  /// running or not.  After a run, the call should return the time at which
  /// the last event was completed.
  virtual const timestamp&
  now()
  {
    return now_;
  }

  void register_stat(stat_logger::ptr stat){
    stats_.push_back(stat);
  }

  /// Set off the given eventhandler at the given time.
  virtual void
  schedule(const timestamp &start_time,
      const eventhandler::ptr &handler,
      const sst_message::ptr &msg) = 0;
};


} // end of namespace sstmac
#endif
