/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/logger.h>

#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/timestamp.h>
#include <deque>

namespace sstmac {

  using namespace std;

  inline std::string TrimStr(const std::string& Src, const std::string& c =
      " \r\n")
  {
    int p2 = Src.find_last_not_of(c);
    if (p2 == std::string::npos)
      return std::string();
    int p1 = Src.find_first_not_of(c);
    if (p1 == std::string::npos)
      p1 = 0;
    return Src.substr(p1, (p2 - p1) + 1);
  }

  //std::string logger::user_param = "";
  std::deque<std::string> logger::user_params;
  std::list<logger*> logger::loggers_;


  const char* logger::black = "\033[0;30m";
  const char* logger::red = "\033[0;31m";
  const char* logger::green = "\033[0;32m";
  const char* logger::yellow = "\033[0;33m";
  const char* logger::blue = "\033[0;34m";
  const char* logger::magenta = "\033[0;35m";
  const char* logger::cyan = "\033[0;36m";
  const char* logger::white = "\033[0;37m";

  bool logger::debug_coloring_ = false;

  logger::logger(std::string namesig, std::ostream *os, bool flush,
      const char* color) :
      active_(false), outstream_(os), my_color_(color), always_flush_(
          flush), is_real_(true)
  {
    size_t left = namesig.find("<");
    size_t right = namesig.find(">");
    if(left != string::npos && right != string::npos){
        category_ = namesig.substr(left+1, right-left - 1);
        signature_ = namesig.substr(right + 1);
    }else{
        category_ = "";
        signature_ = namesig;
    }

   //  std::cout << "cat: " << category_ << std::endl;
   // std::cout << "sig: " << namesig << std::endl;

   // loggers_.push_back(this);
    update_active();
  }

  void logger::update_active()
  {
  //  std::cout << "-------- logger: UPDATE ACTIVE: " << " ------- " << std::endl;
    std::deque<std::string> tok;

    std::string space = "|";
    pst::BasicStringTokenizer::tokenize(signature_, tok, space);

    std::deque<std::string>::iterator it, end = tok.end();
    std::deque<std::string>::iterator paramit, paramend = user_params.end();

    // std::cout << "user param: " << user_param << std::endl;

    for (it = tok.begin(); it != end; it++)
      {
        std::string temp = *it;

        temp = TrimStr(temp);

        // std::cout << "tok: " << temp << std::endl;



        for (paramit = user_params.begin(); paramit != paramend; paramit++)
          {
            std::string t2 = *paramit;


            size_t left = t2.find("<");
            size_t right = t2.find(">");
            string user_cat = "";
            string user_name = t2;
            if(left != string::npos && right != string::npos){
                user_cat = t2.substr(left + 1, right - left - 1);
                user_name = t2.substr(right + 1);
            }

            user_name = TrimStr(user_name);

           // std::cout << "user tok: " << t2 << std::endl;
            if(user_cat == "" || user_cat == category_){
                if (user_name == temp || user_name == "all")
                  {
                    active_ = true;
                    return;
                  }
            }
          }

      }

    //std::cout << "done with update_active()" << std::endl;
  }

  void logger::update_all_active()
  {
    std::list<logger*>::iterator it, end = loggers_.end();
    for (it = loggers_.begin(); it != end; it++)
      {
        (*it)->update_active();
      }
    //std::cout << "done with update_all_active()" << std::endl;
  }

  logger::~logger()
  {

    //std::cout << "logger destructor: " << signature_ << std::endl;
    //loggers_.remove(this);
  }

  void logger::set_user_param(std::string s)
  {
    string tosep = s;


    std::deque<std::string> tok;

    std::string space = "|";
    pst::BasicStringTokenizer::tokenize(tosep, tok, space);
    user_params = tok;

    update_all_active();
  }


  void logger::clear(){
    loggers_.clear();
  }

  logger&
  logger::operator<<(int i)
  {

    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(long i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(size_t i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(long long int i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(unsigned long long int i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(bool i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(const char* i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(double i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(std::string i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(std::pair<int, int> p)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << "(" << p.first << ", " << p.second << ")"
      << black;
    else
      (*outstream_) << "(" << p.first << ", " << p.second << ")";

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }

  logger&
  logger::operator<<(timestamp i)
  {
    if (debug_coloring_)
      (*outstream_) << my_color_ << i << black;
    else
      (*outstream_) << i;

    if (always_flush_)
      outstream_->flush();

    return (*this);
  }
} // end of namespace sstmac
