/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_DRIVER_UTIL_H
#define SSTMAC_DRIVER_UTIL_H

#include <sstmac/sstmac.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/timestamp.h>
#include <ostream>

namespace sstmac {


  using namespace sstmac::hw;
  using namespace sstmac;
  using namespace sstmac::native;
  using namespace sstmac::sw;


  /**
   * The topologies that can be constructed
   */
  enum topologytype {
    TORUS, CROSSBAR, GAMMA, FATTREE, TORUS3, RING, TBIRD, CLOS
  };

  /**
   * A way to convert a topologytype back to a string, for printing help
   * @param os
   * @param nt
   * @return
   */
  static
  std::ostream& operator<<(std::ostream &os, const topologytype &nt) {
    switch(nt) {
    case TORUS:    os << "TORUS";    break;
    case CROSSBAR: os << "CROSSBAR"; break;
    case GAMMA:    os << "GAMMA";    break;
    case FATTREE:  os << "FATTREE";  break;
    case TORUS3:   os << "TORUS3";   break;
    case RING:     os << "RING";     break;
    case TBIRD:    os << "TBIRD";    break;
    case CLOS:     os << "CLOS";     break;
    default:       os << "GARBAGE";  break;

    }
    return os;
  }

  /**
   * The circuit types that can be constructed
   */
  enum circuittype {
    EXCLUSIVE, SHARED, BUFFER
  };

  /**
   * A way to convert a circuittype back to a string, for printing help
   * @param os
   * @param ct
   * @return
   */
  inline
  std::ostream& operator<<(std::ostream &os, const circuittype &ct) {
    switch(ct) {
    case EXCLUSIVE: os << "EXCLUSIVE"; break;
    case SHARED:    os << "SHARED";    break;
    case BUFFER:    os << "BUFFER";    break;
    default:        os << "GARBAGE";   break;
    }
    return os;
  }

  /// The network sharing models recognized by shared-bandwidth interconnects.
   enum netsharing {
     OVERCOMMIT, // Turn the network into a crossbar.
     THROTTLE_OVERCOMMITS,
     ORDERED
   };

   /**
    * A way to convert a netsharing back to a string, for printing help
    * @param os
    * @param ns
    * @return
    */
   inline std::ostream& operator<<(std::ostream &os, netsharing ns) {
     os << "netsharing(";
     switch(ns) {
     case OVERCOMMIT: os << "OVERCOMMIT"; break;
     case THROTTLE_OVERCOMMITS: os << "THROTTLE_OVERCOMMITS"; break;
     case ORDERED: os << "ORDERED"; break;
     default: os << "INVALID_VALUE=" << int(ns);
     }
     os << ")";
     return os;
   }

   /**
    * Constructs a simple simulation, assuming some parameters for the nodes.
    * TODO: make this general and all parameterized
    * @param nodes number of physical nodes to make
    * @param ppn number of processes (tasks) per node to instantiate
    * @param appname name of app, referenced in software/skeletons/skeletonfactory
    * @param app_params list of string to pass to skeletons as parameters
    * @param fab the interconnect model already constructed
    * @return simulation finish time
    */
  static timestamp
  run_simple_sim(int nodes, int ppn, std::string appname, std::list<std::string> app_params, interconnect::ptr fab, indexstrategy::ptr ind = blockindexing::construct(),
		  double memory_bandwidth = 1e11, timestamp memory_latency = timestamp(1e-10))
  {

    manager::ptr mgr = manager::construct();

    mgr->set_interconnect(fab);
    //mgr->set_eventmanager(eventmap::construct());


    simplenode::ptr node = simplenode::construct(memory_latency,
        memory_bandwidth); // bandwidth in bytes/second
    simplenic::ptr nic = simplenic::construct(node, fab);
    mgr->add_nodes(node, nic, nodes);

   launchinfo::ptr l1 = launchinfo::construct(appname, appid(0), app_params, timestamp(0), nodes * ppn);
   std::list<launchinfo::ptr> apps;
   apps.push_back(l1);


   allocationstrategy::ptr alloc = firstavailableallocation::construct(nodes);
    // Construct a launcher
    instantlaunch::ptr job = instantlaunch::construct(apps, alloc, ind);


    timestamp t = mgr->run(job);
    //std::cout << "end time is " << t << std::endl;
    return t;
  }

  /**
     * Parse a string to find a corresponding topologytype enum value
     * @param value name of the topology
     * @param errorflag set to true if nothing matched
     * @return a standard topologytype value
     */
  static
  netsharing parse_flower(const char *value, bool &errorflag) {
    std::string arg = value;
    for(unsigned i = 0; i < arg.size(); ++i) arg[i] = tolower(arg[i]);
    netsharing ns = ORDERED;
    if(arg == "overcommit") ns = OVERCOMMIT;
    else if(arg == "throttle_overcommits" || arg == "throttle-overcommits" ||
        arg == "throttle")
      ns = THROTTLE_OVERCOMMITS;
    else if(arg == "ordered")
      ns = ORDERED;
    else {
        errorflag = true;
        std::cerr << "Invalid netsharing type: " << value << "\n";
    }
    return ns;
  }

  /**
   * Parse a string to find a corresponding topologytype enum value
   * @param value name of the topology
   * @param errorflag set to true if nothing matched
   * @return a standard topologytype value
   */
  static
  topologytype parse_topology(const char *value, bool &errorflag) {
    // TORUS, CROSSBAR, GAMMA, FATTREE, TORUS3
    std::string arg = value;
    for(unsigned i = 0; i < arg.size(); ++i) arg[i] = tolower(arg[i]);
    topologytype nt = TORUS;
    if(arg == "torus") nt = TORUS;
    else if(arg == "crossbar" || arg == "xbar" || arg == "cross") nt = CROSSBAR;
    else if(arg == "gamma" || arg == "gammagraph" || arg == "gamma-graph")
      nt = GAMMA;
    else if(arg == "fattree" || arg == "fat-tree" || arg == "fat") nt = FATTREE;
    else if(arg == "torus3" || arg == "torus3d" || arg == "3dtorus" ||
        arg == "3d-torus")
      nt = TORUS3;
    else if(arg == "ring") nt = RING;
    else if(arg == "clos") nt = CLOS;
    else if(arg == "tbird") nt = TBIRD;
    else {
        errorflag = true;
        std::cerr << "Invalid network type: " << value;
    }
    return nt;
  }

  /**
   * Parses a string to find a corresponding circuittype enum value
   * @param value name of the circuit model
   * @param errorflag set to true if nothing matched
   * @return a standard circuittype value
   */
  static
  circuittype parse_circuit(const char *value, bool &errorflag) {
    // EXCLUSIVE, SHARED
    std::string arg(value);
    for(unsigned i = 0; i < arg.size(); ++i) arg[i] = tolower(arg[i]);
    circuittype ct = SHARED;
    if(arg == "exclusive") ct = EXCLUSIVE;
    else if(arg == "shared") ct = SHARED;
    else if(arg == "buffer") ct = BUFFER;
    else {
        errorflag = true;
        std::cerr << "Invalid circuit type: " << value << "\n";
    }
    return ct;
  }


  /**
   * Construct a topology from a given string name
   * @param args
   * @param netname
   * @return
   */
  static
  topology::ptr
  construct_topology(std::vector<int> &args, const std::string &netname)
  {

    topology::ptr net;
    if (netname == "gamma") {
        if (args.size() == 2) {
            net = sstmac::hw::gamma::construct(args[0],args[1]);
        }
        else if (args.size() == 3) {
            net = sstmac::hw::gamma::construct(args[0],args[1],args[2]);
        }
        else if (args.size() == 4) {
            net = sstmac::hw::gamma::construct(args[0],args[1],args[2],args[3]);
        }
    }
    else if (netname == "hypercube") {
        if (args.size() == 1) {
            net = sstmac::hw::hypercube::construct(args[0]);
        }
    }
    else if (netname == "torus2D") {
        if (args.size() == 2) {
            net = sstmac::hw::torus::construct_2D(args[0],args[1]);
        }
        else if (args.size() == 3) {
            net = sstmac::hw::torus::construct_2D(args[0],args[1],args[2]);
        }
        else if (args.size() == 4) {
            net = sstmac::hw::torus::construct_2D(args[0],args[1],args[2],args[3]);
        }
    }
    else if (netname == "torus3D") {
        if (args.size() == 3) {
            net = sstmac::hw::torus::construct_3D(args[0],args[1],args[2]);
        }
        else if (args.size() == 4) {
            net = sstmac::hw::torus::construct_3D(args[0],args[1],args[2],args[3]);
        }
        else if (args.size() == 5) {
            net = sstmac::hw::torus::construct_3D(args[0],args[1],args[2],args[3],args[4]);
        }
    }
    else if (netname == "hdtorus2D") {
        if (args.size() == 2) {
            net = sstmac::hw::hdtorus::construct_2D(args[0],args[1],1,1);
        }
        else if (args.size() == 3) {
            net = sstmac::hw::hdtorus::construct_2D(args[0],args[1],args[2],1);
        }
        else if (args.size() == 4) {
            net = sstmac::hw::hdtorus::construct_2D(args[0],args[1],args[2],args[3]);
        }
    }
    else if (netname == "hdtorus3D") {
        if (args.size() == 3) {
            net = sstmac::hw::hdtorus::construct_3D(args[0],args[1],args[2],1,1);
        }
        else if (args.size() == 4) {
            net = sstmac::hw::hdtorus::construct_3D(args[0],args[1],args[2],args[3],1);
        }
        else if (args.size() == 5) {
            net = sstmac::hw::hdtorus::construct_3D(args[0],args[1],args[2],args[3],args[4]);
        }
    }
    else if (netname == "ring") {
        if (args.size() == 1) {
            net = sstmac::hw::ring::construct(args[0]);
        }
        else if (args.size() == 2) {
            net = sstmac::hw::ring::construct(args[0],args[1]);
        }
        else if (args.size() == 3) {
            net = sstmac::hw::ring::construct(args[0],args[1],args[2]);
        }
    }
    else if (netname == "fattree") {
        if (args.size() == 2) {
            net = sstmac::hw::fattree::construct(args[0],args[1]);
        }
        if (args.size() == 3) {
            net = sstmac::hw::fattree::construct(args[0],args[1],args[2]);
        }
        if (args.size() == 4) {
            net = sstmac::hw::fattree::construct(args[0],args[1],args[2],args[3]);
        }
    }
    else if (netname == "crossbar") {
        if (args.size() == 1) {
            net = sstmac::hw::crossbar::construct(args[0]);
        }
    }
    else if (netname == "clos") {
        if (args.size() == 1) {
            net = sstmac::hw::clos::construct(args[0]);
        }
    }
    else if (netname == "tbird") {
        if (args.size() == 0) {
            net = sstmac::hw::tbird::construct();
        }
    }

    return net;
  }

}
#endif
