/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_CIRCUIT_EDGE_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_CIRCUIT_EDGE_H_INCLUDED


#include <sstmac/hardware/network/congestion/circuit-transfer.h>
#include <sstmac/common/debug.h>
#include <iostream>
#include <list>

namespace sstmac { namespace hw  {

  /**
   * A handler to manage the state of an edge in a circuit network
   * (reserved vs. available).  Requests for this edge are managed in
   * a FIFO queue.
   *
   * Unlike so many other classes in the simulator, this one is not indended
   * to be individuall reference counted (and must be copy-constructable).
   */
  class circuit::edge {
    /// The state of this edge.
    int nchannel_;
    int nreserved_;

    debug dbg_;

    /// The queue of callers waiting to receive this edge.
    /// Objects are removed from this queue when it is their turn.
    std::list< transfer::ptr > pending_;
    
  public:
    /// Hi.
    edge(int nchannel = 1) : dbg_("<debug> interconnect"),
      nchannel_(nchannel),
      nreserved_(0)
    {}

    /// I just cannot bring myself to use the default (or non-virtual) dtor.
    virtual ~edge() throw() {
      if(nreserved_ > 0) {
        SSTMAC_DEBUG << "WARNING:  ~edge:  Going out of scope while reserved.\n";
      }
    }

    /// Put in a request for this edge.
    void reserve(const transfer::ptr &caller) {
      SSTMAC_DEBUG << "circuit::edge[" << this << "]::reserve(" << caller << ")\n";
      if(nreserved_ < nchannel_) {
        SSTMAC_DEBUG << "    edge given to caller " << caller << "\n";
        nreserved_++;
        caller->edge_available(this);
      }
      else {
        SSTMAC_DEBUG << "    caller " << caller << " added to request stack\n";
        pending_.push_back(caller);
      }
    }

    /// Release current reservation on this edge.
    void release() {
      SSTMAC_DEBUG << "circuit::edge[" << this << "]::release\n";
      nreserved_--;
      if(! pending_.empty()) {
        transfer::ptr next = pending_.front();
        pending_.pop_front();
        nreserved_++;
        SSTMAC_DEBUG << "    edge given to caller " << next
                << " " << nreserved_ << " of " << nchannel_
                << "channels in use\n";
        next->edge_available(this);
      }
      else {
        SSTMAC_DEBUG << ":   no pending callers.  Edge marked available\n";
      }
    }

    /// Set the number of channels that are available for this edge.
    void set_nchannel(int nchannel) { nchannel_ = nchannel; }

  };
  
} } // end of namespace sstmac

#endif
