/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/circuit.h>
#include <sstmac/hardware/network/congestion/circuit-edge.h>

//#include <sstmac/simulator/eventhandler.h>
#include <sstmac/hardware/network/congestion/constantlatency.h>
//#include <sstmac/hardware/network/congestion/transfer_time.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/topology/walk.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <boost/lexical_cast.hpp>

namespace sstmac { namespace hw 
  {

  // Pointer types.


  using boost::lexical_cast;

  //
  // Hi.
  //
  circuit::circuit(const latencymodel::ptr &latency, bandwidth bw,
      const topology::ptr &net,
      const power_config::ptr &pwrcfg) : interconnect(net),
       latency_(latency), bandwidth_(bw), timing_model_(
          ORIGINAL)
  {
    SSTMAC_DEBUG << "circuit(latency=" << latency->to_string() << ", bandwidth="
        << bandwidth_ << ", net=" << net << ")\n";
    if (latency_ == 0)
      {
        throw nullerror("circuit(...):  Null latencymodel pointer.");
      }
    if (bw.bits_per_second() <= 0)
      throw valueerror("circuit(...):  Zero or negative bandwidth.");
    if (topology_ == 0)
      {
        throw nullerror("circuit(...):  null network pointer.");
      }

    //std::cerr << "circuit:  Making room for "
    //          << network_->g().n_edges() << " edges\n";

    edge_.resize(topology_->g().n_edges());

    for (size_t i = 0; i < edge_.size(); i++)
      {
        edge_index ei(i);
        edge_[i].set_nchannel(topology_->edge_width(ei));
      }

    pwr_config_ = pwrcfg;

    if(pwrstat_->is_active())
      pwrstat_->record(pwr_config_->get_rtr_static() * topology_->n());




    if(dbg_.is_active()) {
        size_t i = 0;
        graph::edge_iterator it, end=topology_->g().edges_end();

        for(it = topology_->g().edges_begin(); it != end; ++it) {
            dbg_ << "circuit::circuit:  edge " << &edge_.at(i)
                 << " connects " << it->from() << " and " << it->to() << "\n";
            ++i;
        }

    }

  }



  //
  // New circuit network.
  //
  circuit::ptr circuit::construct(const timestamp latency, double bw,
      const topology::ptr &net)
  {

    return construct(latency, bandwidth(bw, bandwidth::Bps), net);
  }

  //
  // New circuit network
  //
  circuit::ptr circuit::construct(const timestamp latency, bandwidth bw,
      const topology::ptr &net)
  {

    if (latency < timestamp(0))
      {
        throw valueerror("circuit(...):  negative latency.");
      }
    constantlatency::ptr lat = constantlatency::construct(latency);
    return construct(lat, bw, net, power_config::construct());
  }

  //
  // New circuit with a user-defined latency model - no power config
  //
  circuit::ptr circuit::construct(
      const latencymodel::ptr &lat, bandwidth bw,
      const topology::ptr &net)
  {

    circuit::ptr retval(
        new circuit(lat, bw, net, power_config::construct()));

    return retval;
  }

  //
  // New circuit with a user-defined latency model.
  //
  circuit::ptr circuit::construct(
      const latencymodel::ptr &lat, bandwidth bw,
      const topology::ptr &net,
      const power_config::ptr &pwrcfg)
  {

    circuit::ptr retval(new circuit(lat, bw, net, pwrcfg));

    return retval;
  }

  //
  // Goodbye.
  //
  circuit::~circuit() throw ()
    {
    }

  //
  // Required method from parent.
  //
  int64_t circuit::slots() const
  {
    return topology_->n();
  }



  //
  // Send the given message at the given time.
  //
  void circuit::send(const timestamp &time, const sst_message::ptr &msg,
      const eventhandler::ptr& sendhandler,
      const eventhandler::ptr& recvhandler)
  {
    int bytes = msg->get_byte_length();
    std::string me("circuit");
    me += lexical_cast<std::string>(size_t(this)) + "::send";
    SSTMAC_DEBUG << me << "(time=" << time << ", bytes=" << bytes << ", sender="
        << msg->fromaddr_ << ", recver=" << msg->toaddr_ << ", sendhandler="
        << sendhandler << ", recvhandler=" << recvhandler << ")\n";
    walk the_walk;
    the_walk.ev().reserve(topology_->diameter());
    topology_->route(msg->fromaddr_, msg->toaddr_, the_walk);

    SSTMAC_DEBUG << me << the_walk.length() << " step walk from " << msg->fromaddr_
        << " to " << msg->toaddr_ << ": [";
    for (int i = 0; i < the_walk.length(); ++i)
      {
        SSTMAC_DEBUG << the_walk.e(i).from() << " --> ";
        if (i > 0)
          {
            if (the_walk.e(i).from() != the_walk.e(i - 1).to())
              {
                SSTMAC_DEBUG << "ERROR:  Non-contiguous walk\n";
                abort();
              }
          }
      }



    if (dbg_.is_active())
      {
        if (the_walk.length() > 0)
          {
            dbg_ << the_walk.e(the_walk.length() - 1).to() << "]\n";
          }
        else
          {
            dbg_ << "]\n";
          }
      }

    // Timing model stuff shifted down from constructor to permit
    // noisy latency (and eventually bandwidth) models.
    timestamp start, stime, rtime, ttime;
    data payload(bytes, data::Bytes);
    switch (timing_model_)
    {
    case ORIGINAL:
      original_timing_model(payload, start, stime, rtime, ttime);
      break;
    case SPLIT:
      split_timing_model(payload, start, stime, rtime, ttime);
      break;
    default:
      throw valueerror("circuit::send:  Invalid timing model.");
    }
    // Ready to roll
    transfer::ptr trans = transfer::construct(ptr(this), msg,
        the_walk, stime, rtime, ttime, sendhandler, recvhandler);

    SSTMAC_DEBUG << me << ":  Scheduling send for " << (time + start) << "\n";
    schedule(time + start, trans, msg);
  }

  void circuit::original_timing_model(data payload, timestamp &start_time,
      timestamp &send_time, timestamp &recv_time, timestamp &trans_time)
  {
    // Original timing model has no startup cost.
    start_time = timestamp(0);
    // send time is delayed by full latency.
    // The latency model decides what the latency is for any given transfer
    timestamp lat = latency_->latency(payload);
    send_time = lat + payload / bandwidth_;

    /// The model assumes that the sender and receiver finish at the same time.
    recv_time = send_time;

    /// Transfer begins immediately (although it should include a delay).
    trans_time = send_time;
  }

  void circuit::split_timing_model(data payload, timestamp &start_time,
      timestamp &send_time, timestamp &recv_time, timestamp &trans_time)
  {
    /// Split timing model divides the latency onto the sender and receiver.
    timestamp lat = latency_->latency(payload);
    timestamp halflat = lat / 2;

    // Start after half latency.
    start_time = halflat;
    // Transfer time is pure bandwidth contribution.
    trans_time = payload / bandwidth_;
    send_time = trans_time;
    // and receive time is after transfer and half latency.
    recv_time = trans_time + halflat;
  }

  } } // end of namespace sstmac
