/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-arrival.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-departure.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-flowregistration.h>
//#include <sstmac/simulator/eventhandler.h>
//#include <sstmac/simulator/eventmanager.h>
#include <sstmac/hardware/network/congestion/data.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/topology/graph.h>
#include <sstmac/hardware/network/topology/walk.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <algorithm>
#include <math.h>

namespace sstmac { namespace hw 
  {



  //
  // Construct a new sharedbuffer
  //
  sharedbuffer::sharedbuffer(double router_router_bw, double node_router_bw,
      const timestamp node_lat, const timestamp link_lat,
      const timestamp router_lat, double buffer_capacity, int num_vcs,
      const topology::ptr &net) : interconnect(net),
      router_router_bw_(router_router_bw), node_router_bw_(node_router_bw), node_lat_(
          node_lat), link_lat_(link_lat), router_lat_(router_lat), fid_counter_(0), buffer_capacity_(buffer_capacity), num_vcs_(
              num_vcs)
  {
    pwr_config_ = power_config::construct();


    if(pwrstat_->is_active())
        pwrstat_->record(pwr_config_->get_rtr_static() * topology_->n());



    int64_t nlinks = net->nlink();
    edge_bw_.resize(nlinks);
    edge_max_bw_.resize(nlinks);
    available_vcs_.resize(nlinks);
    flows_blocked_on_vc_.resize(nlinks);
    for (int i = 0; i < nlinks; i++)
      {

        available_vcs_[i] = num_vcs;

        edge_index ei(i);
        const lwedge &e = net->g().index_to_edge(ei);
        if (net->is_node(e.from()) || net->is_node(e.to()))
          {
            edge_bw_[i] = node_router_bw;
            edge_max_bw_[i] = node_router_bw;
          }
        else
          {
            edge_bw_[i] = router_router_bw;
            edge_max_bw_[i] = router_router_bw;
          }
      }

    int64_t nvertices = net->g().n_vertices();
    flow_registrations_.resize(nvertices);
    vertices_departed_.resize(nvertices);

    // this may not be correct
    bw_epsilon_ = buffer_capacity / timestamp::max_time();
    buffer_epsilon_ = 10 * std::numeric_limits<double>::epsilon();
  }

  //
  // Construct a new sharedbuffer
  //
  sharedbuffer::ptr sharedbuffer::construct(double router_router_bw,
      double node_router_bw, const timestamp node_lat,
      const timestamp link_lat, const timestamp router_lat,
      double buffer_capacity, int num_vcs, const topology::ptr &net)
  {
    sharedbuffer::ptr retval(
        new sharedbuffer(router_router_bw, node_router_bw, node_lat,
            link_lat, router_lat, buffer_capacity, num_vcs, net));

    return retval;
  }

  //
  // Delete me.
  //
  sharedbuffer::~sharedbuffer() throw ()
    {
    }

  //
  // Slots.
  //
  int64_t sharedbuffer::slots() const
  {
    return topology_->n();
  }




  //
  // Send a message from sender to recver.
  //
  void sharedbuffer::send(const timestamp &start_time,
      const sst_message::ptr &msg,
      const eventhandler::ptr &sendhandler,
      const eventhandler::ptr &recvhandler)
  {
    int bytes = msg->get_byte_length();
    // construct path. all routing is deterministic/oblivious
    walk the_walk;
    the_walk.ev().reserve(topology_->diameter());
    topology_->route(msg->fromaddr_, msg->toaddr_, the_walk);

    for (int i = 0; i < the_walk.length(); ++i)
      if (i > 0)
        if (the_walk.e(i).from() != the_walk.e(i - 1).to())
          throw illformederror(
              "sharedcircuit::send:  Non-contigous walk.");

    // Get the edge indices.
    const std::vector<lwedge> &ev = the_walk.ev();
    std::vector<edge_index> edge_id(ev.size());
    for (size_t i = 0; i < ev.size(); ++i)
      {
        edge_id.at(i) = topology_->g().edge_to_index(ev.at(i));
      }

    // Create new fid for this flow
    flow_id_t fid = fid_counter_;
    fid_counter_++;

    SSTMAC_DEBUG << "sharedbuffer::send: Beginning send - fid: " << fid << "\n";

    // Insert this flow's path to the map.
    flow_edges_.insert(
        std::pair<flow_id_t, std::vector<edge_index> >(fid, edge_id));

    // Record the event send/recv handlers
    recv_handlers_.insert(
        std::pair<flow_id_t, eventhandler::ptr >(fid, recvhandler));
    send_handlers_.insert(
        std::pair<flow_id_t, eventhandler::ptr >(fid, sendhandler));

    recv_messages_.insert(
        std::pair<flow_id_t, sst_message::ptr >(fid, msg));

    // push message to first router. message is assumed fully buffered at injection.
    timestamp arrival_time = start_time + node_lat_;
    arrival::ptr arrive = arrival::construct(ptr(this), fid,
        the_walk.v(0), 0.0, static_cast<double>(bytes));
    schedule(arrival_time, arrive);

  }

  //
  // Update the buffer_occupancy_ of fr.
  //
  void sharedbuffer::update_buffer_usage(flowregistration::ptr fr,
      const timestamp &now)
  {

    double delta_bw = fr->input_bw_ - fr->output_bw_;
    timestamp delta_t(now - fr->last_buffer_update_time_);

    double delta_bytes = delta_bw * delta_t.sec();

    SSTMAC_DEBUG << "sharedbuffer::update_buffer_usage: inputbw " << fr->input_bw_
        << " outputbw " << fr->output_bw_ << " now " << now
        << " last update " << fr->last_buffer_update_time_
        << " old occ " << fr->buffer_occupancy_ << " delta_bytes "
        << delta_bytes << " new occ "
        << fr->buffer_occupancy_ + delta_bytes << "\n";

    fr->buffer_occupancy_ += delta_bytes;
    fr->last_buffer_update_time_ = now;

  }

  //
  // Determine if two doubles are equal within error bounds
  //
  bool sharedbuffer::are_equal(double a, double b)
  {

    double relative_error;
    if (fabs(a) > fabs(b))
      relative_error = fabs((a - b) / a);
    else
      relative_error = fabs((a - b) / b);

    if (relative_error < buffer_epsilon_)
      return true;
    else
      return false;

  }

  } } //end of namespace sstmac

