/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/sharedcircuit.h>
#include <sstmac/hardware/network/congestion/sharedcircuit-stopper.h>
#include <sstmac/hardware/network/congestion/sharedcircuit-starter.h>
#include <sstmac/hardware/network/congestion/constantlatency.h>
#include <sstmac/hardware/network/congestion/data.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/topology/graph.h>
#include <sstmac/hardware/network/topology/walk.h>
#include <sstmac/common/errors.h>

namespace sstmac { namespace hw 
  {



  //
  // Construct a new circuit.
  //
  sharedcircuit::sharedcircuit(const latencymodel::ptr &latency,
      bandwidth bw, const topology::ptr &net,
      const flower::ptr &flow,
      const power_config::ptr &pwrcfg) : interconnect(net),
      flower_(flow), stopper_(stopper::construct(ptr(this), flower_)), latency_(latency), bandwidth_(
          bw)
  {
    pwr_config_ = pwrcfg;


    if(pwrstat_->is_active())
      pwrstat_->record(pwr_config_->get_rtr_static() * topology_->n());


  }

  //
  // Construct a new circuit.
  //
  sharedcircuit::ptr sharedcircuit::construct(const timestamp latency,
      double bw, const topology::ptr &net,
      const flower::ptr &flow)
  {
    sharedcircuit::ptr ret = construct(
        constantlatency::construct(latency), bandwidth(bw, bandwidth::Bps),
        net, flow, power_config::construct());

    return ret;
  }

  //
  // Construct a new circuit.
  //
  sharedcircuit::ptr sharedcircuit::construct(const timestamp latency,
      double bw, const topology::ptr &net,
      const flower::ptr &flow,
      const power_config::ptr &pwrcfg)
  {
    sharedcircuit::ptr ret = construct(
        constantlatency::construct(latency), bandwidth(bw, bandwidth::Bps),
        net, flow, pwrcfg);

    return ret;
  }

  //
  // Construct a new circuit.
  //
  sharedcircuit::ptr sharedcircuit::construct(
      const latencymodel::ptr &latency, bandwidth bw,
      const topology::ptr &net, const flower::ptr &flow,
      const power_config::ptr &pwrcfg)
  {
    sharedcircuit::ptr ret = sharedcircuit::ptr(
        new sharedcircuit(latency, bw, net, flow, pwrcfg));

    return ret;
  }

  //
  // So long, Mr. Despair.
  //
  sharedcircuit::~sharedcircuit() throw ()
    {
    }

  //
  // Required method from parent.
  //
  int64_t sharedcircuit::slots() const
  {
    return topology_->n();
  }

  //


  //
  // Send the given message at the given time.
  //
  void sharedcircuit::send(const timestamp &start_time,
      const sst_message::ptr &msg,

      const eventhandler::ptr &sendhandler,
      const eventhandler::ptr &recvhandler)
  {
    int bytes = msg->get_byte_length();
    // Decide what the latency will be for this transfer.
    data payload(bytes, data::Bytes);
    timestamp latency = latency_->latency(payload);
    timestamp halflat = latency / 2;
    timestamp start_at(start_time + halflat);
    walk the_walk;
    the_walk.ev().reserve(topology_->diameter());
    SSTMAC_DEBUG << "sharedcircuit::send - routing..\n";

    topology_->route(msg->fromaddr_, msg->toaddr_, the_walk);

    SSTMAC_DEBUG << "sharedcircuit::send - checking for a continuous walk \n";
    for (int i = 0; i < the_walk.length(); ++i)
      if (i > 0)
        if (the_walk.e(i).from() != the_walk.e(i - 1).to())
          throw illformederror(
              "sharedcircuit::send:  Non-contigous walk.");

    // Get the edge indices.
    const std::vector<lwedge> &ev = the_walk.ev();
    std::vector<edge_id_t> edge_id(ev.size());
    for (size_t i = 0; i < ev.size(); ++i)
      {
        //edge_id.at(i) = ev.at(i).id();  // THIS YIELDS GARBAGE.
        edge_id.at(i) = topology_->g().edge_to_index(ev.at(i)).value();
      }

    SSTMAC_DEBUG << "sharedcircuit::send - constructing starter \n";
    starter::ptr trans = starter::construct(ptr(this), flower_, stopper_, bytes,
        halflat, edge_id, sendhandler, recvhandler);

    SSTMAC_DEBUG << "sharedcircuit::send  - tracking power \n";



    SSTMAC_DEBUG << "sharedcircuit::send - scheduling transfer \n";
    schedule(start_at, trans, msg);
  }

  } } // end of namespace sstmac
