/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// graph.h: Interface for generic graphs.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_GRAPH_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_GRAPH_H_INCLUDED

#include <sstmac/common/sstmac-config.h>
#include <set>
#include <map>
#include <iosfwd>
#ifdef SSTMAC_HAVE_BOOST_1_37
  #include <boost/unordered_map.hpp>
#endif

#include <sstmac/hardware/network/topology/edge.h>
#include <sstmac/hardware/network/topology/walk.h>

namespace sstmac { namespace hw  {

/** A graph is defined by a set of edges.  Various other quantities are also
    stored with the graph to expedite operations on the graph. Each edge
    and vertex is associated with an integer identifier.
*/
class graph {
  private:
    std::vector<vertex> iv_;
    std::vector<lwedge> ie_;
    std::vector<std::vector<vertex_index> > in_, out_;

#ifdef SSTMAC_HAVE_BOOST_1_37
    // unordered containers for performance
    typedef boost::unordered_map<lwedge, edge_index,
                                 lwedge::lwedge_hash> edge_map;
  #ifdef SSTMAC_HAVE_LIBZ
    typedef boost::unordered_map<vertex, vertex_index,
                                 vertex::vertex_crc32> vertex_map;
  #else
    typedef std::map<vertex, vertex_index> vertex_map;
  #endif
#else
    typedef std::map<lwedge, edge_index> edge_map;
    typedef std::map<vertex, vertex_index> vertex_map;
#endif
    edge_map uei_;
    vertex_map uvi_;

  public:

    /** Using these typedefs protects against future changes
        to edge container.  They are public so other classes 
        which call xxx_begin() or xxx_end() can use them. */
    typedef std::vector<lwedge>::const_iterator edge_iterator;
    typedef std::vector<vertex>::const_iterator vertex_iterator;

    /// Create an empty graph.
    graph();
    /// Create a graph with the given edge and vertex sets.
    graph(const std::set<vertex> &v, const std::set<edge> &a);
    /** Initialize the graph to the given edge and vertex sets.
        The old graph data is first discarded. */
    void init(const std::set<vertex> &v, const std::set<edge> &a);

    /// Return beginning vertex iterator
    const vertex_iterator vertices_begin() const { return iv_.begin(); }

    /// Return ending vertex iterator
    const vertex_iterator vertices_end() const {return iv_.end(); }

    /// Return beginning edge iterator
    const edge_iterator edges_begin() const { return ie_.begin(); }

    /// Return ending edge iterator
    const edge_iterator edges_end() const {return ie_.end(); }

    /// Return the vertex corresponding to the given index.
    const vertex &index_to_vertex(const vertex_index &i) const {
      return iv_[i.value()];
    }

    /** Return the index corresponding to the given vertex. This
        will throw if the vertex is not found. */
    vertex_index vertex_to_index(const vertex & ) const;

    /** Find the integer index for the given edge. This will throw
        if the edge is not found. */
    edge_index edge_to_index(const lwedge &) const;

    /// Return the edge corresponding to the given index.
    const lwedge &index_to_edge(const edge_index &i) const { return ie_[i.value()]; }

    /// Return the number of vertices in the graph.
    int64_t order() const { return iv_.size(); }

    /// Return the number of edges in the graph.
    int64_t n_edges() const { return ie_.size(); }
 
    /// Return the number of vertices in the graph.
    int64_t n_vertices() const { return iv_.size(); }

    /** Returns the maximum number of edges that have the same vertex
        as the end point. */
    int64_t in_degree() const;

    /** Return the maximum number of edges that originate at the same
        vertex. */
    int64_t out_degree() const;

    /// Returns the maximum of in_degree and out_degree.
    int64_t degree() const;

    /// Returns the number of integers that define a vertex.
    int64_t word_length() const;

    /// Verifies that the given walk is valid for this graph.
    void check_walk(const vertex &v0, const vertex &v1, const walk &w) const;

    /// Print basis information about the graph.
    void print(std::ostream &o = std::cout) const;

    /// Print all edges and vertices in the graph.
    void print_all(std::ostream &o = std::cout, bool verbose=false) const;

    /// Print the graph as input to the graphviz dot program.
    void print_dot(std::ostream &o = std::cout) const;

    /// Return the outbound adjacency information for all vertices.
    const std::vector<std::vector<vertex_index> > & in() const { return in_; }

    /// Return the inbound adjacency information for all vertices.
    const std::vector<std::vector<vertex_index> > & out() const { return out_; }

    /// Return an lwedge given two vertices.
    lwedge generate_lwedge(const vertex &v1, const vertex &v2, int id=0) const {
      return lwedge(vertex_to_index(v1),vertex_to_index(v2),id);
    }

    /// Return an lwedge given two vertices.
    lwedge generate_lwedge(vertex_index v1, const vertex &v2, int id=0) const {
      return lwedge(v1,vertex_to_index(v2),id);
    }

    /// Return an lwedge given two vertices.
    lwedge generate_lwedge(const vertex &v1, vertex_index v2, int id=0) const {
      return lwedge(vertex_to_index(v1),v2,id);
    }

    /// Return an lwedge given two vertices.
    lwedge generate_lwedge(vertex_index v1, vertex_index v2, int id=0) const {
      return lwedge(v1,v2,id);
    }
};

} } //end of namespace sstmac

#endif
