/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// network.cc: Implementation for networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <stdexcept>
#include <sstream>
#include <sstmac/hardware/network/topology/topology.h>
#include <algorithm>
#include <sstmac/common/rng.h>

using std::logic_error;
using std::set;
using std::string;
using std::vector;

namespace sstmac { namespace hw  {

///////////////////////////////////////////////////////////////////////
// network

  topology::~topology()
{
}

void
topology::init()
{
  int index = 0;
  inodes_.clear();
  for (set<vertex>::iterator i = nodes_.begin();
       i != nodes_.end(); i++,index++) {
      inodes_.push_back(*i);
      nodesi_.insert(std::make_pair(*i,index));
    }
}

int64_t
topology::shortest_path_length(const vertex &v1, const vertex &v2) const
{
  walk w; shortest_path(v1,v2,w);
  return w.length();
}

double
topology::a() const
{
  double sum = 0.0;
  for (graph::vertex_iterator i = g().vertices_begin();
       i != g().vertices_end();
       i++) {
      for (graph::vertex_iterator j = g().vertices_begin();
           j != g().vertices_end();
           j++) {
          sum += shortest_path_length(*i,*j);
        }
      if (vertex_symmetric()) return sum/g().n_vertices();
    }
  return sum/(g().n_vertices() * g().n_vertices());
}

int64_t
topology::b() const
{
  throw logic_error("topology::b() not implemented");
}

bool
topology::vertex_symmetric() const
{
  throw logic_error("topology::vertex_symmetric() not implemented");
}

void
topology::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
{
  throw logic_error("topology::shortest path() not implemented");
}

void
topology::next_hop(vertex_index loc, vertex_index dst,
                  std::vector<lwedge> &nxt) const
{
  throw logic_error("topology::next_hops() not implemented");
}

void
topology::route(const vertex &v1, const vertex &v2, walk &w) const
{
  shortest_path(v1,v2,w);
}

int64_t
topology::nswitch() const
{
  int64_t nnode = n();
  int64_t nvert = g().order();
  if (nnode == nvert) return nnode;
  return nvert-nnode;
}

bool
topology::direct() const
{
  return g().order() == n();
}

int64_t
topology::nlink() const
{
  int64_t nedge = g().n_edges();
  return nedge;
}

void
topology::route(const nodeaddress::ptr &n1, const nodeaddress::ptr &n2, walk &w) const
{
  const vertex &v1 = inodes_[n1->unique_id()];
  const vertex &v2 = inodes_[n2->unique_id()];
  route(v1,v2,w);
}

void
topology::route(int n1, int n2, walk &w) const
{
  const vertex &v1 = inodes_[n1];
  const vertex &v2 = inodes_[n2];
  route(v1,v2,w);
}

string
topology::name() const
{
  std::ostringstream ostr;
  ostr << "Unknown(" << n() << "," << d() << ")";
  return ostr.str();
}

void
topology::stochastic_max_link_load(vector<int>&v, int64_t nnodes, bool tight_packing) const
{
  if (nnodes == 0) {
      nnodes = nodes().size();
    }
  else if (nnodes > nodes().size()) {
      nnodes = nodes().size();
    }


  int64_t nvert = (tight_packing?nnodes:nodes().size());
  int64_t nedge = g().n_edges();

  vector<int> vert(nvert);
  for (int i=0; i<vert.size(); i++) vert[i] = i;

  vector<int> npath(nedge);

  RNG::UniformInteger::ptr rng
    = RNG::SimpleCombo::construct();


  walk w;
  int64_t npair = nnodes/2;
  for (int i=0; i<v.size(); i++) {
      fill(npath.begin(), npath.end(), 0);
      vector<int> samp = vert;
      RNG::UniformInteger_functor rngf(rng);
      random_shuffle(samp.begin(), samp.end(), rngf);

      for (int j=0; j<npair; j++) {
          walk w;

          route(samp[j], samp[j+npair], w);

          for (int k=0; k<w.length(); k++) {
              npath[g().edge_to_index(w.e(k)).value()]++;
            }

        }

      v[i] = *max_element(npath.begin(), npath.end());
#if 0 // clj -- debug
      // find the edge with the max element and print it
      for (int j=0; j<nedge; j++) {
          edge_index edgej(j);
          if (v[i] == npath[j] && v[i] > 1) {
              std::cout << "offending edge: "
                        << g().index_to_edge(edgej)
                        << " npath = " << v[i]
                        << std::endl;
              // find the paths containing this edge
              for (int m=0; m<npair; m++) {
                  walk w;
                  route(samp[m], samp[m+npair], w);
                  for (int k=0; k<w.length(); k++) {
                      if (edgej == g().edge_to_index(w.e(k))) {
                          std::cout << w << std::endl;
                          break;
                        }
                    }
                }
            }
        }
#endif // clj -- end debug
    }
}

void
topology::init_node_to_switch()
{
  node_to_switch_.clear();
  for (sstmac::hw::graph::edge_iterator i = g().edges_begin();
       i != g().edges_end();
       i++) {
    lwedge e(*i);

    std::set<vertex>::iterator notfound = nodes().end();
    bool to_is_node
      = (nodes().find(g().index_to_vertex(e.to())) != notfound);
    bool from_is_node
      = (nodes().find(g().index_to_vertex(e.from())) != notfound);

    if (from_is_node && to_is_node) {
      throw std::logic_error("topology::init_node_to_switch: "
                             "found a node to node edge");
    }
    else if (from_is_node) {
      node_to_switch_[e.from()] = e.to();
    }
    else if (to_is_node) {
      node_to_switch_[e.to()] = e.from();
    }
  }
}

vertex_index
topology::node_to_switch(vertex_index nd) const
{
  mapvv::const_iterator isw = node_to_switch_.find(nd);
  if (isw == node_to_switch_.end()) {
    if (node_to_switch_.size() == 0) {
      throw std::logic_error("topology::node_to_switch: has not been initialized");
    }
    else {
      throw std::invalid_argument("topology::node_to_switch: node not found");
    }
  }
  return isw->second;
}

int
topology::edge_width(edge_index e) const
{
  return 1;
}

bool
topology::is_switch(vertex_index v) const
{
  if (direct()) return true;
  if (nodes_.find(g().index_to_vertex(v)) == nodes_.end()) {
    return true;
  }
  return false;
}

bool
topology::is_node(vertex_index v) const
{
  if (direct()) return true;
  if (nodes_.find(g().index_to_vertex(v)) == nodes_.end()) {
    return false;
  }
  return true;
}

} } //end of namespace sstmac
