/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_ACTOR_ACTORID_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_ACTOR_ACTORID_H_INCLUDED

#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/ptr_type.h>
#include <iosfwd>
#include <stdint.h>
#include <sstream>

namespace sstmac { namespace sw
{

  /**
   * An identifier that specifies the location of a process in a
   * distributed actor-based application.
   */
  class actorid
  {
    /// Out of simple paranoia, only a proper server is allowed to create a actorid.
   // friend class actorserver;
    friend class simplemailserver;
    friend class actorworld;
    /// The node that this process resides on.
    nodeaddress::ptr node_;

    /// The rank of this process on the node (issued by the actorserver).
    int32_t rank_;

    /// Create a new valid actorid.
    actorid(const nodeaddress::ptr node, int32_t rank);

  public:
    /// Create a blank (invalid) actorid.
    /// Only a actorserver can issue valid actorid objects.
    actorid();

    /// Goodbye.
    virtual
    ~actorid();

    /// Access the node that the process resides on.
     nodeaddress::ptr
    node() const
    {
      return node_;
    }

    /// Access the rank of this task on the node.
    int32_t
    rank() const
    {
      return rank_;
    }

    /// Test whether this is a valid actor identifier for sends 
    /// (not 'invalid' or 'any).
    bool
    is_real() const;

    /// Test whether this actorid matches the given identifier.
    bool
    matches(const actorid &other) const;

    /// An invalid actor identifier.  This is the same as calling actorid().
    static actorid
    invalid();

    /// A wildcard match for any actor identifier.
    static actorid
    any();

    std::string
    to_string() const
    {
      std::stringstream os;
      if (rank() >= 0)
        os << "actorid(" << node()->to_string() << ", " << rank() << ")";
      else if (rank() == -1)
        os << "actorid::invalid()";
      else
        os << "actorid::any()";

      return os.str();
    }

  };

  /// Needed for sorted containers.
  inline bool
  operator<(const actorid &a, const actorid &b)
  {
    if (a.node()->equals(b.node()))
      return a.rank() < b.rank();
    else
      return a.node() < b.node();
  }

  /// Needed for std::find
  inline bool
  operator==(const actorid &a, const actorid &b)
  {
    return ((a.node()->equals(b.node())) && (a.rank() == b.rank()));
  }

  /// Convenience
  inline bool
  operator!=(const actorid &a, const actorid &b)
  {
    return !(a == b);
  }

  /// Utility.
  std::ostream&
  operator<<(std::ostream &os, const actorid &id);

} } // end of namespace sstmac

#endif
