/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/libraries/mpi/mpiserver.h>
#include <sstmac/software/libraries/mpi/mpimessage.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>

namespace sstmac { namespace sw
{

  // Pointer types.
  
  

  //
  // Hello.
  //
  mpiqueue::sendrequest::sendrequest(const mpirequest_t key, std::string libn,
      const mpiqueue::ptr& queue, int64_t handsize, int64_t envelope,
      int64_t mintrans, int count, const mpitype &type, const mpiid &dest,
      const mpitag& tag, const mpicomm &comm, bool rsend, bool ssend,
      mpimessage::category cat, int seqno, const messageid &id,
      const eventhandler::ptr& completion,
          const payload::const_ptr& content) :
    queue_(queue), handsize_(handsize), envelope_(envelope),
        mintrans_(mintrans), count_(count), type_(type), dest_(dest),
        tag_(tag), comm_(comm), rsend_(rsend), ssend_(ssend), cat_(cat),
        seqno_(seqno), id_(id), completion_(completion), content_(content),
        pending_ack_(false), data_sent_(false), libname_(libn), key_(key)
  {


  }

  //
  // Hi there.
  //
  mpiqueue::sendrequest::ptr
  mpiqueue::sendrequest::construct(const mpirequest_t key, std::string libn,
      const mpiqueue::ptr& queue, int64_t handsize,
      int64_t envelope, int64_t mintrans, int count, const mpitype &type,
      const mpiid &dest, const mpitag& tag, const mpicomm &comm, bool rsend,
      bool ssend, mpimessage::category message_type, int seqno,
      const messageid &id, const eventhandler::ptr& completion,
      const payload::const_ptr& content)
  {

    sendrequest::ptr retval(new sendrequest(key, libn, queue, handsize,
        envelope, mintrans, count, type, dest, tag, comm, rsend, ssend,
        message_type, seqno, id, completion, content));

    return retval;
  }

  //
  // Goodbye.
  //
  mpiqueue::sendrequest::~sendrequest() throw ()
  {
   // std::cout << "calling sendrequest destructor" << std::endl;
  }

  //
  // Go.
  //
  void
  mpiqueue::sendrequest::start(const mpiserver::ptr& lib)
  {

    if (lib == 0)
      {
        throw nullerror(
            "mpiqueue::sendrequest::start:  null mpiserver pointer.");
      }
    mpimessage::ptr mess;
    // Decide whether we are sending a full payload or a handshake.
    int bytes = count_ * type_.size;


    if (rsend_ || (bytes <= handsize_))
      {

        // Fire the whole sucker off.
        const taskid pt = comm_.peer_task(dest_);
        timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something
        mess = mpimessage::construct(libname_, sendtime, envelope_, mintrans_,
            count_, type_, comm_.rank(), dest_, tag_, comm_.id(), seqno_, id_,
            ssend_, cat_, comm_.my_task(), pt, content_);
        mess->fromaddr_ = comm_.node_at(comm_.rank());
        mess->toaddr_ = comm_.node_at(dest_);
        data_sent_ = true;
        // If this is not an ssend operation, send the completion to our listener.
        if (!ssend_)
          {

            lib->send(mess);
          }
        else
          {
            //  small ssend operation - send data but don't signal completion
            // until ack comes back.

            pending_ack_ = true;
            queue_->request_ack(mess, sendrequest::ptr(this));

            lib->send(mess);
          }
      }
    else
      {


        timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something
        mess = mpimessage::construct(libname_, sendtime, envelope_, mintrans_,
            -count_, type_, comm_.rank(), dest_, tag_, comm_.id(), seqno_, id_,
            ssend_, cat_, comm_.my_task(), comm_.peer_task(dest_), content_);
        mess->fromaddr_ = comm_.node_at(comm_.rank());
        mess->toaddr_ = comm_.node_at(dest_);
        // Request to be notified of a (future) incoming ack.
        pending_ack_ = true;
        queue_->request_ack(mess, sendrequest::ptr(this));
        // and fire off a handshake packet (negative size indicates a handshake).

        lib->send(mess);
      }

    if (!lib->nic_send_ack_supported())
      {
        complete(mess);
      }
  }

  //
  // We just received an ack.
  //
  void
  mpiqueue::sendrequest::ack(const mpimessage::ptr& response,
      const mpiserver::ptr& lib)
  {


    if (!pending_ack_)
      {
        throw illformederror("mpiqueue::sendrequest::ack:  This request "
          "is not waiting for an ack.");
      }
    if (response->seqnum() != seqno_)
      {
        throw illformederror("mpiqueue::sendrequest::ack:  Received a "
          "non-matching ack.");
      }
    if (!data_sent_)
      {
        // Package up the rest of our stuff and send it off.
        // The receiver for the payload has already signaled that it's present,
        // so we can route the completion for the send directly to our listener.

        timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something
        mpimessage::ptr mess = mpimessage::construct(libname_, sendtime,
            envelope_, mintrans_, count_, type_, comm_.rank(), dest_, tag_,
            comm_.id(), seqno_, id_, false, mpimessage::data, comm_.my_task(),
            comm_.peer_task(dest_), content_);
        mess->fromaddr_ = comm_.node_at(comm_.rank());
        mess->toaddr_ = comm_.node_at(dest_);
        lib->send(mess);
      }
    else
      {
        // We got our ack for an ssend operation.  Our work here is done.

        completion_->handle(response);
      }
  }

  //
  // Test whether we match a given message, for nic send acks.
  //
  bool
  mpiqueue::sendrequest::matches(const mpimessage::ptr& message) const
  {
    bool justHandshake = (message->count() < 0);
    bool sameid = (seqno_ == message->seqnum());
    return (sameid && !justHandshake && comm_.rank().matches(message->source())
        && tag_.matches(message->tag()));
  }

  void
  mpiqueue::sendrequest::complete(const sst_message::ptr &msg)
  {
    mpimessage::ptr mpi =
        boost::dynamic_pointer_cast<mpimessage>(msg);

    key_->complete(mpi);

    if (completion_)
      completion_->handle(msg);

  }

} } // end of namespace sstmac.
