/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpibarrierengine.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/topologyiterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw
{

  // Pointer types.
  

  // A utility type for sendrecv
  // I figured out that I can actually end up recursively inside sendrecv,
  // which will then result in an invalid iterator and major screw-ups.
  struct sendrecvop
  {
    bool sending;
    mpiid partner;
  };

  //
  // Fire off the next set of send and receive operations.
  //
  void
  mpibarrierengine::sendrecv()
  {
    SSTMAC_DEBUG << "mpibarrierengine[" << comm_.rank().id << "]:: sendrecv \n";
    if (pending_recvs_)
      {
        throw illformederror("mpibarrierengine::sendrecvs: Pending receive.");
      }
    std::vector<sendrecvop> ops;
    while (iter_->active())
      {
        if (iter_->sending())
          {
            sendrecvop op =
              { true, mpiid(iter_->dest()) };
            ops.push_back(op);
          }
        if (iter_->receiving())
          {
            sendrecvop op =
              { false, mpiid(iter_->source()) };
            ops.push_back(op);
            // We can only safely handle one receive at a time.
            // (a barrier will generally receive zero or one time,
            // followed by zero or more sends).
            iter_->next();
            break;
          }
        iter_->next();
      }
    SSTMAC_DEBUG << "mpibarrierengine[" << comm_.rank().id << "]:: firing off " << ops.size() << " ops \n";
    // Fire off all sends and receives.
    for (size_t it = 0; it < ops.size(); ++it)
      {
        if (!ops[it].sending)
          {
            ++pending_recvs_;
            this->start_recv(1, mpitype::mpi_byte, tag_, ops[it].partner);
          }
        else
          {
            ++pending_sends_;
            this->start_send(1, mpitype::mpi_byte, tag_, ops[it].partner,
                payload::null());
          }
      }

    // If we have no pending sends or receives, we are done.
    if (pending_sends_ == 0 && pending_recvs_ == 0)
      {
        mpicollective::complete(payload::const_ptr ());
      }
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void
  mpibarrierengine::send_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &dest)
  {
    SSTMAC_DEBUG << "mpibarrierengine[" << comm_.rank().id << "]::send_complete("
          << count << ", " << type.to_string() << ", " << tag.to_string() << ", "
          << dest.to_string() << ")\n";
    --pending_sends_;
    if (!pending_recvs_)
      {
        this->sendrecv();
      }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void
  mpibarrierengine::recv_complete(int count, const mpistatus_t& stat)
  {
    SSTMAC_DEBUG << "mpibarrierengine[" << comm_.rank().id << "]::recv_complete("
          << count << ", " << ")\n";
    --pending_recvs_;
    if (!pending_recvs_)
      {
        this->sendrecv();
      }
  }

  //
  // Copy complete.
  //
  void
  mpibarrierengine::copy_complete(int, const mpitype&, const mpitag&,
      const mpiid&)
  {
    throw illformederror("mpibarrierengine::copy_complete:  "
      "MPI barrier doesn't copy data.");
  }

  //
  // Hi.
  //
  mpibarrierengine::mpibarrierengine(const topologyiterator::ptr& iter,
      const mpirequest::ptr& the_key, const mpiqueue::ptr& queue,
      const mpitag &tag, const mpicomm &comm,
      eventhandler::ptr completion) :
    mpicollective(the_key, queue, comm, completion), iter_(iter), tag_(tag),
        pending_sends_(0), pending_recvs_(0)
  {
    if (tag_ == mpitag::any_tag)
      throw valueerror("mpibarrierengine(...)"
        ":  refusing to send using mpitag::any_tag.");
  }

  //
  // Hi.
  //
  mpibarrierengine::ptr
  mpibarrierengine::construct(const topologyiterator::ptr& iter,
      const mpirequest::ptr& the_key, const mpiqueue::ptr& queue,
      const mpitag &tag, const mpicomm &comm,
      eventhandler::ptr completion)
  {
    mpibarrierengine::ptr retval(new mpibarrierengine(iter, the_key,
        queue, tag, comm, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpibarrierengine::~mpibarrierengine() throw ()
  {
  }

  //
  // Start this kernel.
  //
  void
  mpibarrierengine::start()
  {
    SSTMAC_DEBUG << "mpibarrierengine[" << comm_.rank().id << "]::start \n";
    mpicollective::start();
    this->sendrecv();
  }

} } // end of namespace sstmac

