/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIDIRECTALLTOALL_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIDIRECTALLTOALL_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <vector>
#include <deque>

namespace sstmac { namespace sw {

  // Forward declaration of important types.
  //class mpimagic;

  /**
   * Perform an alltoall (or alltoallv) using concurrent isend and irecv calls.
   * This may not provide the best scaling for all situations, but is probably
   * never outright horrible.
   */
  class mpidirectalltoall : public mpicollective {
    /// Private nested type to keep track of sends that haven't been posted yet
    struct pendsend {
      explicit pendsend(int sendcnt, const mpiid &pal) :
        cnt(sendcnt), dest(pal)
      {}
      int cnt;
      mpiid dest;
    };
    /// Private nested type for receives waiting to be posted.
    struct pendrecv {
      explicit pendrecv(int recvcnt, const mpiid &pal) :
        cnt(recvcnt), source(pal)
      {}
      int cnt;
      mpiid source;
    };

    /// The list of send lengths.
    std::vector<int> sendcnt_;
    /// The send type.
    mpitype sendtype_;
    /// The list of receive lengths.
    std::vector<int> recvcnt_;
    /// The receive type.
    mpitype recvtype_;
    /// The tag for all our communications.
    mpitag tag_;
    /// Our communicator.
    mpicomm comm_;
    /// The magic shortcut to gather data.
   // boost::shared_ptr<mpimagic> magic_;
    /// The maximum number of concurrent sends and receives.
    int max_concurrent_ops_;
    /// Sends that have not been posted yet.
    std::deque<pendsend> pendsend_;
    /// Receives that have not been posted yet.
    std::deque<pendrecv> pendrecv_;
    /// The number of outstanding sends (sends already posted).
    int pending_sends_;
    /// The number of outstanding recvs.
    int pending_recvs_;

    std::vector<payload::const_ptr > send_content_;
    std::vector<payload::const_ptr > recv_content_;

  protected:
    /// Hello.
    mpidirectalltoall(const mpirequest::ptr& thekey,
                      const mpiqueue::ptr& queue,
                      const std::vector<int> &sendcnt, const mpitype &sendtype,
                      const std::vector<int> &recvcnt, const mpitype &recvtype,
                      const mpitag &tag, const mpicomm &comm,
                      const std::vector<payload::const_ptr >& content,
                      eventhandler::ptr completion);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(
                               int count,
                               const mpistatus_t& stat);

    /// Callback to signal copy completion (only on clients).
    void copy_complete(
                       int count,
                       const mpitype &type,
                       const mpitag &tag,
                       const mpiid &source);

  public:
    typedef boost::intrusive_ptr<mpidirectalltoall> ptr;

    /// Hello.
    static mpidirectalltoall::ptr
    construct(const mpirequest::ptr& thekey,
              const mpiqueue::ptr& queue,
              const std::vector<int> &sendcnt, const mpitype &sendtype,
              const std::vector<int> &recvcnt, const mpitype &recvtype,
              const mpitag &tag, const mpicomm &comm,
              const std::vector<payload::const_ptr >& content,
              eventhandler::ptr completion);

    /// Goodbye.
    virtual ~mpidirectalltoall() throw();

    /// Get the maximum number of concurrent sends or receives.
    /// A value of zero or less indicates no maximum number.
    inline int max_concurrent_ops() { return max_concurrent_ops_; }

    /// Define the maximum number of concurrent sends or receives.
    /// Set to zero or less to have no maximum.
    void set_max_concurrent_ops(int max);

    /// Start this kernel.
    virtual void start();
  };

} } // end of namespace sstmac

#endif
