/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpidirectgatherv.h>


#include <sstmac/common/eventhandler.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <deque>

namespace sstmac { namespace sw {

  // Pointer types.


  using boost::dynamic_pointer_cast;

  //
  // Hi.
  //
  mpidirectgatherv::mpidirectgatherv(const mpirequest_t& thekey,
      const mpiqueue::ptr& queue,
      int sendcount, const mpitype &sendtype,
      const std::vector<int> &recvcount,
      const mpitype &recvtype,
      const mpiid &root, const mpitag &tag,
      const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion) :
      mpicollective(thekey, queue, comm, completion),
      sendcount_(sendcount), sendtype_(sendtype),
      recvcount_(recvcount), recvtype_(recvtype),
      root_(root), tag_(tag), comm_(comm),
      pending_sends_(0), pending_recvs_(0)
  {
    SSTMAC_DEBUG << "mpidirectgatherv[" << comm_.rank().id << "]::construct\n";

    bool is_root = comm_.rank() == root_;

    if(is_root){

        std::vector<payload::const_ptr> vals;

        for(int i = 0; i < comm.size().id; i++){
            vals.push_back(payload::const_ptr());
        }

        vals[comm.rank().id] = content;

        content_ = mpicollpayload::construct(vals);

    }else{
        sendcontent_ = content;
    }

  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpidirectgatherv::send_complete( int count,
      const mpitype &type, const mpitag &tag,
      const mpiid &dest)
  {
    SSTMAC_DEBUG << "mpidirectgatherv[" << comm_.rank().id << "]::send_complete(" << ", " << count
        << ", " << type.to_string() << ", " << tag.to_string() << ", " << dest.to_string() << ")\n";

    --pending_sends_;
    if(! pending_sends_) {
        payload::const_ptr content;
        if((comm_.rank() == root_)) {

            content = content_;
        }
        mpicollective::complete(content);
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpidirectgatherv::recv_complete(int count,
      const mpistatus_t& stat)
  {
    SSTMAC_DEBUG<< "mpidirectgatherv[" << comm_.rank().id << "]::recv_complete(" << ", " << count
        << ", " << stat << ")\n";
    //this->start_funneled_copy(count, stat->recv_type(),
    //                         stat->tag(), stat->source());

    payload::const_ptr load = stat->content();

    mpiid src = stat->source();

    content_->set_content(load, src.id);

    --pending_recvs_;
    if(! pending_recvs_) {
        if((comm_.rank() == root_)) {
            mpicollective::complete(content_);
        }
    }
  }

  //
  // Callback to signal copy completion (only on clients).
  //
  void mpidirectgatherv::copy_complete( int count,
      const mpitype &type, const mpitag &tag,
      const mpiid &source)
  {
    SSTMAC_DEBUG << "mpidirectgatherv[" << comm_.rank().id << "]::copy_done(" << ", " << count << ", "
        << type.to_string() << ", " << tag.to_string() << ", " << source.to_string() << ")\n";
    --pending_recvs_;
    if(! pending_recvs_) {
        payload::const_ptr content;
        if((comm_.rank() == root_)) {

        }
        mpicollective::complete(content);
    }
  }

  //
  // Hello.
  //
  mpidirectgatherv::ptr
  mpidirectgatherv::construct(const mpirequest_t& thekey,
      const mpiqueue::ptr& queue,
      int sendcount, const mpitype &sendtype,
      const std::vector<int> &recvcount,
      const mpitype &recvtype,
      const mpiid &root, const mpitag &tag,
      const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion)
  {
    mpidirectgatherv::ptr retval
    (new mpidirectgatherv(thekey, queue, sendcount, sendtype, recvcount,
        recvtype, root, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpidirectgatherv::~mpidirectgatherv() throw() {
  }

  //
  // Start this kernel.
  //
  void mpidirectgatherv::start()
  {
    SSTMAC_DEBUG << "mpidirectgatherv[" << comm_.rank().id << "]::start()\n";
    mpicollective::start();
    if(comm_.rank() == root_) {
        // Root node -- fire off a bunch of receives.
        mpiid size = comm_.size();
        pending_recvs_ = comm_.size().id - 1;
        if(pending_recvs_ == 0) {
            payload::const_ptr content;
            if((comm_.rank() == root_)) {
                content = content_;
            }
            mpicollective::complete(content);
        }
        else {
            for(mpiid dest(0); dest < size; ++dest) {
                if(dest != root_) {
                    this->start_recv(recvcount_.at(dest.id), recvtype_, tag_, dest);
                }
            }
        }
    }
    else {
        // Client node -- fire off a single send.
        pending_sends_ = 1;
        this->start_send(sendcount_, sendtype_, tag_, root_, sendcontent_);
    }
  }

} } // end of namespace sstmac
