/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpiringallgather.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw
{

  // Pointer types.
  
  using boost::dynamic_pointer_cast;

  //
  // Hello.
  //
  mpiringallgather::mpiringallgather(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, int sendcnt, const mpitype &sendtype,
      int recvcnt, const mpitype &recvtype, const mpitag &tag,
      const mpicomm &comm, const payload::const_ptr& content,
      eventhandler::ptr completion) :
    mpicollective(thekey, queue, comm, completion), sendcnt_(sendcnt),
        sendtype_(sendtype), recvcnt_(recvcnt), recvtype_(recvtype), tag_(tag),
        comm_(comm), iteration_(0), target_iterations_(comm.size().id - 1),
        source_((comm.size().id + comm.rank().id - 1) % comm.size().id), dest_(
            (comm.rank().id + 1) % comm.size().id), pending_sends_(0),
        pending_recvs_(0)
  {
    // Clone the content if it is non-null.
    //std::cerr << "mpiringallgather(...):  " << comm << " of size " << comm.size() << ": adding collective payload for tag " << tag << "\n";
    // magic_->add_collective_payload(typeid(*this), tag, comm,
    //                               (content ? content->clone() : content),
    //                               true);
    std::vector<payload::const_ptr > vals;
    for (int i = 0; i < comm.size().id; i++)
      {
        if (i == comm.rank().id)
          vals.push_back(content);
        else
          vals.push_back(payload::const_ptr ());
      }

    content_ = mpicollpayload::construct(vals);
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void
  mpiringallgather::send_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &dest)
  {
    if (dbg_.is_active())
      dbg_ << "mpiringallgather::send_complete(" << ", " << count << ", "
          << type.id << ", " << tag.id << ", " << dest.id << ")\n";
    --pending_sends_;
    if (pending_sends_ == 0 && pending_recvs_ == 0)
      {
        //std::cerr << "mpiringallgather::send_complete:  No pending messages on " << comm_ << " with tag " << tag_ << "\n";
        //  if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        //   throw illformederror("mpiringallgather::send_complete:  Transfers done "
        //                        "but not all payload has been gathered.");
        // payload::const_ptr content =
        //   magic_->get_collective_payload(typeid(*this), tag_, comm_);
        mpicollective::complete(content_);
      }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void
  mpiringallgather::recv_complete(int count,
      const mpistatus_t& stat)
  {
    SSTMAC_DEBUG << "mpiringallgather::recv_complete(" << ", " << count << ", "
          << stat << ")\n";
    --pending_recvs_;
    ++iteration_;

    payload::const_ptr pay = stat->content();
    if (pay)
      {
        mpicollpayload::ptr other = mpicollpayload::construct(
            dynamic_pointer_cast<const mpicollpayload> (stat->content()->clone()));

        std::vector<payload::const_ptr > realload =
            other->get_content();

        std::vector<payload::const_ptr >::iterator it, end =
            realload.end();

        int i = 0;
        for (it = realload.begin(); it != end; it++)
          {
            if (*it)
              {
                SSTMAC_DEBUG << "mpiringallgather(" << comm_.rank().id
                      << "]: setting content for position " << i << "\n";

                content_->set_content(*it, i);
              }
            i++;
          }
      }

    if (iteration_ < target_iterations_)
      {
        // Next send/receive pair.
        ++pending_sends_;
        ++pending_recvs_;
        this->start_recv(recvcnt_, recvtype_, tag_, source_);
        this->start_send(sendcnt_, sendtype_, tag_, dest_, content_);
      }
    if (pending_sends_ == 0 && pending_recvs_ == 0)
      {

        mpicollective::complete(content_);
      }
  }

  //
  // Callback to signal copy completion (ignored).
  //
  void
  mpiringallgather::copy_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &source)
  {
    SSTMAC_DEBUG << "mpiringallgather::copy_complete(" << ", " << count << ", "
          << type.id << ", " << tag.id << ", " << source.id << ")\n";
  }

  //
  // Hello.
  //
  mpiringallgather::ptr
  mpiringallgather::construct(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, int sendcnt, const mpitype &sendtype,
      int recvcnt, const mpitype &recvtype, const mpitag &tag,
      const mpicomm &comm, const payload::const_ptr& content,
      eventhandler::ptr completion)
  {
    mpiringallgather::ptr retval(new mpiringallgather(thekey, queue,
        sendcnt, sendtype, recvcnt, recvtype, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpiringallgather::~mpiringallgather() throw ()
  {
  }

  //
  // Get busy child.
  //
  void
  mpiringallgather::start()
  {
    if (dbg_.is_active())
      dbg_ << "mpiringallgather::start(" << ")\n";
    mpicollective::start();
    // Catch situations where comm_.size() is 1.
    if (target_iterations_ == 0)
      {

        mpicollective::complete(content_);
      }
    else
      {
        // Set up the first send/receive pair.
        iteration_ = 0;
        pending_sends_ = pending_recvs_ = 1;
        this->start_recv(recvcnt_, recvtype_, tag_, source_);
        this->start_send(sendcnt_, sendtype_, tag_, dest_, content_);
      }
  }

} } // end of namespace sstmac
