/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPISTRATEGIES_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPISTRATEGIES_H_INCLUDED

#include <sstmac/common/sstmac-config.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitag.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpiid.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitype.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpiop.h>
#include <sstmac/software/libraries/mpi/mpirequest.h>
#include <sstmac/software/libraries/mpi/mpimessage.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/software/process/os_api.h>
#include <sstmac/common/ptr_type.h>
#include <vector>
#include <stdint.h>

namespace sstmac { namespace sw {

  // Forward declaration of important types:
class mpistrategybase : public ptr_type {
public:
  virtual std::string to_string() const{
          return "mpistrategy";
        }
};

  //
  /// Base strategy for sending messages.
  //
  class mpisendstrategy : public mpistrategybase  {
  public:
    typedef boost::intrusive_ptr<mpisendstrategy> ptr;

    /// Virtual destructor is critical here.
    virtual ~mpisendstrategy() throw() {}
    /// Post a send request to the given node and associate the
    /// completion of this request with the given request handle.
    /// All sends are non-blocking; if return receipts are required the
    /// completion of the request must be associated with the arrival
    /// of the return receipt.
    ///
    /// Like all strategy methods, this method must be reentrant.
    virtual void
    execute(
        const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int count, const mpitype &type, const mpiid &dest,
          const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  /// Base strategy for receiving messages.
  //
  class mpirecvstrategy  : public mpistrategybase {
  public:
    typedef boost::intrusive_ptr<mpirecvstrategy> ptr;

    /// Must be virtual
    virtual ~mpirecvstrategy() throw() {}
    /// Post a receive request from the given node to be matched by the
    /// given request handle.  All receives are non-blocking; it is up
    /// to the strategy to handle return receipts or other synchronization.
    virtual void
    execute(
        const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int count, const mpitype &type, const mpiid &source,
          const mpitag &tag, const mpicomm &comm,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  /// Base strategy for broadcast operations.
  //
  class mpibcaststrategy  : public mpistrategybase {
  public:
      typedef boost::intrusive_ptr<mpibcaststrategy> ptr;
    /// Must be virtual
    virtual ~mpibcaststrategy() throw() {}
    /// Broadcast from root to all the mpiids in the given list.
    virtual void
    execute(
        const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int count, const mpitype &type, const mpiid &root,
          const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
      };

  //
  /// Base strategy for reduce operations.
  //
  class mpireducestrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpireducestrategy> ptr;
    /// Must be virtual.
    virtual ~mpireducestrategy() throw() {}
    /// Reduce between root mpiid and all other participants.
    /// Associates the completion of the reduction operation with the
    /// given request handle.
    virtual void
    execute(
        const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int count, const mpitype &type, const mpiop &op,
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  /// Base strategy for allreduce operations.
  //
  class mpiallreducestrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpiallreducestrategy> ptr;
    /// Must be virtual.
    virtual ~mpiallreducestrategy() throw() {}
    /// Reduce between all nodes.
    /// Associates the completion of the reduction operation with the
    /// given request handle.
    virtual void
    execute(const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int count, const mpitype &type, const mpiop &op,
          const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for gather operations.
  //
  class mpigatherstrategy  : public mpistrategybase {
  public:
      typedef boost::intrusive_ptr<mpigatherstrategy> ptr;
    /// Must be virtual.
    virtual ~mpigatherstrategy() throw() {}
    /// Perform a gather operation
    virtual void
    execute(
        const mpirequest_t& thekey,
        const mpiqueue::ptr& queue,
          int sendcount, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype,
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  /// Base strategy for gatherv operations.
  //
  class mpigathervstrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpigathervstrategy> ptr;
    /// Must be virtual.
    virtual ~mpigathervstrategy() throw() {}
    /// Gather data to root node from all other participants.
    virtual void
    execute(const mpirequest_t& thekey,
        const mpiqueue::ptr& queue,
          int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcounts, const mpitype &recvtype, 
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content, OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for allgather operations.
  //
  class mpiallgatherstrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpiallgatherstrategy> ptr;
    /// Must be virtual.
    virtual ~mpiallgatherstrategy() throw() {}
    /// Gather data to all nodes from all participants.
    virtual void
    execute(const mpirequest_t& thekey,
        const mpiqueue::ptr& queue,
          int sendcount, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for allgatherv operations.
  //
  class mpiallgathervstrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpiallgathervstrategy> ptr;
    /// Must be virtual.
    virtual ~mpiallgathervstrategy() throw() {}
    /// Gather data to all nodes from all participants.
    virtual void
    execute(const mpirequest_t& thekey,
        const mpiqueue::ptr& queue,
          int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcount, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for all-to-all communication with fixed data lengths.
  //
  class mpialltoallstrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpialltoallstrategy> ptr;
    /// Must be virtual.
    virtual ~mpialltoallstrategy() throw() {}
    /// Gather data from all nodes to all nodes.
    virtual void
    execute(
        const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int sendcnt, const mpitype &sendtype,
          int recvcnt, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const std::vector<payload::const_ptr >& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for all-to-all communication with variable data lengths.
  //
  class mpialltoallvstrategy  : public mpistrategybase {
  public:
      typedef boost::intrusive_ptr<mpialltoallvstrategy> ptr;
    /// Must be virtual.
    virtual ~mpialltoallvstrategy() throw() {}
    /// Gather data from all nodes to all nodes.
    virtual void
    execute(
        const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          const std::vector<int> &sendcnts, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const std::vector<payload::const_ptr >& content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for scatter operations.
  //
  class mpiscatterstrategy  : public mpistrategybase {
  public:
      typedef boost::intrusive_ptr<mpiscatterstrategy> ptr;
    /// Must be virtual.
    virtual ~mpiscatterstrategy() throw() {}
    /// Perform a scatter operation
    virtual void
    execute(const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int sendcount, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype,
          const mpiid &root, const mpitag &tag,
          const mpicomm &comm,
          const std::vector<payload::const_ptr > &content,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for scatterv operations.
  //
  class mpiscattervstrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpiscattervstrategy> ptr;
    /// Must be virtual.
    virtual ~mpiscattervstrategy() throw() {}
    /// Scatter data from root node to all other participants.
    virtual void
    execute(const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          const std::vector<int> &sendcounts, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype, 
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const std::vector< payload::const_ptr > &cc,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for fused reduce-scatter operations.
  //
  class mpireduce_scatterstrategy : public mpistrategybase  {
  public:
      typedef boost::intrusive_ptr<mpireduce_scatterstrategy> ptr;
    /// Must be virtual.
    virtual ~mpireduce_scatterstrategy() throw() {}
    /// Perform a fused reduce-scatter operation
    virtual void
    execute(const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          const std::vector<int> &recvcnts, const mpitype &type,
          const mpiop &op, const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content, OS_API::ptr &os
          ) const = 0;
  };

  class mpibarrierstrategy  : public mpistrategybase {
  public:
      typedef boost::intrusive_ptr<mpibarrierstrategy> ptr;
    /// Must be virtual.
    virtual ~mpibarrierstrategy() throw() {}
    /// Synchronize nodes.
    virtual void
    execute(const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          const mpitag &tag, const mpicomm &comm,
          OS_API::ptr &os
          ) const = 0;
  };

  //
  // Base strategy for scan operations.
  //
  class mpiscanstrategy  : public mpistrategybase {
  public:
      typedef boost::intrusive_ptr<mpiscanstrategy> ptr;
    /// Must be virtual.
    virtual ~mpiscanstrategy() throw() {}
    /// execute a kernel to perform a scan operation
    virtual void
    execute(const mpirequest_t& thekey,
          const mpiqueue::ptr& queue,
          int count, const mpitype &type, const mpiop &op,
          const mpitag &tag, const mpicomm &comm,
          const payload::const_ptr& content,
          OS_API::ptr &os
          ) const = 0;
  };

} } // end of namespace sstmac.

#endif
