/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpiapi.h>
#include <sstmac/software/libraries/mpi/mpiapi-persistent.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/software/libraries/mpi/mpitracer.h>
#include <sstmac/software/libraries/mpi/mpirequest.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>

#include <sstmac/software/process/os_api.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpistrategy.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicommfactory.h>

#include <sstmac/software/libraries/mpi/mpiserver.h>
#include <sstmac/common/errors.h>

#include <sstmac/software/libraries/unblockhandler.h>

#include <sstmac/software/libraries/mpi/dumpitracer.h>

#include <time.h>

#include <boost/lexical_cast.hpp>

namespace sstmac { namespace sw
  {

  // Pointer types.

  using namespace std;
  using boost::dynamic_pointer_cast;
  using boost::lexical_cast;


  debug mpiapi::dbg_;

  mpitracer::ptr mpiapi::tracer_;
  std::string mpiapi::tracername_ = "";


  // Catch ugly null pointers.
  template<typename T>
  void
  nonnull(const boost::intrusive_ptr<T> &p, const char *caller)
  {
    if (p == 0)
      throw nullerror(caller);
  }



  //
  // Catch calls on an un-initialized MPI environment.
  //
  inline void
  mpiapi::assert_initialized(const char *caller) const
  {
    if (status_ != is_initialized)
      {
        throw illformederror(
            std::string(caller) + ":  mpiapi is not active (initialized).");
      }
  }

  // Test whether a communicator is valid for sending or receiving.
  inline void
  validcomm(const mpicomm &comm, const char *here)
  {
    if (comm == mpicomm::comm_null)
      throw valueerror(
          std::string(here) + ": Can't communicate on MPI_COMM_NULL.");
  }

  // Test whether we are sending a valid number of elements.
  inline void
  validcount(int count, const char *here)
  {
    if (count < 0)
      throw valueerror(std::string(here) + ": Negative array length.");
  }

  // Test whether the send type is acceptable.
  inline void
  validtype(const mpitype &type, const char *here)
  {
    if (type.size <= 0)
      throw valueerror(
          std::string(here) + ":  MPI type \"" + type.label
          + "\" does not have size greater than zero.");
  }

  // Test whether the given id is valid for sending.
  inline void
  validsendid(const mpiid &target, const mpicomm &comm, const char *here)
  {
    if (target.id < 0 || target >= comm.size())
      throw valueerror(
          std::string(here) + ": Invalid destination rank "
          + lexical_cast<std::string>(target));
  }

  // Test whether the given id is valid for receiving.
  inline void
  validrecvid(const mpiid &source, const mpicomm &comm, const char *here)
  {
    if ((source.id < 0 || source >= comm.size())
        && (source != mpiid::any_source))
      {
        std::cerr << "Failing on source rank" << source << "\n";
        throw valueerror(
            std::string(here) + ": Invalid source rank"
            + lexical_cast<std::string>(source));
      }
  }

  // Test wheter the given tag is valid for sending.
  inline void
  validsendtag(const mpitag &tag, const char *here)
  {
    if (tag.id < 0)
      throw valueerror(std::string(here) + ": Send-tag must be >= 0.");
  }

  // Test wheter the given tag is valid for receiving.
  inline void
  validrecvtag(const mpitag &tag, const char *here)
  {
    if ((tag.id < 0) && (tag != mpitag::any_tag))
      throw valueerror(
          std::string(here) + ": Receive-tag must be >= 0 or mpitag::any_tag.");
  }

  // Test paramters for sending a single element.
  inline void
  validsend(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm, const char *here)
  {
    validcount(count, here);
    validtype(type, here);
    validcomm(comm, here);
    validsendid(target, comm, here);
    validsendtag(tag, here);
  }

  // Test paramters for receiving a single element.
  inline void
  validrecv(int count, const mpitype &type, const mpiid &source,
      const mpitag &tag, const mpicomm &comm, const char *here)
  {
    validcount(count, here);
    validtype(type, here);
    validcomm(comm, here);
    validrecvid(source, comm, here);
    validrecvtag(tag, here);
  }

  // Test paramters for sending a vector of elements.
  inline void
  validsends(const std::vector<int> &counts, const mpitype &type,
      const mpiid &target, const mpitag &tag, const mpicomm &comm,
      const char *here)
  {
    std::vector<int>::const_iterator it, end = counts.end();
    for (it = counts.begin(); it != end; ++it)
      validcount(*it, here);
    validtype(type, here);
    validcomm(comm, here);
    validsendid(target, comm, here);
    validsendtag(tag, here);
  }

  // Test paramters for receiving a vector of elements.
  inline void
  validrecvs(const std::vector<int> &counts, const mpitype &type,
      const mpiid &source, const mpitag &tag, const mpicomm &comm,
      const char *here)
  {
    std::vector<int>::const_iterator it, end = counts.end();
    for (it = counts.begin(); it != end; ++it)
      validcount(*it, here);
    validtype(type, here);
    validcomm(comm, here);
    validrecvid(source, comm, here);
    validrecvtag(tag, here);
  }

  // Test arguments for a broadcast.
  inline void
  validbcast(int count, const mpitype &type, const mpiid &root,
      const mpitag &tag, const mpicomm &comm, const char *here)
  {
    validcount(count, here);
    validtype(type, here);
    validcomm(comm, here);
    validsendid(root, comm, here);
    validsendtag(tag, here);
  }

  // Test whether all the given ranks are valid in a communicator.
  inline void
  validranks(const std::vector<mpiid> &rank, const mpicomm &comm,
      const char *here)
  {
    const int ub = comm.size().id;
    std::vector<mpiid>::const_iterator it, end = rank.end();
    for (it = rank.begin(); it != end; ++it)
      {
        if (it->id < 0 || it->id >= ub)
          {
            throw valueerror(std::string(here) + ":  Invalid rank found.");
          }
      }
  }

  std::string
  mpiapi::get_lib_name()
  {
    std::stringstream ss;
    ss << "mpiapi" << id_.app_.id << "-" << id_.task_.id;
    return ss.str();
  }

  //
  // Build a new mpiapi.
    //
    mpiapi::mpiapi(softwareid id) :
        library(id), rank_(id.task_.id), status_(is_fresh)
    {
      if(!dbg_.is_real()){
           dbg_ = debug("<debug> mpi | mpiapi");
           logger::register_logger(&dbg_);
       }

      if (tracername_ == "")
        {
          stringstream timestr;
          time_t t = time(NULL);
          timestr << "dumpi-trace-";
          timestr << t;
          tracername_ = timestr.str();
        }
    }

  //
  // Build a new object.
  //
  mpiapi::ptr
  mpiapi::construct(softwareid id)
  {
    mpiapi::ptr retval(new mpiapi(id));

    return retval;
  }

  //
  // Adieu.
  //
  mpiapi::~mpiapi()
  {

  }


  //
  // Get the world.
  //
  const mpicomm&
  mpiapi::comm_world() const
  {
    return worldcomm_;
  }

  //
  // Get self.
  //
  const mpicomm&
  mpiapi::comm_self() const
  {
    return selfcomm_;
  }

  /*
   * Methods exposing MPI calls and other key methods to
   * derived objects.
   */

  //
  // Get count.
  //
  int
  mpiapi::get_count(const_mpistatus_t status, const mpitype &type)
  {
    assert_initialized("mpiapi::get_count");
    nonnull(status, "mpiapi::get_count:  Null status pointer.");
    int count = status->bytes_received() / type.size;
    if (tracer_->is_active())
      tracer_->get_count(os_->now(), os_->now(), status, type, count);
    return count;
  }

  /* Set up and tear down */

  //
  // Test whether MPI has bee initialized.
  //
  bool
  mpiapi::initialized()
  {
    if (tracer_->is_active())
      tracer_->initialized(os_->now(), os_->now(), status_ == is_initialized);
    return (status_ == is_initialized);
  }

  //
  // Test whether MPI has been finalized.
  //
  bool
  mpiapi::finalized()
  {
    if (tracer_->is_active())
      tracer_->finalized(os_->now(), os_->now(), status_ == is_finalized);
    return (status_ == is_finalized);
  }

  //
  // Initialize MPI.
  //
  timestamp
  mpiapi::init()
  {

    if (!os_)
      {
        throw nullerror("mpiapi::init - os has not been initialized yet");
      }

    environment::const_ptr env = os_->get_env(id_.app_);
    if (!env)
      {
        throw nullerror("mpiapi::init - no environment found");
      }

    library::ptr mpilib = os_->get_lib(mpiserver::lib_name(id_));

    if (mpilib == library::ptr()) // only do one mpiserver per app per node
      {
        mpilib = mpiserver::construct(id_);
        os_->register_lib(mpilib);

      }
    mpiserver::ptr mpiserv = boost::dynamic_pointer_cast<mpiserver>(
        mpilib);

    queue_ = mpiqueue::construct(mpiserv, id_.task_);

    mpiserv->add_task(id_.task_, queue_);

    comm_factory_ = mpicommfactory::construct(id_.app_, mpiapi::ptr(this));
    comm_factory_->init(env, rank_);

    tracer_ = tracer_t();

    worldcomm_ = comm_factory_->get_world();
    selfcomm_ = comm_factory_->get_self();

    strategy_ = mpistrategy::construct();

    status_ = is_initialized;

    if(!tracer_){

        tracer_ = dumpitracer::construct(tracername_);

        if(tracer_->is_active()){
            tracer_->init(os_->now(), timestamp(100000), worldcomm_, selfcomm_);
        }
    }


    barrier(worldcomm_, false);

    return os_->now();

  }

  //
  // Finalize MPI.
  //
  timestamp
  mpiapi::finalize()
  {

    barrier(worldcomm_);

    status_ = is_finalized;

    if (worldcomm_.rank().id == 0)
      {
        debug mpicheck("<debug> mpicheck");
        if(mpicheck.is_active()){
            mpicheck << "----- MPI API ----------" << "\n";
            mpicheck << "At sim time: " << os_->now() << "\n";
            mpicheck << "Rank 0 is passed the barrier in finalize," << "\n";
            mpicheck << "which means the simulation can probably be" << "\n";
            mpicheck << "killed, if it's not already dead." << "\n";
            mpicheck << "------------------------ " << "\n";
        }
      }

    return os_->now();
  }

  //
  // Get current time.
  //
  timestamp
  mpiapi::wtime()
  {
    return os_->now();
  }

  /* Create and destroy communicators. */

  //
  // Duplicate a communicator.
  //
  timestamp
  mpiapi::comm_dup(const mpicomm &input, mpicomm &output)
  {
    assert_initialized("mpiapi::comm_dup");
    timestamp entry = os_->now();
    output = comm_factory_->comm_dup(input);
    if (tracer_->is_active())
      tracer_->comm_dup(entry, os_->now(), input, output);
    return os_->now();

    throw unimplementederror("mpiapi::comm_dup not implemented");
  }

  //
  // Create a communicator containing a subset of an existing comm.
  //
  timestamp
  mpiapi::comm_create(const mpicomm &input, const std::vector<mpiid> &members,
      mpicomm &output)
  {
    static const char* here = "mpiapi::comm_create";
    assert_initialized(here);
    timestamp entry = os_->now();
    validranks(members, input, here);
    output = comm_factory_->comm_create(input, members);
    if (tracer_->is_active())
      tracer_->comm_create(entry, os_->now(), input, members, output);
    return os_->now();

    throw unimplementederror("mpiapi::comm_create not implemented");
  }

  //
  // Split a communicator.  This one is a little weird.
  //
  timestamp
  mpiapi::comm_split(const mpicomm &incomm, int color, int key, mpicomm &outcomm)
  {
    static const char* here = "mpiapi::comm_split";
    //std::cerr << here << "(" << incomm << ", " << color << ", "
    //      << key << ", &outcomm)\n";
    assert_initialized(here);
    outcomm = comm_factory_->comm_split(incomm, color, key);
    if (tracer_->is_active())
      tracer_->comm_split(os_->now(), os_->now(), incomm, color, key,
          outcomm);
    return os_->now();

    throw unimplementederror("mpiapi::comm_split not implemented");
  }

  //
  // Destroy a communicator.
  //
  timestamp
  mpiapi::comm_free(mpicomm &comm)
  {
    assert_initialized("mpiapi::comm_create");
    if (tracer_->is_active())
      tracer_->comm_free(os_->now(), os_->now(), comm);
    return os_->now();

    throw unimplementederror("mpiapi::comm_free not implemented");
  }

  /* Basic point-to-point operations. */

  //
  // Blocking send.
  //
  timestamp
  mpiapi::send(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::send";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    timestamp entry = os_->now();
    key::ptr block_key = key::construct();

    mpirequest_t req = mpirequest::construct(block_key);
    this->do_isend(count, type, target, tag, comm, req, content);

    //only block if the machine model supports nic send acks
    if (!req->is_complete())
      os_->block(block_key);

    if (tracer_->is_active())
      tracer_->send(entry, os_->now(), count, type, target, tag, comm);

    return os_->now();
  }

  //
  // Blocking send using a user-defined buffer (see buffer_attach).
  //
  timestamp
  mpiapi::bsend(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::bsend";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    this->do_ibsend(count, type, target, tag, comm, req, content);
    timestamp finish = os_->block(req->get_key());
    if (tracer_->is_active())
      tracer_->bsend(entry, finish, count, type, target, tag, comm);
    return finish;
  }

  //
  // Blocking syncronous send.
  //
  timestamp
  mpiapi::ssend(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::comm_ssend";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    mpirequest_t req = mpirequest::construct(key::construct());
    timestamp entry = os_->now();
    this->do_issend(count, type, target, tag, comm, req, content);
    timestamp finish = os_->block(req->get_key());
    if (tracer_->is_active())
      tracer_->ssend(entry, finish, count, type, target, tag, comm);
    return finish;
  }

  //
  // Ready send (as in "ready or not, here comes the data").
  timestamp
  mpiapi::rsend(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::rsend";
    assert_initialized(here);
    validsend(count, type, target, tag, comm, here);
    mpirequest_t req = mpirequest::construct(key::construct());
    timestamp entry = os_->now();
    this->do_irsend(count, type, target, tag, comm, req, content);
    timestamp finish = os_->block(req->get_key());
    if (tracer_->is_active())
      tracer_->ssend(entry, finish, count, type, target, tag, comm);
    return finish;
  }

  //
  // Nonblocking send.
  //
  timestamp
  mpiapi::isend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::isend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = os_->now();
    req = mpirequest::construct(key::construct());
    this->do_isend(count, type, dest, tag, comm, req, content);
    timestamp finish = os_->now();
    if (tracer_->is_active())
      tracer_->isend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  /// Nonblocking send.
  /// wait for this request has the same completion as send.
  /// \return time when the request has been registered (likely current time).
  timestamp
  mpiapi::do_isend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, const mpirequest_t &req,
      const payload::const_ptr& content)
  {
    SSTMAC_DEBUG << "mpiapi[" << rank_.id << "]::do_isend(" << count << ","
      << type.label << "," << dest.id << ", " << tag.id << ", "
      << comm.id().id << ", " << "[req]" << ")\n";

    //// shared_ptr<mpikernel> ker = strategy_->send()->build(req, queue_, count,
    ////     type, dest, tag, comm, content, magic_);

    strategy_->send()->execute(req, queue_, count, type, dest, tag, comm, content,
        os_);

    ////  dbg_ << "mpiapi::isend: adding send kernel.\n";
    ////os_->add(os_->now(), ker);
    return os_->now();
  }

  //
  // Nonblocking user-buffered send.
  //
  timestamp
  mpiapi::ibsend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::ibsend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = os_->now();
    req = mpirequest::construct(key::construct());
    this->do_ibsend(count, type, dest, tag, comm, req, content);
    timestamp finish = os_->now();
    if (tracer_->is_active())
      tracer_->ibsend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  //
  // Nonblocking user-buffered send.
  //
  timestamp
  mpiapi::do_ibsend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, const mpirequest_t &req,
      const payload::const_ptr& content)
  {
    std::cerr << "WARNING:  mpiapi::ibsend does not use "
        << "user-defined buffer\n";
    return this->do_isend(count, type, dest, tag, comm, req, content);
  }

  //
  // Nonblocking synchronous send.
  //
  timestamp
  mpiapi::issend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::issend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = os_->now();
    req = mpirequest::construct(key::construct());
    this->do_issend(count, type, dest, tag, comm, req, content);
    timestamp finish = os_->now();
    if (tracer_->is_active())
      tracer_->issend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  //
  // Nonblocking synchronous send.
  //
  timestamp
  mpiapi::do_issend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, const mpirequest_t &req,
      const payload::const_ptr& content)
  {
    strategy_->ssend()->execute(req, queue_, count, type, dest, tag, comm,
        content, os_);

    return os_->now();
  }

  //
  // Nonblocking ready send.
  //
  timestamp
  mpiapi::irsend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr& content)
  {
    static const char *here = "mpiapi::irsend";
    assert_initialized(here);
    validsend(count, type, dest, tag, comm, here);
    timestamp entry = os_->now();
    req = mpirequest::construct(key::construct());
    this->do_irsend(count, type, dest, tag, comm, req, content);
    timestamp finish = os_->now();
    if (tracer_->is_active())
      tracer_->irsend(entry, finish, count, type, dest, tag, comm, req);
    return finish;
  }

  //
  // Nonblocking ready send.
  //
  timestamp
  mpiapi::do_irsend(int count, const mpitype &type, const mpiid &dest,
      const mpitag &tag, const mpicomm &comm, const mpirequest_t &req,
      const payload::const_ptr& content)
  {
    strategy_->rsend()->execute(req, queue_, count, type, dest, tag, comm,
        content, os_);

    return os_->now();
  }

  //
  // Synchronous receive.
  //
  timestamp
  mpiapi::recv(int count, const mpitype &type, const mpiid &source,
      const mpitag &tag, const mpicomm &comm, const_mpistatus_t &stat)
  {
    static const char *here = "mpiapi::recv";
    assert_initialized(here);
    validrecv(count, type, source, tag, comm, here);
    timestamp entry = os_->now();
    key::ptr block_key = key::construct();

    mpirequest_t req = mpirequest::construct(block_key);
    this->do_irecv(count, type, source, tag, comm, req);
    //threadstatus_t treq = dynamic_pointer_cast<ThreadStatus> (req); //this is the obnoxious part about boost
    //timestamp fini = os_->wait(treq, (int) mpirequest::ARRIVED);
    timestamp fini = entry;
    if (!req->is_complete())
      fini = os_->block(block_key);

    if (tracer_->is_active())
      tracer_->recv(entry, fini, count, type, source, tag, comm, stat);
    stat = req->get_status();
    return fini;
  }

  //
  // Nonblocking receive
  //
  timestamp
  mpiapi::irecv(int count, const mpitype &type, const mpiid &source,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req)
  {
    static const char *here = "mpiapi::irecv";
    assert_initialized(here);
    validrecv(count, type, source, tag, comm, here);

    timestamp entry = os_->now();

    req = mpirequest::construct(key::construct());

    this->do_irecv(count, type, source, tag, comm, req);
    timestamp finish = os_->now();
    if (tracer_->is_active())
      tracer_->irecv(entry, finish, count, type, source, tag, comm, req);
    return finish;
  }

  //
  // Asynchronous receive.
  //
  timestamp
  mpiapi::do_irecv(int count, const mpitype &type, const mpiid &source,
      const mpitag &tag, const mpicomm &comm, const mpirequest_t &req)
  {
    static const char *here = "mpiapi::irecv";
    assert_initialized(here);
    validrecv(count, type, source, tag, comm, here);
    SSTMAC_DEBUG << "mpiapi[" << rank_.id << "]::do_irecv(" << count << "," << type.label
      << "," << source.id << ", " << tag.id << ", " << comm.id().id << ", "
      << "[req]" << ")\n";

    strategy_->recv()->execute(req, queue_, count, type, source, tag, comm, os_);

    ////  dbg_ << "mpiapi[" << id_.id << "]::irecv:  adding recv kernel.\n";
    ////os_->add(os_->now(), ker);
    return os_->now();
  }

  //
  // Send/receieve exchange.
  //
  timestamp
  mpiapi::sendrecv(int sendcount, const mpitype &sendtype, const mpiid &target,
      const mpitag &sendtag, int recvcount, const mpitype &recvtype,
      const mpiid &source, const mpitag &recvtag, const mpicomm &comm,
      const_mpistatus_t &stat)
  {
    static const char *here = "mpiapi::sendrecv";
    assert_initialized(here);
    validsend(sendcount, sendtype, target, sendtag, comm, here);
    validrecv(recvcount, recvtype, source, recvtag, comm, here);
    timestamp entry = os_->now();
    mpirequest_t req1 = mpirequest::construct(key::construct());
    mpirequest_t req2 = mpirequest::construct(key::construct());
    payload::const_ptr cont;
    this->do_isend(sendcount, sendtype, target, sendtag, comm, req1, cont);
    this->do_irecv(recvcount, recvtype, source, recvtag, comm, req2);
    std::vector<mpirequest_t> reqs;
    reqs.push_back(req1);
    reqs.push_back(req2);
    std::vector<const_mpistatus_t> stats;
    waitall(reqs, stats);
    stat = dynamic_pointer_cast<const mpistatus> (req2->get_status());
    timestamp fini = os_->now();
    if (tracer_->is_active())
      tracer_->sendrecv(entry, fini, sendcount, sendtype, target, sendtag,
          recvcount, recvtype, source, recvtag, comm, stat);
    return fini;
  }

  //
  // Send and receive using a single buffer.
  //
  timestamp
  mpiapi::sendrecv_replace(int count, const mpitype &type, const mpiid &dest,
      const mpitag &sendtag, const mpiid &source, const mpitag &recvtag,
      const mpicomm &comm, const_mpistatus_t &stat)
  {
    static const char *here = "mpiapi::sendrecv";
    assert_initialized(here);
    validsend(count, type, dest, sendtag, comm, here);
    validrecv(count, type, source, recvtag, comm, here);
    timestamp entry = os_->now();
    mpirequest_t req1 = mpirequest::construct(key::construct());
    mpirequest_t req2 = mpirequest::construct(key::construct());
    payload::const_ptr cont;
    this->do_isend(count, type, dest, sendtag, comm, req1, cont);
    this->do_irecv(count, type, source, recvtag, comm, req2);
    std::vector<mpirequest_t> reqs;
    reqs.push_back(req1);
    reqs.push_back(req2);
    std::vector<const_mpistatus_t> stats;
    waitall(reqs, stats);
    stat = dynamic_pointer_cast<const mpistatus> (req2->get_status());
    timestamp fini = os_->now();
    if (tracer_->is_active())
      tracer_->sendrecv_replace(entry, fini, count, type, dest, sendtag,
          source, recvtag, comm, stat);
    return fini;
  }

  /* Stepwise point-to-point (not implemented, but
   * should be fairly straight-forward to complete). */

  //
  // Initiate a send.
  //
  timestamp
  mpiapi::send_init(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr &content)
  {
    static const char *here = "mpiapi::send_init";
    assert_initialized(here);
    req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_isend, count,
        type, target, tag, comm, content);
    return this->wtime();
  }

  //
  // Initiate a buffered send.
  //
  timestamp
  mpiapi::bsend_init(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr &content)
  {
      static const char *here = "mpiapi::bsend_init";
   assert_initialized(here);
   validsend(count, type, target, tag, comm, here);
   req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_ibsend, count,
   type, target, tag, comm, content);
   return this->wtime();
  }

  //
  // Initiate a synchronous send.
  //
  timestamp
  mpiapi::ssend_init(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr &content)
  {
    static const char *here = "mpiapli::ssend_init";
   assert_initialized(here);
   validsend(count, type, target, tag, comm, here);
   req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_issend, count,
   type, target, tag, comm, content);
   return this->wtime();
  }

  //
  // Initiate a ready send.
  //
  timestamp
  mpiapi::rsend_init(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req,
      const payload::const_ptr &content)
  {
   static const char *here = "mpiapi::rsend_init";
   assert_initialized(here);
   validsend(count, type, target, tag, comm, here);
   req = persistentsend::construct(mpiapi::ptr(this), &mpiapi::do_irsend, count,
   type, target, tag, comm, content);
   return this->wtime();
  }

  //
  // Initiate a receive.
  //
  timestamp
  mpiapi::recv_init(int count, const mpitype &type, const mpiid &target,
      const mpitag &tag, const mpicomm &comm, mpirequest_t &req)
  {
    static const char *here = "mpiapi::recv_init";
   assert_initialized(here);
   validrecv(count, type, target, tag, comm, here);
   req = persistentrecv::construct(mpiapi::ptr(this), count, type, target, tag,
   comm);
   return this->wtime();
  }

  //
  // Start a previously initiated request.
  //
  timestamp
  mpiapi::start(mpirequest_t req)
  {
    persistent::ptr pers = dynamic_pointer_cast<persistent> (req);
    if (!pers)
      throw illformederror("mpiapi::start:  Incorrect request handle type.");
    pers->start();
    return this->wtime();
  }

  //
  // Start a collection of previously initiated requests.
  //
  timestamp
  mpiapi::startall(std::vector<mpirequest_t> &requests)
  {
    typedef std::vector<mpirequest_t>::iterator iterator_t;
    for (iterator_t it = requests.begin(); it != requests.end(); ++it)
      this->start(*it);
    return this->wtime();
  }

  //
  // Cancel a request that had been previously initiated.
  //
  timestamp
  mpiapi::cancel(mpirequest_t req)
  {
    static bool first = true;
    assert_initialized("mpiapi::cancel");
    throw unimplementederror("mpiapi::cancel:  Not implemented yet.");
    if (first)
      std::cerr << "Warning:  mpiapi::cancel is not implemented "
      << "(but we can currently live without it)\n";
    first = false;
    return this->wtime();
  }

  //
  // Test whether a request has been cancelled (takes no time).
  //
  bool
  mpiapi::test_cancelled(const const_mpistatus_t &req)
  {
    assert_initialized("mpiapi::test_cancelled");
    throw unimplementederror("mpiapi::test_cancelled:  Not implemented yet.");
  }

  /* Completion of outstanding requests */

  //
  // Wait for a request to finish.
  //
  timestamp
  mpiapi::wait(mpirequest_t req, const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::wait");
    timestamp entry = os_->now();
    timestamp fini = entry;
    if (req.get() == NULL)
      {
        // Null request is instantly complete.
        return os_->now();
      }
    if (!req->is_complete())
      {
        fini = os_->block(req->get_key());
      }
    stat = req->get_status();
    if (tracer_->is_active())
      tracer_->wait(entry, fini, req, stat);
    return fini;
  }

  //
  // Wait for all the given requests to finish.
  //
  timestamp
  mpiapi::waitall(std::vector<mpirequest_t> &req,
      std::vector<const_mpistatus_t> &statuses)
  {
    assert_initialized("mpiapi::waitall");
    timestamp entry = os_->now(), fini = entry;
    if (!req.empty())
    {
    	typedef std::vector<mpirequest_t>::iterator iterator_t;
    	key::ptr thekey;

    	statuses.clear();
    	for (iterator_t it = req.begin(); it != req.end(); ++it)
    	{
    		if(!(*it)){
    			SSTMAC_DEBUG << "mpiapi::waitall: null request, skipping \n";
    		}else{
    			if (!(*it)->is_complete())
    			{
    				SSTMAC_DEBUG << "mpiapi[" << worldcomm_.rank().id
    						<< "]: req not finished, blocking \n";
    				os_->block((*it)->get_key());
    			}
    			else
    			{
    				SSTMAC_DEBUG << "mpiapi[" << worldcomm_.rank().id
    						<< "]: req already finished, continuing \n";
    			}
    			const_mpistatus_t stat;
    			if (it->get() != NULL)
    			{
    				stat = dynamic_pointer_cast<const mpistatus>((*it)->get_status());
    				if (stat == 0)
    				{
    					throw illformederror("mpiapi::waitall:  Wait request did not "
    							"return a valid mpistatus object.");
    				}
    			}
    			statuses.push_back(stat);
    		}
    	}

    }
    else
      {
        statuses.clear();
      }
    if (tracer_->is_active())
      tracer_->waitall(entry, fini, req, statuses);
    return fini;
  }

  //
  // Wait for at least one of the given requests to complete.
  //
  timestamp
  mpiapi::waitany(std::vector<mpirequest_t> &req, int &index,
      const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::waitany");
    timestamp entry = os_->now(), fini = entry;
    if (!req.empty())
      {
        key::ptr thekey = key::construct();

        index = -1;
        for(int i = 0; i < req.size(); i++){
            if(req[i]->is_complete()){
                SSTMAC_DEBUG << "mpiapi[" << worldcomm_.rank().id << "]: request " << i << " already complete\n";
                index = i;
                break;
            }else{
                req[i]->get_key()->add_blocker(os_->get_current_context());
            }
        }

        if(index < 0){

            os_->block(thekey);


            for(int i = 0; i < req.size(); i++){
                if(!req[i]->get_key()->still_blocked()){
                    //this is the one
                    SSTMAC_DEBUG << "mpiapi[" << worldcomm_.rank().id << "]: unblocked by request " << i << "\n";
                    index = i;
                    break;
                }
            }
        }

        if (index >= 0)
          {
            stat = req.at(index)->get_status();
            if (stat == 0)
              {
                throw illformederror("mpiapi::waitany:  Wait request did not "
                    "return a valid mpistatus object.");
              }
          }

      }
    else
      {
        index = -1;
      }
    fini = os_->now();
    if (tracer_->is_active())
      tracer_->waitany(entry, fini, req, index, stat);
    return fini;
  }

  //
  // Wait for some of the given requests to complete.
  //
  timestamp
  mpiapi::waitsome(std::vector<mpirequest_t> &req, std::vector<int> &indices,
      std::vector<const_mpistatus_t> &statuses)
  {
    assert_initialized("mpiapi::waitany");
    timestamp entry = os_->now(), fini = entry;
    if (!req.empty())
      {
        key::ptr thekey = key::construct();

        indices.clear();
        for(int i = 0; i < req.size(); i++){
            if(req[i]->is_complete()){
                SSTMAC_DEBUG << "mpiapi[" << worldcomm_.rank().id << "]: request " << i << " already complete\n";
                indices.push_back(i);

            }

            req[i]->get_key()->add_blocker(os_->get_current_context());

        }

        if(indices.size() == 0){

            os_->block(thekey);

            for(int i = 0; i < req.size(); i++){
                if(!req[i]->get_key()->still_blocked()){
                    //this is the one
                    SSTMAC_DEBUG << "mpiapi[" << worldcomm_.rank().id << "]: unblocked by request " << i << "\n";
                    indices.push_back(i);
                }
            }
        }

        if (indices.size() > 0)
          {
            for(int i = 0; i < indices.size(); i++){
                const_mpistatus_t stat = req.at(indices[i])->get_status();
                if (stat == 0)
                  {
                    throw illformederror("mpiapi::waitany:  Wait request did not "
                        "return a valid mpistatus object.");
                  }

                statuses.push_back(stat);
            }
          }

      }
    fini = os_->now();
    if (tracer_->is_active())
      tracer_->waitsome(entry, fini, req, indices, statuses);
    return fini;
  }

  /* Testing of outstanding requests */

  /// Test whether a request has finished.
  /// \return the completion time for the call (current time).
  timestamp
  mpiapi::test(mpirequest_t req, bool &flag, const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::test");
    timestamp entry = os_->now(), fini = entry;
    if (!req)
      {
        flag = true;
        stat = const_mpistatus_t();
      }
    else
      {
        flag = req->is_complete();
        if (flag)
          {

            stat = dynamic_pointer_cast<const mpistatus> (req->get_status());
          }
      }
    if (tracer_->is_active())
      tracer_->test(entry, fini, req, flag, stat);
    return fini;
  }

  //
  // Test whether all the given requests have finished.
  //
  timestamp
  mpiapi::testall(std::vector<mpirequest_t> &req, bool &flag,
      std::vector<const_mpistatus_t> &stats)
  {
    assert_initialized("mpiapi::testall");
    timestamp entry = os_->now();
    if (!req.empty())
      {
        flag = true;
        for (size_t id = 0; id < req.size(); ++id)
          {
            if (req.at(id).get() != NULL && !req.at(id)->is_complete())
              {
                flag = false;
                break;
              }
          }
        if (flag)
          {
            // All requests have completed.
            stats.resize(req.size());
            for (size_t id = 0; id < req.size(); ++id)
              {
                if (req.at(id).get() != NULL)
                  {

                    stats.at(id) = dynamic_pointer_cast<const mpistatus> (
                        req.at(id)->get_status());
                  }
                else
                  stats.at(id) = const_mpistatus_t();
              }
          }
      }
    else
      {
        stats.clear();
        flag = false;
      }
    timestamp fini = os_->now();
    if (tracer_->is_active())
      tracer_->testall(entry, fini, req, flag, stats);
    return fini;
  }

  //
  // Test whether at least one of the given requests has completed.
  //
  timestamp
  mpiapi::testany(std::vector<mpirequest_t> &req, int &index, bool &flag,
      const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::testany");
    timestamp entry = os_->now();
    if (!req.empty())
      {
        flag = false;
        for (size_t id = 0; id < req.size(); ++id)
          {
            if (req.at(id).get() != NULL && req.at(
                id)->is_complete())
              {

                index = id;
                flag = true;
                stat = dynamic_pointer_cast<const mpistatus> (
                    req.at(id)->get_status());
                break;
              }
          }
      }
    else
      {
        flag = false;
      }
    timestamp fini = os_->now();
    if (tracer_->is_active())
      tracer_->testany(entry, fini, req, index, flag, stat);
    return fini;
  }

  //
  // Test whether some of the given requests have completed.
  //
  timestamp
  mpiapi::testsome(std::vector<mpirequest_t> &req, std::vector<int> &indices,
      std::vector<const_mpistatus_t> &statuses)
  {
    assert_initialized("mpiapi::testsome");
    timestamp entry = os_->now();
    if (!req.empty())
      {
        indices.clear();
        statuses.clear();
        for (size_t id = 0; id < req.size(); ++id)
          {
            if (req.at(id).get() != NULL && req.at(
                id)->is_complete())
              {

                indices.push_back(id);
                statuses.push_back(
                    dynamic_pointer_cast<const mpistatus> (
                        req.at(id)->get_status()));
              }
          }
      }
    else
      {
        indices.clear();
        statuses.clear();
      }
    timestamp fini = os_->now();
    if (tracer_->is_active())
      tracer_->testsome(entry, fini, req, indices, statuses);
    return fini;
  }

  //
  // Blocking probe for a message that matches the given signature.
  //
  timestamp
  mpiapi::probe(const mpiid &source, const mpitag &tag, const mpicomm &comm,
      const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::probe");
    timestamp entry = os_->now();
    key::ptr thekey = key::construct();
    mpirequest_t req = mpirequest::construct(thekey);
    eventhandler::ptr unblock = unblockhandler::construct(os_, thekey);
    queue_->probe(req, comm, source, tag, unblock);
    if (!req->is_complete()){
      os_->block(thekey);
    }
    stat = req->get_status();
    timestamp fini = os_->now();
    if (tracer_->is_active())
      tracer_->probe(entry, fini, source, tag, comm, stat);
    return fini;
  }

  //
  // Non-blocking probe for messages.
  //
  timestamp
  mpiapi::iprobe(const mpiid &source, const mpitag &tag, const mpicomm &comm,
      bool &flag, const_mpistatus_t &stat)
  {
    assert_initialized("mpiapi::iprobe");
    timestamp entry = os_->now();
    mpistatus_t statout = queue_->iprobe(comm, source, tag);
    if (statout != 0)
      {
        flag = true;
        stat = statout;
      }
    else
      {
        flag = false;
      }
    timestamp fini = os_->now();
    if (tracer_->is_active())
      tracer_->iprobe(entry, fini, source, tag, comm, flag, stat);
    return fini;
  }

  /* Mucking with MPI state. */

  //
  // Attach a user-defined buffer for *bsend operations.
  //
  timestamp
  mpiapi::buffer_attach(int bytes)
  {
    assert_initialized("mpiapi::buffer_attach");
    throw unimplementederror("mpiapi::buffer_attach:  Not implemented.");
  }

  //
  // Detach the user-defined buffer.
  //
  timestamp
  mpiapi::buffer_detach(int &bytes)
  {
    assert_initialized("mpiapi::buffer_detach");
    throw unimplementederror("mpiapi::buffer_detach:  Not implemented.");
  }

  /* Collective operations */

  //
  // Set a barrier.
  //
  timestamp
  mpiapi::barrier(const mpicomm &comm, bool tracethis)
  {
    assert_initialized("mpiapi::barrier");
    mpitag barriertag = comm.next_collective_tag();
    SSTMAC_DEBUG << "mpiapi[" << worldcomm_.rank().id << "]: barrier: next tag = " << barriertag.id << "\n";
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->barrier()->execute(req, queue_, barriertag, comm, os_);

    if (!req->is_complete())
      os_->block(req->get_key());

    timestamp fini = os_->now();
    if (tracethis && tracer_->is_active())
      tracer_->barrier(entry, fini, comm);
    return fini;
  }

  //
  // Broadcast data without payload
  //
  timestamp
  mpiapi::bcast(int count, const mpitype &type, const mpiid &root,
      const mpicomm &comm)
  {
    payload::const_ptr content = payload::null();
    return this->bcast(count, type, root, comm, content);
  }

  //
  // Broadcast data with payload.
  //
  timestamp
  mpiapi::bcast(int count, const mpitype &type, const mpiid &root,
      const mpicomm &comm, payload::const_ptr &content)
  {
    static const char *here = "mpiapi::bcast";
    mpitag bcasttag = comm.next_collective_tag();
    assert_initialized(here);
    validbcast(count, type, root, bcasttag, comm, here);
    timestamp entry = os_->now();

    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->bcast()->execute(req, queue_, count, type, root, bcasttag, comm,
        content, os_);

    timestamp fini = os_->block(req->get_key());

    mpistatus::const_ptr stat = req->get_status();

    if (!stat)
      throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");

    content = stat->content();

    if (tracer_->is_active())
      tracer_->bcast(entry, fini, count, type, root, comm);
    return fini;
  }

  //
  // Scatter data uniformly.
  //
  timestamp
  mpiapi::scatter(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpiid &root, const mpicomm &comm)
  {
    std::vector<payload::const_ptr > vals;

    if (comm.rank() == root)
      {
        for (mpiid id = mpiid(0); id < comm.size(); ++id)
          vals.push_back(payload::null());
      }

    payload::const_ptr result;
    timestamp fini = this->scatter(sendcount, sendtype, recvcount, recvtype, root,
        comm, vals, result, false);
    return fini;
  }

  //
  // Scatter data with payload.
  //
  timestamp
  mpiapi::scatter(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpiid &root, const mpicomm &comm,
      const std::vector<payload::const_ptr > &cc,
      payload::const_ptr &result, bool usingpayload)
  {
    static const char *here = "mpiapi::scatter";
    mpitag scattertag = comm.next_collective_tag();
    assert_initialized(here);

    if (comm.rank() == root)
      {
        if (comm.size().id != cc.size())
          {
            throw valueerror(
                "mpiapi: scatter - content size does not match comm size");
          }
        validsend(sendcount, sendtype, root, scattertag, comm, here);
      }
    else
      {
        validrecv(recvcount, recvtype, root, scattertag, comm, here);
      }

    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->scatter()->execute(req, queue_, sendcount, sendtype, recvcount,
        recvtype, root, scattertag, comm, cc, os_);

    timestamp fini = os_->block(req->get_key());
    mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus>(
        req->get_status());
    if (!stat)
      throw illformederror("mpiapi::scatter: NULL status pointer.");
    payload::const_ptr collpay = boost::dynamic_pointer_cast<
        const payload>(stat->content());

    if (usingpayload && !collpay)
      throw valueerror("mpiapi:scatter - null mpicollpayload");

    result = collpay;

    if (tracer_->is_active())
      tracer_->scatter(entry, fini, sendcount, sendtype, recvcount, recvtype, root,
          comm);
    return fini;
  }

  //
  // Scatter data non-uniformly.
  //
  timestamp
  mpiapi::scatterv(const std::vector<int> &sendcounts, const mpitype &sendtype,
      int recvcount, const mpitype &recvtype, const mpiid &root,
      const mpicomm &comm)
  {
    std::vector<payload::const_ptr > vals;

    if (comm.rank() == root)
      {
        for (mpiid id = mpiid(0); id < comm.size(); ++id)
          vals.push_back(payload::null());
      }

    payload::const_ptr result;
    timestamp fini = this->scatterv(sendcounts, sendtype, recvcount, recvtype, root,
        comm, vals, result, false);
    return fini;
  }

  //
  // Scatter data non-uniformly.
  //
  timestamp
  mpiapi::scatterv(const std::vector<int> &sendcounts, const mpitype &sendtype,
      int recvcount, const mpitype &recvtype, const mpiid &root,
      const mpicomm &comm,  const std::vector<payload::const_ptr > &cc,
      payload::const_ptr &result, bool usingpayload)
  {
    static const char *here = "mpiapi::scatter";
    mpitag scattertag = comm.next_collective_tag();
    assert_initialized(here);
    if (comm.rank() == root){
        if (comm.size().id != cc.size())
          {
            throw valueerror(
                "mpiapi: scatter - content size does not match comm size");
          }
        validsends(sendcounts, sendtype, root, scattertag, comm, here);
    }
    else
      validrecv(recvcount, recvtype, root, scattertag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->scatterv()->execute(req, queue_,
        sendcounts, sendtype, recvcount, recvtype, root, scattertag, comm, cc,
        os_);


    if(!req->is_complete())
      os_->block(req->get_key());

    timestamp fini = os_->now();
    mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus>(
        req->get_status());
    if (!stat)
      throw illformederror("mpiapi::scatter: NULL status pointer.");
    payload::const_ptr collpay = boost::dynamic_pointer_cast<
        const payload>(stat->content());

    if (usingpayload && !collpay)
      throw valueerror("mpiapi:scatter - null mpicollpayload");

    result = collpay;

    if (tracer_->is_active())
      tracer_->scatterv(entry, fini, sendcounts, sendtype, recvcount, recvtype,
          root, comm);
    return fini;
  }

  //
  // Gather uniformly distributed data.
  //
  timestamp
  mpiapi::gather(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpiid &root, const mpicomm &comm)
  {
    std::vector<payload::const_ptr > rv;
    return gather(sendcount, sendtype, recvcount, recvtype, root, comm,
        payload::null(), rv);
  }

  //
  // Gather uniformly distributed data.
  //
  timestamp
  mpiapi::gather(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpiid &root, const mpicomm &comm,
      const payload::const_ptr& content,
      std::vector<payload::const_ptr > &cc)
  {
    static const char *here = "mpiapi::gather";
    mpitag gathertag = comm.next_collective_tag();
    assert_initialized(here);
    if (comm.rank() == root)
      validrecv(recvcount, recvtype, root, gathertag, comm, here);
    else
      validsend(sendcount, sendtype, root, gathertag, comm, here);

    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->gather()->execute(req, queue_, sendcount, sendtype, recvcount,
        recvtype, root, gathertag, comm, content, os_);

    if(!req->is_complete())
          os_->block(req->get_key());

        timestamp fini = os_->now();

    if (comm.rank() == root)
      {
        mpistatus::const_ptr stat = dynamic_pointer_cast<
            const mpistatus>(req->get_status());
        if (!stat)
          throw illformederror("mpiapi::gather: NULL status pointer.");
        mpicollpayload::const_ptr coll = dynamic_pointer_cast<const mpicollpayload>(stat->content());
        if (!coll)
          throw illformederror("mpiapi::gather:  NULL mpicollpayload.");

        cc = coll->get_content();

      }
    if (tracer_->is_active())
      tracer_->gather(entry, fini, sendcount, sendtype, recvcount, recvtype, root,
          comm);
    return fini;
  }

  //
  // Gather non-uniformly distributed data.
  //
  timestamp
  mpiapi::gatherv(int sendcount, const mpitype &sendtype,
      const std::vector<int> &recvcnts, const mpitype &recvtype, const mpiid &root,
      const mpicomm &comm)
  {
    std::vector<payload::const_ptr > cc;
    return gatherv(sendcount, sendtype, recvcnts, recvtype, root, comm,
        payload::null(), cc);
  }

  //
  // Gather non-uniformly distributed data.
  //
  timestamp
  mpiapi::gatherv(int sendcount, const mpitype &sendtype,
      const std::vector<int> &recvcnts, const mpitype &recvtype, const mpiid &root,
      const mpicomm &comm, const payload::const_ptr &cc,
      std::vector<payload::const_ptr > &result)
  {
    static const char *here = "mpiapi::gatherv";
    mpitag scattertag = comm.next_collective_tag();
    assert_initialized(here);
    if (comm.rank() == root)
      validrecvs(recvcnts, recvtype, root, scattertag, comm, here);
    else
      validsend(sendcount, sendtype, root, scattertag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->gatherv()->execute(req, queue_,
        sendcount, sendtype, recvcnts, recvtype, root, scattertag, comm,
        cc, os_);

    if(!req->is_complete())
      os_->block(req->get_key());

    timestamp fini = os_->now();


    if (comm.rank() == root)
      {
        mpistatus::const_ptr stat = dynamic_pointer_cast<
            const mpistatus>(req->get_status());
        if (!stat)
          throw illformederror("mpiapi::gather: NULL status pointer.");
        mpicollpayload::const_ptr coll = dynamic_pointer_cast<const mpicollpayload>(stat->content());
        if (!coll)
          throw illformederror("mpiapi::gather:  NULL mpicollpayload.");

        result = coll->get_content();

      }
    if (tracer_->is_active())
      tracer_->gatherv(entry, fini, sendcount, sendtype, recvcnts, recvtype,
          root, comm);
    return fini;
  }

  //
  // Gather uniformly distributed data and make it available to all nodes.
  //
  timestamp
  mpiapi::allgather(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpicomm &comm)
  {
    std::vector<payload::const_ptr > rv;
    return allgather(sendcount, sendtype, recvcount, recvtype, comm,
        payload::null(), rv);
  }

  //
  // Gather uniformly distributed data and make it available to all nodes.
  //
  timestamp
  mpiapi::allgather(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpicomm &comm,
      const payload::const_ptr& content,
      std::vector<payload::const_ptr > &result)
  {
    static const char *here = "mpiapi::allgather";
    mpitag mytag = comm.next_collective_tag();
    assert_initialized(here);
    validsend(sendcount, sendtype, comm.rank(), mytag, comm, here);
    validrecv(recvcount, recvtype, comm.rank(), mytag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->allgather()->execute(req, queue_, sendcount, sendtype, recvcount,
        recvtype, mytag, comm, content, os_);

    timestamp fini = os_->block(req->get_key());
    mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus>(
        req->get_status());
    if (!stat)
      throw illformederror("mpiapi::allgather: NULL status pointer.");

    mpicollpayload::const_ptr coll = dynamic_pointer_cast<
        const mpicollpayload>(stat->content());

    if (!coll)
      throw illformederror("mpiapi::allgather:  NULL mpicollpayload.");

    result = coll->get_content();

    if (tracer_->is_active())
      tracer_->allgather(entry, fini, sendcount, sendtype, recvcount, recvtype,
          comm);
    return fini;
  }

  //
  // Gather non-uniformly distributed data onto all nodes.
  //
  timestamp
  mpiapi::allgatherv(int sendcnt, const mpitype &sendtype,
      const std::vector<int> &recvcnts, const mpitype &recvtype,
      const mpicomm &comm)
  {
    std::vector<payload::const_ptr > rv;
    return allgatherv(sendcnt, sendtype, recvcnts, recvtype, comm,
        payload::null(), rv);
  }

  //
  // Gather non-uniformly distributed data onto all nodes.
  //
  timestamp
  mpiapi::allgatherv(int sendcnt, const mpitype &sendtype,
      const std::vector<int> &recvcnts, const mpitype &recvtype,
      const mpicomm &comm, const payload::const_ptr& content,
      std::vector<payload::const_ptr > &result)
  {
    static const char *here = "mpiapi::allgatherv";
    mpitag mytag = comm.next_collective_tag();
    assert_initialized(here);
    validsend(sendcnt, sendtype, comm.rank(), mytag, comm, here);
    validsends(recvcnts, recvtype, comm.rank(), mytag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());

    strategy_->allgatherv()->execute(req, queue_,
        sendcnt, sendtype, recvcnts, recvtype, mytag, comm, content, os_);

    timestamp fini = os_->block(req->get_key());
    mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus>(
        req->get_status());
    if (!stat)
      throw illformederror("mpiapi::allgather: NULL status pointer.");

    mpicollpayload::const_ptr coll = dynamic_pointer_cast<
        const mpicollpayload>(stat->content());

    if (!coll)
      throw illformederror("mpiapi::allgather:  NULL mpicollpayload.");

    result = coll->get_content();

    if (tracer_->is_active())
      tracer_->allgatherv(entry, fini, sendcnt, sendtype, recvcnts, recvtype,
          comm);
    return fini;
  }

  //
  // Send uniformly distributed data from all to all processes.
  //
  timestamp
  mpiapi::alltoall(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpicomm &comm)
  {
    std::vector<payload::const_ptr > cn;
    std::vector<payload::const_ptr > rv;
    return this->alltoall(sendcount, sendtype, recvcount, recvtype, comm,
        cn, rv);
  }

  //
  // Send uniformly distributed data from all to all processes.
  //
  timestamp
  mpiapi::alltoall(int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpicomm &comm,
      const std::vector<payload::const_ptr >& content,
      std::vector<payload::const_ptr > &result)
  {
    static const char *here = "mpiapi::alltoall";
    mpitag a2atag = comm.next_collective_tag();
    assert_initialized(here);
    validsend(sendcount, sendtype, comm.rank(), a2atag, comm, here);
    validrecv(recvcount, recvtype, comm.rank(), a2atag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());

    strategy_->alltoall()->execute(req, queue_, sendcount, sendtype, recvcount,
        recvtype, a2atag, comm, content, os_);

    timestamp fini = os_->block(req->get_key());
    mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus>(
        req->get_status());
    if (!stat)
      throw illformederror("mpiapi::alltoall: NULL status pointer.");

    mpicollpayload::const_ptr pay = dynamic_pointer_cast<
        const mpicollpayload>(stat->content());

    if (!pay)
      throw illformederror("mpiapi::alltoall:  NULL mpicollpayload.");

    result = pay->get_content();

    if (tracer_->is_active())
      tracer_->alltoall(entry, fini, sendcount, sendtype, recvcount, recvtype,
          comm);
    return fini;
  }

  //
  // Send non-uniformly distributed data from all to all processors.
  //
  timestamp
  mpiapi::alltoallv(const std::vector<int> &sendcnts, const mpitype &sendtype,
      const std::vector<int> &recvcnts, const mpitype &recvtype,
      const mpicomm &comm)
  {
    std::vector<payload::const_ptr > cn(comm.size().id, payload::const_ptr());
    std::vector<payload::const_ptr > rv(comm.size().id, payload::const_ptr());
    return alltoallv(sendcnts, sendtype, recvcnts, recvtype, comm,
        cn, rv);
  }

  //
  // Send non-uniformly distributed data from all to all processors.
  //
  timestamp
  mpiapi::alltoallv(const std::vector<int> &sendcnts, const mpitype &sendtype,
      const std::vector<int> &recvcnts, const mpitype &recvtype,
      const mpicomm &comm, const std::vector<payload::const_ptr >& content,
      std::vector<payload::const_ptr > &result)
  {
    static const char *here = "mpiapi::alltoallv";
    mpitag a2avtag = comm.next_collective_tag();
    assert_initialized(here);
    validsends(sendcnts, sendtype, comm.rank(), a2avtag, comm, here);
    validrecvs(recvcnts, recvtype, comm.rank(), a2avtag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->alltoallv()->execute(req, queue_,
        sendcnts, sendtype, recvcnts, recvtype, a2avtag, comm, content, os_);

    timestamp fini = os_->block(req->get_key());
    mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus>(
        req->get_status());
    if (!stat)
      throw illformederror("mpiapi::alltoall: NULL status pointer.");

    mpicollpayload::const_ptr pay = dynamic_pointer_cast<
       const mpicollpayload>(stat->content());

    if (!pay)
      throw illformederror("mpiapi::alltoall:  NULL mpicollpayload.");

    result = pay->get_content();
    if (tracer_->is_active())
      tracer_->alltoallv(entry, fini, sendcnts, sendtype, recvcnts, recvtype,
          comm);
    return fini;
  }

  //
  // Reduce data from all nodes onto the root node.
  //
  timestamp
  mpiapi::reduce(int count, const mpitype &type, const mpiop &op,
      const mpiid &root, const mpicomm &comm)
  {
    payload::const_ptr result;
    return reduce(count, type, op, root, comm, payload::null(), result);
  }

  //
  // Reduce data from all nodes onto the root node.
  //
  timestamp
  mpiapi::reduce(int count, const mpitype &type, const mpiop &op,
      const mpiid &root, const mpicomm &comm,
      const payload::const_ptr& content,
      payload::const_ptr &result)
  {
    static const char *here = "mpiapi::reduce";
    mpitag reducetag = comm.next_collective_tag();
    assert_initialized(here);
    validsend(count, type, comm.rank(), reducetag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->reduce()->execute(req, queue_, count, type, op, root, reducetag,
        comm, content, os_);

    timestamp fini = os_->block(req->get_key());

    if (comm.rank() == root)
      {
        mpistatus::const_ptr stat = req->get_status();
        if (!stat)
          throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");

        result = stat->content();
      }
    else
      {
        result = payload::const_ptr();
      }
    if (tracer_->is_active())
      tracer_->reduce(entry, fini, count, type, op, root, comm);
    return fini;
  }

  //
  // Reduce data from all nodes to all nodes.
  //
  timestamp
  mpiapi::allreduce(int count, const mpitype &type, const mpiop &op,
      const mpicomm &comm)
  {
    payload::const_ptr result;
    return allreduce(count, type, op, comm, payload::null(), result);
  }

  timestamp
  mpiapi::allreduce(int count, const mpitype &type, const mpiop &op,
      const mpicomm &comm, const payload::const_ptr& content,
      payload::const_ptr &result)
  {
    static const char *here = "mpiapi::allreduce";
    mpitag allreducetag = comm.next_collective_tag();
    assert_initialized(here);
    validsend(count, type, comm.rank(), allreducetag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->allreduce()->execute(req, queue_, count, type, op, allreducetag,
        comm, content, os_);

    timestamp fini = os_->block(req->get_key());

    mpistatus::const_ptr stat = req->get_status();

    if (!stat)
      throw illformederror("mpiapi::bcast: NULL status pointer after bcast.");

    result = stat->content();

    if (tracer_->is_active())
      tracer_->allreduce(entry, fini, count, type, op, comm);
    return fini;
  }

  //
  // Reduce data and scatter results.
  //
  timestamp
  mpiapi::reduce_scatter(const std::vector<int> &recvcnts, const mpitype &type,
      const mpiop &op, const mpicomm &comm)
  {
    /*static const char *here = "mpiapi::reduce_scatter";
    static mpitag rscattag(0xFFAA1A);
    assert_initialized(here);
    validsends(recvcnts, type, comm.rank(), rscattag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->reduce_scatter()->execute(req, queue_,
        recvcnts, type, op, rscattag, comm, payload::null(), os_);

    timestamp fini = os_->block(req->get_key());
    //mpistatus::const_ptr stat = req->get_status();
    //if (!stat)
    //  throw illformederror("mpiapi::reduce_scatter: NULL status pointer.");
   // result = stat->content();
    //if (!result)
    //  throw illformederror("mpiapi::alltoallv:  NULL mpicollpayload.");
    if (tracer_->is_active())
      tracer_->reduce_scatter(entry, fini, recvcnts, type, op, comm);
    return fini;*/

    throw unimplementederror("mpiapi::reduce_scatter is not supported yet");
  }

  //
  // Reduce data and scatter results.
  //
  timestamp
  mpiapi::reduce_scatter(const std::vector<int> &recvcnts, const mpitype &type,
      const mpiop &op, const mpicomm &comm,
      const payload::const_ptr& content, payload::const_ptr &result)
  {
    throw unimplementederror("mpiapi::reduce_scatter with a payload is not supported yet");
  }

  //
  // Compute partial reductions on data.
  //
  timestamp
  mpiapi::scan(int count, const mpitype &type, const mpiop &op,
      const mpicomm &comm)
  {
    /*shared_ptr<const mpicollpayload> result;
 return this->scan(count, type, op, comm, payload::null(), result);*/

    throw unimplementederror("mpiapi::scan: not yet implemented in this version of SST/macro");
  }

  //
  // Compute partial reductions and aggregate data.
  //
  timestamp
  mpiapi::scan(int count, const mpitype &type, const mpiop &op,
      const mpicomm &comm, const payload::const_ptr& content,
      payload::const_ptr &result)
  {
    /*static const char *here = "mpiapi::scan";
    mpitag scantag = comm.next_collective_tag();
    assert_initialized(here);
    validsend(count, type, comm.rank(), scantag, comm, here);
    //
    timestamp entry = os_->now();
    mpirequest_t req = mpirequest::construct(key::construct());
    strategy_->scan()->execute(req, queue_, count, type, op, scantag, comm, content,
        os_);

    timestamp fini = os_->block(req->get_key());
    mpistatus::const_ptr stat = dynamic_pointer_cast<const mpistatus>(
        req->get_status());
    if (!stat)
      throw illformederror("mpiapi::scan: NULL status pointer.");
    result = dynamic_pointer_cast< mpicollpayload>(stat->content());
    if (tracer_->is_active())
      tracer_->scan(entry, fini, count, type, op, comm);
    return fini;*/

    throw unimplementederror("mpiapi::scan: not yet implemented in this version of SST/macro");
  }

  } } // end of namespace sstmac

