/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpiserver.h>

#include <sstmac/software/libraries/mpi/mpimessage.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>

#include <sstmac/common/errors.h>

namespace sstmac { namespace sw
{
  

  using boost::dynamic_pointer_cast;

  using namespace std;

  debug mpiserver::dbg_;

  std::string
  mpiserver::get_lib_name()
  {
    return lib_name(library::id_);
  }

  std::string
  mpiserver::lib_name(softwareid id)
  {
    std::stringstream ss;
    ss << "mpiserver" << id.app_.id;
    return ss.str();
  }

  //
  // Hello.
  //
  mpiserver::mpiserver( softwareid id) :
      library(id)
  {
    if(!dbg_.is_real()){
         dbg_ = debug("<debug> mpiserver | mpi | library");
         logger::register_logger(&dbg_);
     }
  }



  //
  // Hello.
  //
  mpiserver::ptr
  mpiserver::construct(
      softwareid id)
  {
    mpiserver::ptr retval(new mpiserver(id));

    return retval;
  }

  //
  // Goodbye.
  //
  mpiserver::~mpiserver() throw ()
      {
      }

  void
  mpiserver::computor::run()
  {
    while (!killed_)
      {
        while (pending_recv_.size() > 0 || pending_send_.size() > 0)
          {

            if (pending_recv_.size() > 0) // do recv's first
              {
                //first, get enough time on the processor
                size_t numcomputing = pending_recv_.size(); //save the number we are computing for, in case some arrive while we're computing

                int bytes = 0;

                list<mpimessage::ptr >::iterator it =
                    pending_recv_.begin();
                for (size_t i = 0; i < numcomputing; i++)
                  {
                    mpimessage::ptr mess = *it;
                    bytes += mess->get_byte_length();
                    it++;

                  }

                if (compute_lib_inst_->supported())
                  {
                    compute_lib_inst_->compute_block_access(bytes);
                  }
                else
                  {
                    compute_lib_time_->compute(timestamp(bytes * 1e-9));
                  }

                for (size_t i = 0; i < numcomputing; i++)
                  {
                    mpimessage::ptr mess = pending_recv_.front();
                    pending_recv_.pop_front();
                    parent_->handle_recv(mess);
                  }
              }

            if (pending_send_.size() > 0) // now do sends
              {
                //first, get enough time on the processor
                size_t numcomputing = pending_send_.size(); //save the number we are computing for, in case some arrive while we're computing

                int bytes = 0;

                list<mpimessage::ptr >::iterator it =
                    pending_send_.begin();
                for (size_t i = 0; i < numcomputing; i++)
                  {
                    mpimessage::ptr mess = *it;
                    bytes += mess->get_byte_length();
                    it++;

                  }

                if (compute_lib_inst_->supported())
                  {
                    compute_lib_inst_->compute_block_access(bytes);
                  }
                else
                  {
                    compute_lib_time_->compute(timestamp(bytes * 1e-9));
                  }

                for (size_t i = 0; i < numcomputing; i++)
                  {
                    mpimessage::ptr mess = pending_send_.front();
                    pending_send_.pop_front();
                    parent_->handle_send(mess);
                  }
              }

          }

        sleeping_ = true;
        os_->block(sleepkey_);
        sleeping_ = false;
      }
  }
  //
  // Define a task for this server.  Each task can only be defined once
  //
  void
  mpiserver::add_task(const taskid &tid, const mpiqueue::ptr& peer)
  {
    if (peer == 0)
      {
        throw nullerror("mpiserver::add_task:  null peer pointer.");
      }
    queue_t::iterator it = queue_.find(tid);
    if (it != queue_.end())
      {
        throw illformederror("mpiserver::add_task:  Attempt to redefine peer.");
      }
    queue_[tid] = peer;
  }

  //
  // Test whether the given task is defined on this node.
  //
  bool
  mpiserver::has_task(const taskid &tid) const
  {
    return (queue_.find(tid) != queue_.end());
  }

  //
  // Send data.
  //
  void
  mpiserver::send(const mpimessage::ptr& payload)
  {
    if(!compute_){
        compute_ = computor::construct(id_, mpiserver::ptr(this));
           os_->start_thread(compute_);
    }
    compute_->send(payload);
  }

  void
  mpiserver::handle(const sst_message::ptr &msg)
  {
    SSTMAC_DEBUG << "[t=" << library::os_->now() << "] -- mpiserver["
        << library::id_.to_string() << "]: received message, passing it on to computor \n";

    if(!compute_){
           compute_ = computor::construct(id_, mpiserver::ptr(this));
              os_->start_thread(compute_);
       }
    compute_->recv(msg);
  }

  void
  mpiserver::handle_send(mpimessage::ptr &payload)
  {
    SSTMAC_DEBUG << "[t=" << library::os_->now() << "] -- mpiserver["
      << library::id_.to_string() << "]: sending...\n";
    taskid recver = payload->dest_task();
    queue_t::iterator it = queue_.find(recver);
    if (it != queue_.end())
      {
        // Transfer between processes on local node.
        SSTMAC_DEBUG << "[t=" << library::os_->now() << "] -- copy to local task \n";
        it->second->incoming_message(payload);

        //push a nic ack to the source also
        mpimessage::ptr nicack = boost::dynamic_pointer_cast<mpimessage>(
            payload->clone(sst_message::SEND_ACK));

        taskid sender = payload->source_task();
        queue_t::iterator it2 = queue_.find(sender);
        if (it2 != queue_.end())
          {
            it2->second->incoming_message(nicack);
          }
      }
    else
      {
        // Network transfer.
        SSTMAC_DEBUG << "[t=" << library::os_->now()
          << "] -- executing os kernel to send remote \n";

        library::os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, payload);
      }
  }

  //
  // Complete a receive.
  void
  mpiserver::handle_recv(mpimessage::ptr &mess)
  {
    SSTMAC_DEBUG << "[t=" << library::os_->now() << "] -- mpiserver["
      << library::id_.to_string() << "]: handling recv - " << mess->to_string()
      << "\n";
    if (mess == 0)
      {
        throw illformederror("mpiserver::incoming_message:  Message is not "
            "of type mpimessage.");
      }

    queue_t::iterator it;
    if (mess->get_type() == sst_message::SEND_ACK)
      {
        taskid sender = mess->source_task();
        it = queue_.find(sender);
      }
    else
      {
        taskid recver = mess->dest_task();
        it = queue_.find(recver);
      }

    if (it == queue_.end())
      {

        std::cerr << "mpiserver[" << library::id_.to_string() << "]::incoming_message:  "
            << "Failed to find receiver " << mess->dest_task() << " or sender "
            << mess->source_task() << " in the task queues.  " << "My tasks are: [ ";
        for (queue_t::iterator a = queue_.begin(); a != queue_.end(); ++a)
          {
            std::cerr << a->first << ", ";
          }
        std::cerr << "]\n";
        throw illformederror("mpiserver::incoming_message:  Message for task "
            "not registered with this server.");

      }
    else
      {
        SSTMAC_DEBUG << "[t=" << library::os_->now() << "] -- mpiserver["
          << library::id_.to_string() << "]: delivering to queue "
          << it->first.to_string() << "\n";
        it->second->incoming_message(mess);
      }
  }

} } // end of namespace sstmac
