/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPISERVER_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPISERVER_H_INCLUDED

#include <sstmac/software/process/thread.h>
#include <sstmac/software/process/taskid.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/software/process/os_api.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/libraries/mpi/mpimessage.h>
//#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/software/libraries/compute/lib_compute_time.h>

#include <sstmac/common/ptr_type.h>

#include <map>
#include <queue>

namespace sstmac
{
  namespace sw
  {

    /// Important types.
    class mpiqueue;

    /**
     * A server that acts as a single entry/exit point for
     * mpi messages for an application instance on a node.
     * this can have multiple tasks (ranks) attached to it.
     */
    class mpiserver : public library, virtual public nocopy
    {
    public:
      typedef boost::intrusive_ptr<mpiserver> ptr;

    private:

      /// A nested type to handle the computation necessary for all mpi sends and receives, local or remote
      class localtransfer;

      class computor : public thread
      {
        mpiserver::ptr parent_;

        timestamp time_per_recv_;
        timestamp time_per_send_;

        std::list<mpimessage::ptr> pending_recv_;
        std::list<mpimessage::ptr> pending_send_;

        lib_compute_memmove::ptr compute_lib_inst_;
        lib_compute_time::ptr compute_lib_time_;

        computor(softwareid id, mpiserver::ptr p) :
            parent_(p), killed_(false), sleeping_(false), sleepkey_(
                key::construct()), //computekey_(key::construct()),// computemsg_(sst_message::construct(get_lib_name(), timestamp(0))),
            time_per_recv_(1E-8), time_per_send_(1E-8)
        {
          compute_lib_inst_ = lib_compute_memmove::construct(
              softwareid(id.app_, taskid(-1)));
          register_lib(compute_lib_inst_);

          compute_lib_time_ = lib_compute_time::construct(
              softwareid(id.app_, taskid(-1)));
          register_lib(compute_lib_time_);


        }
      public:
        typedef boost::intrusive_ptr<computor> ptr;

        static ptr
        construct(softwareid id, mpiserver::ptr p)
        {
          return ptr(new computor(id, p));
        }

        virtual void
        run();

        void
        send(const mpimessage::ptr& payload)
        {
          pending_send_.push_back(payload);

          if (sleeping_)
            {
              os_->unblock(sleepkey_);
            }
        }

        void
        recv(const sst_message::ptr& msg)
        {
          mpimessage::ptr mess = boost::dynamic_pointer_cast < mpimessage
              > (msg);

          pending_recv_.push_back(mess);

          if (sleeping_)
            {
              os_->unblock(sleepkey_);
            }
        }

        virtual void
        kill()
        {
          killed_ = true;
        }

        bool killed_;
        bool sleeping_;
        key::ptr sleepkey_;
      };

      /// The message queue for each of the peers residing on this node
      /// (indexed by their task id).
      /// These objects handle unfulfilled inbound and outbound requests,
      /// ack-ing messages, etc.
      typedef std::map<taskid, boost::intrusive_ptr<mpiqueue> > queue_t;
      queue_t queue_;

      computor::ptr compute_;

    protected:
      /// Hello.
      mpiserver(softwareid id);


      virtual void
      handle_recv(mpimessage::ptr &msg);

      virtual void
      handle_send(mpimessage::ptr &msg);

      static debug dbg_;

    public:

      /// Hello.
      static mpiserver::ptr
      construct(softwareid id);

      /// Goodbye.
      virtual
      ~mpiserver() throw ();

      /// Define a task for this server.  Each task can only be defined once
      virtual void
      add_task(const taskid &tid, const boost::intrusive_ptr<mpiqueue>& peer);

      /// Test whether the given task is defined on this node.
      virtual bool
      has_task(const taskid &tid) const;

      /// Send data.h gc
      /// The backend has the "right" to handle local transfers differently
      /// than network transfers.
      /// The data will be sent exactly as is, it is up to the queue to
      /// decide on when to do handshakes, ready-sends, etc.
      virtual void
      send(const mpimessage::ptr& payload);

      /// CALLBACK:  Complete a receive.
      /// Forwarded by the node when an incoming message arrives.
      /// This method will forward the message to the corresponding queue
      /// using mpiqueue::incoming_message.
      virtual void
      handle(const sst_message::ptr &msg);



      virtual std::string
      get_lib_name();

      appid
      get_appid()
      {
        return library::id_.app_;
      }

      /// the mpi queue should be able to find this out, to tell
      /// if it should complete sends immediately or not
      bool
      nic_send_ack_supported()
      {
        return library::os_->kernel_supported(AMI::AMI_COMM,
            AMI::AMI_COMM_NIC_SEND_ACK);
      }

      static std::string
      lib_name(softwareid id);
    };

  }
} // end of namespace sstmac.

#endif
