/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_operatingsystem_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_operatingsystem_H_INCLUDED

#include <sstmac/software/process/os_api.h>
//#include <sstmac/software/threading/context.h>
#include <sstmac/software/threading/stackalloc.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/ami/hardware_descriptor.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/software/process/thread.h>
#include <sstmac/software/process/app.h>
#include <sstmac/common/ptr_type.h>


#include <sstmac/software/threading/threading_interface.h>

#include <stack>
#include <list>
#include <string>

namespace sstmac
{
  namespace sw
  {

    class environment;
    class service;

    class operatingsystem : public OS_API
    {

      friend class service;

    protected:
      operatingsystem(hardware_descriptor::ptr &hw);

      /// This is how we keep our chucks of stack.
      static stackalloc stackalloc_;

      /// The currently active position in the list.
      //  threadcontext_t *current_;

      static debug dbg_;

      /// The caller context (main DES thread).  We go back
      /// to this context on every context switch.
      threading_interface *des_context_;

      void
      add_thread(const thread::ptr &t);

      void
      switch_to_thread(threading_interface *tothread);

    public:
      typedef boost::intrusive_ptr<operatingsystem> ptr;

      virtual
      ~operatingsystem();

      static operatingsystem::ptr
      construct(hardware_descriptor::ptr &hw);
      // --------------------------------------//

    public:
      // ---- OS_API functions ------//
      // -- these are exposed to libraries --
      virtual void
      execute_kernel(AMI::AMI_CAT cat, int kernelName, sst_message::ptr data);
      virtual ptr_type::ptr
      execute_kernel(AMI::AMI_CAT cat, int kernelName);

      virtual bool
      kernel_supported(AMI::AMI_CAT cat, int kernelName);

      virtual timestamp
      block(const key::ptr &req);
      virtual timestamp
      unblock(const key::ptr &req);
      virtual void
      start_thread(const thread::ptr &t);
      //virtual void
      //join_thread(const thread::ptr &t, threadcontext_t con);
      virtual void
      complete_thread(thread *t);

      virtual void
      register_lib(const library::ptr &lib);

      virtual void
      register_ser(const boost::intrusive_ptr<service> &lib);

      virtual library::ptr
      get_lib(std::string name);

      virtual timestamp
      now();

      virtual nodeaddress::ptr
      get_task_addr(softwareid sid);

      virtual environment::const_ptr
      get_env(appid aid)
      {

        if (env_.find(aid) == env_.end())
          {
            std::cout << "appid - " << aid.to_string() << std::endl;
            throw nullerror("os: environment not found");
          }
        return env_[aid];
      }

      // ---------------------

      // ---- functions which can be called by services ------- //
      virtual nodeaddress::ptr
      get_my_addr();

      // --- functions here to be called by node / core / hardware --- //

      void
      setAMI(AMI::AMI_CAT cat, AMI* a);

      void
      add_application(app::ptr &a, const std::list<std::string> &params);

      void
      set_environment(const environment::const_ptr &e, appid aid)
      {
        SSTMAC_DEBUG << "OS: setting environment for app " << aid.to_string()
              << "\n";
        env_[aid] = e;
      }

      void
      start_apps();

      void
      handle_message(const sst_message::ptr &msg);

      void
      apps_kill();



      // -------------------------------------- //

      virtual threading_interface*
      get_current_context()
      {
        return threadstack_.top();
      }
    protected:
      // scheduler scheduler_;
      ///hardware_descriptor::ptr hw_desc_;

      std::map<sstmac::sw::appid, sstmac::sw::environment::const_ptr> env_;

      AMI*
      getAMI(AMI::AMI_CAT cat);

      std::map<AMI::AMI_CAT, AMI*> amis_;
      std::map<std::string, eventhandler::ptr> libs_;

      std::list<thread::ptr> threads_;

      std::stack<threading_interface*> threadstack_;

      std::deque<sst_message::ptr> pending_messages_;

      // threading_interface *threader_;
    };

  }
} //end of namespace sstmac
#endif
