/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <algorithm>
#include <exception>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/common/errors.h>
#include <iterator>

namespace sstmac { namespace sw
  {

  //////////////////////////////////////////////////////////////////////
  // allocationstrategy

  debug allocationstrategy::dbg_;

  allocationstrategy::~allocationstrategy() throw()
  {
  }

  void
  allocationstrategy::check_request(int nnode_requested,
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation)
  {
    if(nnode_requested <= 0
       || nnode_requested > available_.size()) {
      std::cerr << "allocationstrategy:  Request for "
                << nnode_requested
                << " nodes when " << available_.size()
                << " are available.\n";
      throw valueerror("allocationstrategy: Not enough nodes available.");
    }
  }

  //////////////////////////////////////////////////////////////////////
  // firstavailableallocation

  firstavailableallocation::firstavailableallocation(long num_avail) : allocationstrategy(num_avail)
  {
  }

  firstavailableallocation::~firstavailableallocation() throw()
  {
  }

  boost::intrusive_ptr<firstavailableallocation>
  firstavailableallocation::construct(long num_avail)
  {
    return boost::intrusive_ptr<firstavailableallocation>(new firstavailableallocation(num_avail));
  }

  void
  firstavailableallocation::allocate(int nnode_requested,
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation)
  {
    check_request(nnode_requested,allocation);

   // allocation.resize(nnode_requested);
    for (int i=0; i<nnode_requested; i++) {
        SSTMAC_DEBUG << "firstavailable_: i=" << i << "\n";
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp>::const_iterator nextnode = available_.begin();
      SSTMAC_DEBUG << "firstavailable_: nextnode= " << (*nextnode)->to_string() << "\n";
      allocation.insert(*nextnode);
      allocated_.insert(*nextnode);
      available_.erase(nextnode);
    }
  }

  //////////////////////////////////////////////////////////////////////
  // randomallocation

  randomallocation::randomallocation(long num_avail, unsigned seed) : allocationstrategy(num_avail)
  {
    rng_ = RNG::SimpleCombo::construct(seed);
  }

  randomallocation::randomallocation
  (long num_avail, const RNG::UniformInteger::ptr &rng) : allocationstrategy(num_avail)
  {
    rng_ = rng;
  }

  randomallocation::~randomallocation() throw()
  {
  }

  boost::intrusive_ptr<randomallocation>
  randomallocation::construct(long num_avail, unsigned seed)
  {
    return boost::intrusive_ptr<randomallocation>(new randomallocation(num_avail, seed));
  }

  boost::intrusive_ptr<randomallocation>
  randomallocation::construct(long num_avail, const RNG::UniformInteger::ptr &rng)
  {
    return boost::intrusive_ptr<randomallocation>(new randomallocation(num_avail, rng));
  }

  void
  randomallocation::allocate(int nnode_requested,
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation)
  {
    check_request(nnode_requested,allocation);
    /*
    std::cout << "randomallocation::allocate:"
              << " nnode_requested = " << nnode_requested
              << " available_.size() = " << available_.size()
              << " allocated_.size() = " << allocated_.size()
              << std::endl;
    */
    //allocation.resize(nnode_requested);
    std::vector<nodeaddress::ptr> availvec(available_.size());
    std::copy(available_.begin(), available_.end(), availvec.begin());
    RNG::UniformInteger_functor rngf(rng_);
    std::random_shuffle(availvec.begin(), availvec.end(), rngf);
    for (int i=0; i<nnode_requested; i++) {
      nodeaddress::ptr node = availvec[i];
      allocation.insert(node);
      allocated_.insert(node);
      available_.erase(node);
    }
    // we only want to randomly pick nodes, not randomize the nodes
    // picked, so sort them
    //std::sort(allocation.begin(), allocation.end());
  }
  
  //////////////////////////////////////////////////////////////////////
  // fixedallocation
  
  fixedallocation::fixedallocation( const std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &nodes): allocationstrategy(0),
  nodes_(nodes)
  {
  }
  
  fixedallocation::~fixedallocation() throw()
  {
  }
  
  boost::intrusive_ptr<fixedallocation>
  fixedallocation::construct( const std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &nodes)
  {
    return boost::intrusive_ptr<fixedallocation>(new fixedallocation(nodes));
  }

  void
  fixedallocation::allocate(int nnode_requested,
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation)
  {
   /* static std::string here("fixedallocation::allocate: ");
    check_request(nnode_requested,allocation);
    if (nnode_requested != nodes_.size()) {
    	std::cout << "nnode_requested: " << nnode_requested << ", nodes.size: " << nodes_.size() << std::endl;
      throw illformederror(here + "nnode_requested"
                           " does not match size of node vector");
    }
    for (int i=0; i<nnode_requested; i++) {

      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp>::iterator node = available_.find(nodes_[i]);
      if (node == available_.end()) {
          std::cout << "looking for: " << nodes_[i]->to_string() << std::endl;
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp>::iterator it, end = available_.end();
          for(it = available_.begin(); it != end; it++){
              std::cout << "available_: " << (*it)->to_string() << "\n";
          }
          throw illformederror(here + "required node not available_");
      }
      available_.erase(node);
      allocated_.insert(nodes_[i]);
      allocation.insert(*node);
    }*/

    allocation = nodes_;
  }

} } //end of namespace sstmac
