/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_MINI_MINIMD_MINIMD_COMM_H_INCLUDED
#define SSTMAC_SOFTWARE_SKELETONS_MINI_MINIMD_MINIMD_COMM_H_INCLUDED

#include <sstmac/software/skeletons/mini/minimd/minimd.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-atom.h>
#include <sstmac/software/skeletons/mini/minimd/lib_compute_minimd.h>
#include <sstmac/software/libraries/mpi/mpiapi.h>

namespace sstmac { namespace sw {

  

  /**
   * Mimic the communication routines in miniMD.
   */
  class minimd::comm  : public ptr_type {
  private:
    lib_compute_minimd::ptr execution_;
    mpiapi::ptr mpi_;

    template <typename T>
    T min(T a, T b) {
      return (a < b ? a : b);
    }
    template <typename T>
    T max(T a, T b) {
      return (a > b ? a : b);
    }

  public:
    virtual std::string to_string() const{
            return "comm";
          }

    typedef boost::intrusive_ptr<comm> ptr;

    std::vector<mpiid> sendproc; // of dimension nswap
    std::vector<mpiid> recvproc; // of dimension nswap
    std::vector<int> comm_send_size, comm_recv_size;
    std::vector<int> comm_reverse_send_size, comm_reverse_recv_size;
    int nswap; // the number of swaps to perform
    mpiid procneigh[3][2]; // the 6 nearest neighbors.
    int procgrid[3]; // the processor grid.
    int need[3]; // how many processors we need in each dimension
 
    void init(const lib_compute_minimd::ptr& exe,
              const mpiapi::ptr& mpi)
    {
      nswap = -1;
      for(int r = 0; r < 3; ++r) {
        procgrid[r] = -100;
        need[r] = -100;
        for(int c = 0; c < 2; ++c)
          procneigh[r][c] = mpiid(-100);
      }
      execution_ = exe;
      mpi_ = mpi;
    }

    /// Setup spatial decomposition.
    void setup(double cutneigh, const atom::ptr& atm) {
      mpicomm world = mpi_->comm_world();
      int nprocs = world.size().id;
      mpiid me = world.rank();
      int rank = me.id;
      const double prd[] = { atm->box.xprd, atm->box.yprd, atm->box.zprd };
      double area[] = { prd[0]*prd[1], prd[0]*prd[2], prd[1]*prd[2] };
      double bestsurf = 2 * area[0] * area[1] * area[2];
      // Figure out the best factorization for the grid.
      for(int ipx = 1; ipx <= nprocs; ++ipx) {
        if(nprocs % ipx == 0) {
          int nremain = nprocs / ipx;
          for(int ipy = 1; ipy <= nprocs; ++ipy) {
            if(nremain % ipy == 0) {
              int ipz = nremain / ipy;
              double surf = area[0]/ipx/ipy + area[1]/ipx/ipz + area[2]/ipy/ipz;
              if(surf < bestsurf) {
                bestsurf = surf;
                procgrid[0] = ipx;
                procgrid[1] = ipy;
                procgrid[2] = ipz;
              }
            }
          }
        }
      }
      // I am not using the cart comm trick from miniMD because I still haven't
      // implemented the needed mpi functionality in the simulator.
      // Instead, we simply stick the nodes on the grid in sequential order.
      // Let
      //    stride = [procgrid[1]*procgrid[2], procgrid[2], 1]
      // A node with rank r has the coordinates:
      //    xpos = rank / stride[0]
      //    ypos = (rank - xpos*stride[0]) / stride[1]
      //    zpos = rank - xpos*stride[0] - ypos*stride[1]
      // Similarly, the grid location [xpos, ypos, zpos] has the rank
      //    xpos*stride[0] + ypos*stride[1] + zpos*stride[2]
      // Functionally, this has the same effect as a "vanilla" MPI_Cart_shift
      // with periodic boundaries under mpich2.
      int stride[] = { procgrid[1]*procgrid[2], procgrid[2], 1 };
      int myloc[] = { rank/stride[0],
                       (rank - myloc[0]*stride[0]) / stride[1],
                       rank - myloc[0]*stride[0] - myloc[1]*stride[1] };
      int nb[3] = {myloc[0], myloc[1], myloc[2]};
      // shift.  Make "source" the downward direction and "dest" the upward.
      for(int axis = 0; axis < 3; ++axis) {
        nb[axis] = (myloc[axis] > 0 ? myloc[axis]-1 : procgrid[axis]-1);
        procneigh[axis][0] = mpiid(nb[0]*stride[0]+
                                   nb[1]*stride[1]+
                                   nb[2]*stride[2]);
        nb[axis] = (myloc[axis] < procgrid[axis]-1 ? myloc[axis]+1 : 0);
        procneigh[axis][1] = mpiid(nb[0]*stride[0]+
                                   nb[1]*stride[1]+
                                   nb[2]*stride[2]);
        nb[axis] = myloc[axis];
      }
#if 0
      // Debug dump.
      std::cerr << me << " dumping procneigh:\n";
      for(int axis = 0; axis < 3; ++axis) {
        std::cerr << "    procneigh[" << axis << "] = [" << procneigh[axis][0]
                  << ", " << procneigh[axis][1] << "]\n";
      }
#endif // !0
      // The number of boxes from which I need information in each direction.
      for(int axis = 0; axis < 3; ++axis) {
        need[axis] = int(cutneigh * procgrid[axis] / prd[axis] + 1);
      }
      // Set up parameters for each exchange
      nswap = 0;
      for(int idim = 0; idim < 3; ++idim) {
        for(int ineed = 0; ineed < 2*need[idim]; ++ineed) {
          if(sendproc.size() <= size_t(nswap)) sendproc.resize(nswap+1);
          if(recvproc.size() <= size_t(nswap)) recvproc.resize(nswap+1);
          if(ineed % 2 == 0) {
            sendproc.at(nswap) = procneigh[idim][0];
            recvproc.at(nswap) = procneigh[idim][1];
          } else {
            sendproc.at(nswap) = procneigh[idim][1];
            recvproc.at(nswap) = procneigh[idim][0];
          }
          nswap++;
        }
      }
      // finally, account for all the computation and allocations.
      //mpi_->compute(execution_->get("Comm::setup"));
      //unsigned int seedp = rank;
      //double randnoise = (double(rand_r(&seedp)) / RAND_MAX - 0.5) * 0.25;
      timestamp t(1.024e-05);
      execution_->compute(t);
    }

    /// Forward communicate (at every timestep)
    void communicate(const atom::ptr&) {
      static const mpitag tag(0);
      mpicomm world = mpi_->comm_world();
      mpiid rank = world.rank();
      mpirequest_t request;
      const_mpistatus_t status;
      for (int iswap = 0; iswap < nswap; iswap++) {
        if (sendproc[iswap] != rank) {
          mpi_->irecv(comm_recv_size.at(iswap), mpitype::mpi_double,
                      recvproc.at(iswap), tag, world ,request);
          mpi_->send(comm_send_size.at(iswap), mpitype::mpi_double,
                     sendproc.at(iswap), tag, world);
          mpi_->wait(request,status);
        }
      }
    }

    /// Reverse communicate (at every timestep).
    void reverse_communicate(const atom::ptr&) {
      mpicomm world = mpi_->comm_world();
      mpiid me = world.rank();
      mpirequest_t request;
      const_mpistatus_t status;
      for (int iswap = 0; iswap < nswap; iswap++) {
        if (sendproc[iswap] != me) {
          mpi_->irecv(comm_reverse_recv_size.at(iswap), mpitype::mpi_double,
                      sendproc[iswap], mpitag(0), world, request);
          mpi_->send(comm_reverse_send_size.at(iswap), mpitype::mpi_double,
                     recvproc.at(iswap), mpitag(0), world);
          mpi_->wait(request,status);
        }
      }
    }
    
    /// Exchange atoms that have left their boxes
    /// This will be guessestimated statistically.
    void exchange(const atom::ptr& atm) {
//#define SSTMAC_DB(command) std::cerr << rank << " " << #command << "\n"; command
#define SSTMAC_DB(command) command
      static const mpitag tag(0);
      mpicomm world = mpi_->comm_world();
      mpiid rank = world.rank();
      SSTMAC_DB( mpirequest_t request );
      SSTMAC_DB( const_mpistatus_t status );
      for(int idim = 0; idim < 3; ++idim) {
        // Only exchange if there's more than one processor in this dimension
        if(procgrid[idim] <= 1)
          continue;
        // Filling buffers etc.
        //SSTMAC_DB( mpi_->compute(execution_->get("Comm::exchange",  0,  0)) );
        // This is fast enough that we don't even worry about it (~ 1 us).
        // Get the estimated paramters.
        /// nsend, nrecv, and nrecv2 are the lengths of packed arrays of
        /// coordinate and velocity values.  It looks like the sends
        /// are almost always empty (based on trace files), which is
        /// very strange.
        int nsend = (6 * atm->nlocal) / 50;
        int nrecv1 = (6 * atm->neighatoms.at(procneigh[idim][1].id)) / 50;
        int nrecv2 = (6 * atm->neighatoms.at(procneigh[idim][0].id)) / 50;
        // The actual exchange.  First send nsend and receive nrecv1
        mpi_->send(1, mpitype::mpi_int, procneigh[idim][0], tag, world);
        mpi_->recv(1, mpitype::mpi_int, procneigh[idim][1], tag, world, status);
        if (procgrid[idim] > 2) {
          // send nsend and receive nrecv2
          mpi_->send(1,mpitype::mpi_int,procneigh[idim][1],tag,world);
          mpi_->recv(1,mpitype::mpi_int,procneigh[idim][0],tag,world,status);
        }
        // Exchange packed data arrays
        SSTMAC_DB( mpi_->irecv(nrecv1, mpitype::mpi_double, procneigh[idim][1], tag, 
                    world, request) );
        SSTMAC_DB( mpi_->send(nsend, mpitype::mpi_double, procneigh[idim][0], tag, world) );
        SSTMAC_DB( mpi_->wait(request, status) );
        
        if (procgrid[idim] > 2) {
          // Exchange packed data arrays in the other direction
          SSTMAC_DB( mpi_->irecv(nrecv2, mpitype::mpi_double, procneigh[idim][0], tag, 
                      world, request) );
          SSTMAC_DB( mpi_->send(nsend, mpitype::mpi_double, procneigh[idim][1], tag, world) );
          SSTMAC_DB( mpi_->wait(request, status) );
        }
      }
#undef SSTMAC_DB
    }

    /// Make a list of border atoms to send to neighboring procs.
    void borders(const atom::ptr&) {
      static const mpitag tag(0);
      mpicomm world = mpi_->comm_world();
      mpiid rank = world.rank();
      mpirequest_t request;
      const_mpistatus_t status;
//#define SSTMAC_DB(command) std::cerr << rank << " " << #command << "\n"; command
#define SSTMAC_DB(command) command
      // Do swaps over 3 dimensions.
      int iswap = 0;
      for(int idim = 0; idim < 3; ++idim) {
        for(int ineed = 0; ineed < 2*need[idim]; ++ineed) {
          // Get the estimated paramters.
          SSTMAC_DB( int nsend = execution_->get_int("Comm::borders::nsend", iswap) );
          SSTMAC_DB( int nrecv = execution_->get_int("Comm::borders::nrecv", iswap) );
          if(comm_send_size.size() <= size_t(iswap)) {
            SSTMAC_DB( comm_send_size.resize(iswap+1) );
          }
          if(comm_recv_size.size() <= size_t(iswap)) {
            SSTMAC_DB( comm_recv_size.resize(iswap+1) );
          }
          if(comm_reverse_send_size.size() <= size_t(iswap)) {
            SSTMAC_DB( comm_reverse_send_size.resize(iswap+1) );
          }
          if(comm_reverse_recv_size.size() <= size_t(iswap)) {
            SSTMAC_DB( comm_reverse_recv_size.resize(iswap+1) );
          }
          SSTMAC_DB( comm_send_size.at(iswap) = nsend * 3 ); // Atom::comm_size is 3
          SSTMAC_DB( comm_recv_size.at(iswap) = nrecv * 3 ); // Same as above.
          SSTMAC_DB( comm_reverse_send_size.at(iswap) = nrecv * 3 ); // Same as above.
          SSTMAC_DB( comm_reverse_recv_size.at(iswap) = nsend * 3 ); // Same as above.
          SSTMAC_DB( mpi_->send(1, mpitype::mpi_int, sendproc.at(iswap), tag, world) );
          SSTMAC_DB( mpi_->recv(1, mpitype::mpi_int, recvproc.at(iswap), tag, world, status) );
          SSTMAC_DB( mpi_->irecv(nrecv*3, mpitype::mpi_double, 
                          recvproc.at(iswap), tag, world, request) );
          SSTMAC_DB( mpi_->send(nsend*3, mpitype::mpi_double, 
                         sendproc.at(iswap), tag, world) );
          SSTMAC_DB( mpi_->wait(request, status) );
          SSTMAC_DB( ++iswap );
        }
      }
#undef SSTMAC_DB
    }
  };

} } // end of namespace sstmac

#endif
