/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/osu/osubw.h>
#include <sstmac/software/libraries/mpi/mpiapi.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

namespace sstmac { namespace sw {

  //
  // Hi.
  //
  osubw::osubw(softwareid id) :
    mpiapp(id), max_req_num(1000), max_alignment(65536), max_msg_size(1<<22),
    loop(100), window_size(64), skip(10), loop_large(20), window_size_large(64),
    skip_large(2), large_message_size(8192)
  {}

  //
  // It's creation time.
  //
  osubw::ptr osubw::construct(softwareid id) {
    return new osubw(id);
  }

  //
  // Goodbye.
  //
  osubw::~osubw() throw() {
  }

  //
  // Get a copy.
  //
  app::ptr osubw::clone(softwareid id) {
    osubw::ptr myclone = construct(id);
    myclone->max_req_num = this->max_req_num;
    myclone->max_alignment = this->max_alignment;
    myclone->max_msg_size = this->max_msg_size;
    myclone->loop = this->loop;
    myclone->window_size = this->window_size;
    myclone->skip = this->skip;
    myclone->loop_large = this->loop_large;
    myclone->window_size_large = this->window_size_large;
    myclone->skip_large = this->skip_large;
    myclone->large_message_size = this->large_message_size;
    return myclone;
  }

  //
  // Go.
  //
  void osubw::skeleton_main() {
    int i, j, size;
    mpiid myid, numprocs;
    timestamp t_start, t_end, t;
    static const mpiid zeroid(0), oneid(1);
    int winsize = window_size;
    int currloop = loop;

    FILE *fout = stdout;
    
    mpi()->init();
    mpicomm world = mpi()->comm_world();

    myid = world.rank();
    numprocs = world.size();

    //align_size = getpagesize();
    //assert(align_size <= MAX_ALIGNMENT);

    if(numprocs.id != 2) {
      if(myid == zeroid) {
        fprintf(stderr, "osubw: requires exactly two processes\n");
      }

      mpi()->finalize();
      return;
    }

    if(myid == zeroid) {
      fprintf(fout, "# %s %s\n", "SST/macro OSU bandwidth", "v3.1.1");
      fprintf(fout, "%-9s %20s\n", "# Size", "Bandwidth (MB/s)");
      fflush(fout);
    }

    /* Bandwidth test */
    for(size = 1; size <= max_msg_size; size *= 2) {
      if(size > large_message_size) {
        currloop = loop_large;
        skip = skip_large;
        winsize = window_size_large;
      }
      if(myid == zeroid) {
        for(i = 0; i < currloop + skip; i++) {
          std::vector<mpirequest_t> request;
          if(i == skip) {
            t_start = mpi()->wtime();
          }
          for(j = 0; j < winsize; j++) {
            mpirequest_t req;
            mpi()->isend(size, mpitype::mpi_char, oneid, mpitag(100), world,
                         req);
            request.push_back(req);
          }
          std::vector<const_mpistatus_t> reqstat(1);
          mpi()->waitall(request, reqstat);
          mpi()->recv(4, mpitype::mpi_char, oneid, mpitag(101), world,
                      reqstat.at(0));
        }

        t_end = mpi()->wtime();
        t = t_end - t_start;
      }

      else if(myid == oneid) {
        std::vector<mpirequest_t> request;
        for(i = 0; i < currloop + skip; i++) {
          for(j = 0; j < winsize; j++) {
            mpirequest_t req;
            mpi()->irecv(size, mpitype::mpi_char, zeroid, mpitag(100),
                         world, req);
            request.push_back(req);
          }
          std::vector<const_mpistatus_t> reqstat(1);
          mpi()->waitall(request, reqstat);
          mpi()->send(4, mpitype::mpi_char, zeroid, mpitag(101), world);
        }
      }

      if(myid == zeroid) {
        double tmp = size / 1e6 * currloop * winsize;

        fprintf(fout, "%-9d %20.2f\n", size, tmp / t.sec());
        fflush(fout);
      }
    }
 
    mpi()->finalize();
  }

  //
  // This object gets access to the mpi api.
  //
  void
  osubw::consume_params(const std::list<std::string> &params)
  {
  }

} } // end of namespace sstmac.
