/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/sstmac.h>
#include <unistd.h> // getopt
#include <sstmac_testutil.h>
#include <gtest/gtest.h>
#include <stdlib.h>
#include <stdio.h>




using namespace sstmac::hw;
using namespace sstmac;
using namespace sstmac::native;
using namespace sstmac::sw;
using namespace std;

void
meshnet(double latency, long int bandwidth, int n1, int n2, int n3,
    interconnect::ptr &fab, topology::ptr &net)
{
  net = composite::construct(line::construct(n1),
      composite::construct(line::construct(n2), ring::construct(n3)));
  fab = circuit::construct(timestamp(latency), bandwidth, net);
}
void
fullnet(double latency, long int bandwidth, interconnect::ptr &fab,
    topology::ptr &net)
{
  fab = fullnetwork::construct(timestamp(latency), bandwidth);
}

void
fatnet(double latency, long int bandwidth, int levels, int radix,
    interconnect::ptr &fab, topology::ptr &net)
{
  net = fattree::construct(levels, radix / 2);
  fab = circuit::construct(timestamp(latency), bandwidth, net);
}


std::string
get_token(std::string &tokens, int *narg = 0)
{
  std::string result;

  if (tokens.size() != 0) {
      size_t pos = tokens.find_first_of("_(),");

      if (pos == std::string::npos) {
          result = tokens;
          tokens.clear();
      }
      else {
          result = tokens.substr(0,pos);
          tokens = tokens.substr(pos+1,std::string::npos);
      }
  }

  if (result.size() > 0 && narg) *narg = *narg + 1;

  return result;
}




enum FLOWER_TYPES {ORIGINAL_FLOWER, ORDERED_FLOWER, THROTTLE_FLOWER };

interconnect::ptr
construct_network(std::vector<double> &args, topology::ptr top, const std::string &netname)
{
  timestamp latency;
  double bw;


  if(args.size() < 2){
      std::cout << "construct_network - must provide latency, bandwidth" << std::endl;
      exit(1);
  }

  latency = timestamp(args[0]);
  bw = args[1];


  if (netname == "fullnetwork") {
      if(args.size() < 3){
          std::cout << "construct_network - fullnetwork - must provide latency, bandwidth, and size" << std::endl;
          exit(1);
      }

      int size = args[2];
      return sstmac::hw::fullnetwork::construct(latency, bw, size);
  }else if(netname == "circuit"){
      return sstmac::hw::circuit::construct(latency, bw, top);
  }else if(netname == "sharedcircuit"){
      if(args.size() < 3){
          std::cout << "construct_network - sharedcircuit - must provide latency, bandwidth, and flower name" << std::endl;
          exit(1);
      }

      flower::ptr flow;
      if(args[2] == ORDERED_FLOWER){
          if(args.size() < 4){
              std::cout << "construct_network - sharedcircuit - assuming bandwidth is node_router bandwidth, must also supply router_router bandwidth" << std::endl;
              exit(1);
          }
          flow = ordered_flower::construct(top, bw, args[3]);
      }else if(args[2] == THROTTLE_FLOWER){
          if(args.size() < 4){
              std::cout << "construct_network - sharedcircuit - assuming bandwidth is node_router bandwidth, must also supply router_router bandwidth" << std::endl;
              exit(1);
          }
          flow = throttle_flower::construct(top, bw, args[3]);
      }else{
          std::cout << "construct_network - unknown flower: " << args[3] << std::endl;
          exit(1);
      }

      return sharedcircuit::construct(latency, bw, top, flow);
  }else if(netname == "sharedbuffer"){
      if(args.size() < 7){
          std::cout << "construct_network - for sharedbuffer must provide router_router_bw, node_router_bw, " << std::endl;
          std::cout << "\t node_lat, link_lat, router_lat, buffer_capacity, num_vcs" << std::endl;
          exit(1);
      }

      return sharedbuffer::construct(args[2], bw, latency, timestamp(args[3]), timestamp(args[4]), args[5], args[6], top);
  }else{
      std::cout << "construct_network - unknown interconnect name: " << netname << std::endl;
      exit(1);
  }

}


struct test_results : public ptr_type {
  timestamp end_;
  int numsent_;
  int numrecv_;
  int nummsgs_;

  test_results() : end_(timestamp(0)), numsent_(0), numrecv_(0), nummsgs_(0) {}

  virtual std::string
  to_string() const{
    return "testnetworks::test_results";
  }
};

class message_sent_handler : public eventhandler
{
  eventmanager::ptr em_;
  test_results* res_;

  message_sent_handler(eventmanager::ptr e, test_results* res) : em_(e), res_(res) {}

public:
  virtual void
  handle(const sst_message::ptr &msg)
  {
    // std::cout << em_->now() << ": sent from " << msg->fromaddr_->to_string() << " to " << msg->toaddr_->to_string() << std::endl;
    res_->numsent_++;
  }

  virtual std::string
  to_string() const
  {
    return "test_networks::message_sent_handler";
  }

  static boost::intrusive_ptr<message_sent_handler>
  construct(eventmanager::ptr e, test_results* res)
  {
    boost::intrusive_ptr<message_sent_handler> ret(new message_sent_handler(e, res));
    return ret;
  }

};



class message_recv_handler : public eventhandler
{
  eventmanager::ptr em_;
  test_results* res_;

  message_recv_handler(eventmanager::ptr e, test_results* res) : em_(e), res_(res) {}

public:
  virtual void
  handle(const sst_message::ptr &msg)
  {

    // std::cout << em_->now() << ": recv from " << msg->fromaddr_->to_string() << " to " << msg->toaddr_->to_string() << std::endl;
    res_->numrecv_++;
  }

  virtual std::string
  to_string() const
  {
    return "test_networks::message_recv_handler";
  }



  static boost::intrusive_ptr<message_recv_handler>
  construct(eventmanager::ptr e, test_results* res)
  {
    boost::intrusive_ptr < message_recv_handler
    > ret(new message_recv_handler(e, res));
    return ret;
  }



};

class test_message : public sst_message {

  static int ID_CNT;
  int msg_id_;
  test_message() : sst_message("dummy", timestamp(0), sst_message::SST, 0){
    msg_id_ = ID_CNT++;
  }

public:
  static boost::intrusive_ptr<test_message>
  construct(){
    boost::intrusive_ptr<test_message> ret(new test_message());
    return ret;
  }

  int bytes_;

  virtual int
  get_byte_length() const
  {
    return bytes_;
  }

  virtual std::string
  to_string() const
  {
    return "test_networks::test_message";
  }

  virtual sst_message::ptr
  clone() const
  {
    boost::intrusive_ptr<test_message> ret = construct();
    ret->bytes_ = bytes_;
    ret->fromaddr_ = fromaddr_;
    ret->toaddr_ = toaddr_;
    return ret;
  }

  virtual sst_message::ptr
  clone(MESSAGE_TYPES t) const
  {
    boost::intrusive_ptr<test_message> ret = construct();
    ret->bytes_ = bytes_;
    ret->msgtype_ = msgtype_;
    ret->fromaddr_ = fromaddr_;
    ret->toaddr_ = toaddr_;
    return ret;
  }
};

int test_message::ID_CNT = 0;

void
test_network(interconnect::ptr net, std::string testfile, test_results* res){


  eventmanager::ptr em = eventmap::construct();



  net->set_eventmanager(em);

  boost::intrusive_ptr<message_sent_handler> sent_handler = message_sent_handler::construct(em, res);
  boost::intrusive_ptr<message_recv_handler> recv_handler = message_recv_handler::construct(em, res);

  //for(int i = 0; i < testfiles.size(); i++){
  //std::cout << "running test " << testfile << std::endl;
  std::string line;
  std::ifstream myfile;
  myfile.open(testfile.c_str());
  if(!myfile.is_open()){
      std::cout << "testnetworks::test_network -- ERROR - cannot open test file: " << testfile << std::endl;
      exit(1);
  }


  while(!myfile.eof()){
      std::getline(myfile, line);
      std::istringstream iss(line);
      std::string temp;
      iss >> temp;
      timestamp when = timestamp(atof(temp.c_str()));
      iss >> temp;
      nodeid::ptr from = nodeid::construct(atoi(temp.c_str()));
      iss >> temp;
      nodeid::ptr to = nodeid::construct(atoi(temp.c_str()));
      iss >> temp;
      int size = atoi(temp.c_str());

      boost::intrusive_ptr<test_message> the_msg = test_message::construct();
      the_msg->fromaddr_ = from;
      the_msg->toaddr_ = to;
      the_msg->bytes_ = size;
      res->nummsgs_++;

      net->send(when, the_msg, sent_handler, recv_handler);

  }

  em->run();
  res->end_ = em->now();

  // std::cout << "done with test " << testfile << std::endl;

  em->clear(timestamp(0));

  //}

  //std::cout << "testing complete" << std::endl;
}

void
print_help(const char *name)
{
  std::cout << name << " [-h] [-f file] [-p file_prefix] [-d debug_string]" << std::endl;
}

std::vector<std::string> test_file_names;
std::string prefix;

int
main(int argc, char **argv)
{
  int opt;
  int nodes = 10;

  ::testing::InitGoogleTest(&argc, argv);

  std::vector<int> top_args;

  std::vector<string> files;
  bool helpflag = false;
  while(!helpflag && (opt = getopt(argc, argv, "hf:p:d:")) != -1) {
      switch(opt) {
      case 'h':
        helpflag = true;
        break;
      case 'f':
        // files.push_back(optarg);
        test_file_names.push_back(optarg);
        break;

      case 'p':
        prefix = optarg;
        prefix += "/";
        break;

      case 'd':
        logger::set_user_param(optarg);
        break;

      default:
        std::cerr << "Unknown argument -" << opt << "\n";
        helpflag = true;
      }
  }
  if (helpflag) {
      print_help(argv[0]);
  }


  if(test_file_names.size() == 0){
      test_file_names.push_back(prefix+"testnetworks_tests/all2all.txt");
      test_file_names.push_back(prefix+"testnetworks_tests/hotspot.txt");
      test_file_names.push_back(prefix+"testnetworks_tests/neighbor.txt");
      test_file_names.push_back(prefix+"testnetworks_tests/tornado.txt");
      test_file_names.push_back(prefix+"testnetworks_tests/random.txt");
  }

  return RUN_ALL_TESTS();


}

TEST(testnetworks, fullnetwork)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.000011000000));
  expected.push_back(timestamp(1.000101000000));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000101000000));

  for(int i = 0; i < test_file_names.size(); i++){
      topology::ptr top;

      std::vector<double> net_args;
      net_args.push_back(1e-6);
      net_args.push_back(1e9);
      net_args.push_back(16);
      interconnect::ptr net = construct_network(net_args, top, "fullnetwork");

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "fullnetwork, t=" << res->end_  << ", expecting " << expected[i] << std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-9);
      if(!good){
          std::cout << "ERROR: expected: " << expected[i] << ", got: " << res->end_ << std::endl;
      }
      EXPECT_TRUE(good);
      EXPECT_TRUE(res->nummsgs_ == res->numsent_);
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_);
      delete res;
  }

}

TEST(testnetworks, circuit)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.000011000000));
  expected.push_back(timestamp(1.000808000000));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000004000000));
  expected.push_back(timestamp(1.000101000000));

  for(int i = 0; i < test_file_names.size(); i++){


      topology::ptr top = torus::construct_2D(4, 4);

      std::vector<double> net_args;
      net_args.push_back(1e-6);
      net_args.push_back(1e9);

      interconnect::ptr net = construct_network(net_args, top, "circuit");

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "circuit, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-9);
      if(!good){
          std::cout << "ERROR: expected: " << expected[i] << ", got: " << res->end_ << std::endl;
      }
      EXPECT_TRUE(good);
      EXPECT_TRUE(res->nummsgs_ == res->numsent_);
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_);
      delete res;
  }
}

TEST(testnetworks, sharedcircuit_ordered)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.000011000000));
  expected.push_back(timestamp(1.000801000000));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000003000000));
  expected.push_back(timestamp(1.000101000000));

  for(int i = 0; i < test_file_names.size(); i++){
      topology::ptr top = torus::construct_2D(4, 4);

      std::vector<double> net_args;
      net_args.push_back(1e-6);
      net_args.push_back(1e9);
      net_args.push_back(ORDERED_FLOWER);
      net_args.push_back(1.5e9);

      interconnect::ptr net = construct_network(net_args, top, "sharedcircuit");

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "sharedcircuit_ordered, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-9);
      if(!good){
          std::cout << "ERROR: expected: " << expected[i] << ", got: " << res->end_ << std::endl;
      }
      EXPECT_TRUE(good);
      EXPECT_TRUE(res->nummsgs_ == res->numsent_);
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_);
      delete res;
  }
}

TEST(testnetworks, sharedcircuit_throttle)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.000011000000));
  expected.push_back(timestamp(1.011787666666));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000003500000));
  expected.push_back(timestamp(1.000101000000));

  for(int i = 0; i < test_file_names.size(); i++){
      topology::ptr top = torus::construct_2D(4, 4);

      std::vector<double> net_args;
      net_args.push_back(1e-6);
      net_args.push_back(1e9);
      net_args.push_back(THROTTLE_FLOWER);
      net_args.push_back(1.5e9);

      interconnect::ptr net = construct_network(net_args, top, "sharedcircuit");

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      //std::cout << "sharedcircuit_throttle, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-9);
      if(!good){
          std::cout << "ERROR: expected: " << expected[i] << ", got: " << res->end_ << std::endl;
      }
      EXPECT_TRUE(good);
      EXPECT_TRUE(res->nummsgs_ == res->numsent_);
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_);
      delete res;
  }
}

TEST(testnetworks, sharedbuffer)
{
  std::vector<timestamp> expected;

  expected.push_back(timestamp(254.000007110001));
  expected.push_back(timestamp(1.000052110001));
  expected.push_back(timestamp(1.000002610001));
  expected.push_back(timestamp(1.000002610001));
  expected.push_back(timestamp(1.000052110001));

  for(int i = 0; i < test_file_names.size(); i++){
      topology::ptr top = torus::construct_2D(4, 4);

      std::vector<double> net_args;

      net_args.push_back(1e-6); //   node_lat
      net_args.push_back(2e9); //   node_router_bw
      net_args.push_back(1e9); //   router_router_bw
      net_args.push_back(1e-8); //   link_lat
      net_args.push_back(1e-7); //   router_lat
      net_args.push_back(8192); //   buffer_capacity
      net_args.push_back(2); //   num_vcs

      interconnect::ptr net = construct_network(net_args, top, "sharedbuffer");

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "sharedbuffer, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-9);
      if(!good){
          std::cout << "ERROR: expected: " << expected[i] << ", got: " << res->end_ << std::endl;
      }
      EXPECT_TRUE(good);
      EXPECT_TRUE(res->nummsgs_ == res->numsent_);
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_);

      delete res;
  }
}

