/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
#include <sstmac/software/skeletons/undumpi/dumpimeta.h>
#include <sstmac/software/skeletons/undumpi/parsedumpi.h>
#include <sstmac/sstmac.h>
#include <bin/undumpi.h>
#include <unistd.h> // getopt
#include <math.h> // pow
#include <iomanip>
#include <sys/time.h>
#include <sys/resource.h>


using namespace sstmac;
using namespace sstmac::hw;
using namespace sstmac::sw;
using namespace sstmac::native;


double get_wall_time()
{
  struct timeval tv;
  gettimeofday(&tv,0);
  return tv.tv_sec + 0.000001 * tv.tv_usec;
}

sstmac::timestamp
run_parsedumpi(const std::string &metafname,
    const std::string &mapfile,
    int nodes, int ppn,
    const std::string &indexing,
    const interconnect::ptr &intercon,
    const topology::ptr &net,
    size_t stack)
{
  dumpimeta::ptr meta = dumpimeta::construct(metafname);
  manager::ptr mgr = manager::construct();
  mgr->set_interconnect(intercon);

  const double memory_bandwidth(1e10); // bandwidth in bytes/second
  const timestamp memory_latency(1e-9); // latency in seconds
  simplenode::ptr node = simplenode::construct(memory_latency,
      memory_bandwidth); // bandwidth in bytes/second
  simplenic::ptr nic = simplenic::construct(node, intercon);
  mgr->add_nodes(node, nic, nodes);


  std::list<std::string> app_params;
  app_params.push_back(meta->dirplusfileprefix_);
  app_params.push_back("1");




  indexstrategy::ptr is;
  allocationstrategy::ptr as;
  if (indexing == "random")          is = randomindexing::construct();
  else if (indexing == "block")      is = blockindexing::construct();
  else if (indexing == "roundrobin") is = roundrobinindexing::construct();
  else if (indexing == "dumpi")      layout(mapfile, meta, net, as, is);
  else throw sstmac::valueerror(indexing+": invalid indexing strategy");

  int nproc = meta->getnumprocs();
  int nnode_needed = nproc/ppn;
  if (nproc%ppn) {
      std::string errmsg = "testparsedumpi: trace nproc must be divisible by ppn";
      std::cout << errmsg << std::endl;
      throw sstmac::valueerror(errmsg);
  }

  launchinfo::ptr l1 = launchinfo::construct("parsedumpi", appid(0), app_params, timestamp(1), meta->getnumprocs());
        std::list<launchinfo::ptr> apps;
        apps.push_back(l1);

         instantlaunch::ptr job = instantlaunch::construct(apps, as, is);

  std::cout << "testparsedumpi: launching on " << nnode_needed << " nodes " << std::endl;

  sstmac::timestamp t = mgr->run(job);
  return t;
}

flower::ptr
get_flower(const topology::ptr &net,
    double node_router_bw, double router_router_bw)
    {
  ordered_flower::ptr
  fl = ordered_flower::construct(net, node_router_bw, router_router_bw);

  return fl;
    }

void
run_parsedumpis(const std::string &tfname,
    const std::string &mapfile,
    int nodes, int ppn,
    const std::string &indexing, int repeats, size_t stack)
{

  double node_router_bw = 1.0e9;
  double router_router_bw = 1.5e9;
  double latency = 1e-6;

  int nxyz = 1;
  while (nxyz*nxyz*nxyz < nodes) nxyz++;

  topology::ptr net = hdtorus::construct_3D(nxyz, nxyz, nxyz, 1, 1);

  for (int j=0; j<repeats; j++) {
      flower::ptr flower = get_flower(net,node_router_bw, router_router_bw);
      interconnect::ptr intercon
      = sharedcircuit::construct(sstmac::timestamp(latency), 0.0, net, flower);
      timestamp t = run_parsedumpi(tfname, mapfile,
          nodes, ppn, indexing, intercon, net, stack);
      std::cout << " " << std::fixed << std::setprecision(6) << 1e6*t.sec() << std::endl;
  }
}

void
print_help(const char *name)
{
  std::cout << name << " [-h] [-n nnode] [-p ppn]"
      << " [-I indexing] [-s stacksize] [-r repeats]"
      << " [-t meta_file_name] [-m map_file_name]"
      << "\n";
}

int
main(int argc, char **argv)
{
  bool helpflag = false;
  int opt;
  int nodes = 200, ppn = 4;
  std::string indexing("dumpi");
  std::string metafile("trace.meta");
  std::string mapfile("trace.map");
  int repeats = 1;
  int stacksize = 1<<16;
  std::string dbg_str = "";
  while(!helpflag && (opt = getopt(argc, argv, "hn:p:d:i:r:s:I:t:m:")) != -1) {
      switch(opt) {
      case 'h':
        helpflag = true;
        break;
      case 'd':
        dbg_str = optarg;
        break;
      case 'n':
        nodes = atoi(optarg);
        break;
      case 'p':
        ppn = atoi(optarg);
        break;
      case 'I':
        indexing = optarg;
        break;
      case 't':
        metafile = optarg;
        break;
      case 'm':
        mapfile = optarg;
        break;
      case 's':
        stacksize = atoi(optarg);
        break;
      case 'r':
        repeats = atoi(optarg);
        break;
      default:
        std::cerr << "Unknown argument -" << opt << "\n";
        helpflag = true;
      }
  }
  if(! helpflag) {
      bool errflag = false;
      if(nodes <= 0) {
          std::cerr << "Invalid node count \"" << nodes << "\"\n";
          errflag = true;
      }
      if(ppn <= 0) {
          std::cerr << "Invalid ppn count \"" << ppn << "\"\n";
          errflag = true;
      }
      if (errflag) {
          return 1;
      }
  }
  else {
      print_help(argv[0]);
      return 0;
  }

  sstmac::logger::set_user_param(dbg_str);
  sstmac::sw::skeletonfactory::init_skeletons();
  //  try {
  run_parsedumpis(metafile, mapfile, nodes, ppn, indexing, repeats, stacksize);
  //  } catch (std::exception &e) {
  //    std::cerr << "Caught exception: " << e.what() << std::endl;
  //  }

  return 0;
}
