/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// gtest.cc: Driver code for testing networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <sstmac_testutil.h>
#include <cstdlib>
#include <fstream>
#include <algorithm>
#include <stdexcept>
#include <memory>
#include <iomanip>
#include <cmath>

extern "C" {
  extern double sqrt(double);
};

using std::setw;
using std::binary_function;
using std::auto_ptr;
using std::ofstream;
using std::set;
using std::atoi;
using std::cout;
using std::endl;
using std::vector;
using std::min_element;
using std::max_element;
using std::logic_error;
using std::multiset;

void
help(const char *argv0)
{
  std::cout << "Usage: " << argv0
      << " [-n name] [-d] [-h]"
      << std::endl;
  std::cout
  << "The options are:" << std::endl
  << "  -h      Print help information and exit." << std::endl
  << "  -d      Write a dot file to net.dot." << std::endl
  << "  -n name Create the named network." << std::endl
  << "Examples of network names are:" << std::endl
  << "  gamma(degree,diameter[,nnodeperswitch[,width]])" << std::endl
  << "  hybercube(dimension)" << std::endl
  << "  ring(nx[,nnodeperswitch[,width]])" << std::endl
  << "  torus2D(nx,ny[,nnodeperswitch[,width]])" << std::endl
  << "  torus3D(nx,ny,nz[,nnodeperswitch[,width]])" << std::endl
  << "  hdtorus2D(nx,ny[,nnodeperswitch[,width]])" << std::endl
  << "  hdtorus3D(nx,ny,nz[,nnodeperswitch[,width]])" << std::endl
  << "  crossbar(nnode)" << std::endl
  << "  fattree(nlevel,half_degree[,nnodeperswitch[,width]])" << std::endl
  << "  clos(half_degree)" << std::endl
  << "  tbird" << std::endl
  << "If no networks are specified, then all except rings will be used."
  << std::endl;
}

static double
average(const std::vector<int> &d)
{
  double a = 0.0;
  for (int i=0; i<d.size(); i++) a+=d[i];
  return a/d.size();
}

static double
stddev(const std::vector<int> &d)
{
  double a = average(d);
  double s = 0.0;
  for (int i=0; i<d.size(); i++) s+=(a-d[i])*(a-d[i]);
  return std::sqrt(s/(d.size()-1));
}

std::string
get_token(std::string &tokens, int *narg = 0)
{
  std::string result;

  if (tokens.size() != 0) {
      size_t pos = tokens.find_first_of("_(),");

      if (pos == std::string::npos) {
          result = tokens;
          tokens.clear();
      }
      else {
          result = tokens.substr(0,pos);
          tokens = tokens.substr(pos+1,std::string::npos);
      }
  }

  if (result.size() > 0 && narg) *narg = *narg + 1;

  return result;
}

static void
write_analysis_header()
{
  cout << setw(20) << ""
      << setw(6) << ""
      << " "
      << setw(6) << ""
      << setw(8) << ""
      << setw(6) << ""
      << setw(4) << ""
      << setw(4) << ""
      << setw(8) << "Hops "
      << setw(20) << "Max Link Load    "
      << endl;
  cout << setw(20) << "Graph  "
      << setw(6) << "Ncomm"
      << " "
      << setw(6) << "Nnode"
      << setw(8) << "Nswitch"
      << setw(6) << "Nlink"
      << setw(4) << "Deg"
      << setw(4) << "Dia"
      << setw(8) << "Ave "
      << setw(4) << "Max"
      << setw(8) << "Ave "
      << setw(8) << "StdDev"
      << endl;
}

static void
analyze_one_line(topology &n, int maxb, int maxp, bool tight_packing)
{
  vector<int> sc(maxb);
  n.stochastic_max_link_load(sc, maxp, tight_packing);

  int minsc = *min_element(sc.begin(), sc.end());
  int maxsc = *max_element(sc.begin(), sc.end());

  double avesc = average(sc);
  double devsc = stddev(sc);

  //   int nmaxsc = 0;
  //   for (int i=0; i<maxb; i++) { if (sc[i] == maxsc) { nmaxsc++; } }
  //   double percent_max = 100*double(nmaxsc)/maxb;

  int maxp_p = (maxp==0?n.n():maxp);
  if (maxp_p > n.n()) maxp_p = n.n();

  cout.setf(std::ios_base::fixed, std::ios_base::floatfield);
  cout << setw(20) << n.name()
           << setw(6) << maxp_p
           << setw(1) << (tight_packing?'p':' ')
           << setw(6) << n.n()
           << setw(8) << n.nswitch()
           << setw(6) << n.nlink()
           << setw(4) << n.d()
           << setw(4) << n.diameter()
           << setw(8) << std::setprecision(4) << n.a()
           << setw(4) << maxsc
           //        << setw(4) << int(percent_max+0.5)
           << setw(8) << std::setprecision(4) << avesc
           << setw(8) << std::setprecision(4) << devsc
           << endl;
}

template <class T>
struct order_less: public binary_function<T, T, bool>
{
  bool operator() (const T &a, const T &b) {
    if (a->n() < b->n()) return true;
    if (a->n() > b->n()) return false;
    return a->d() < b->d();
  }
};

template <class T>
struct degree_less: public binary_function<T, T, bool>
{
  bool operator() (const T &a, const T &b) {
    if (a->d() < b->d()) return true;
    if (a->d() > b->d()) return false;
    return a->n() < b->n();
  }
};

class network_sorter: public binary_function<topology::ptr,
topology::ptr,
bool>
{
  bool node_most_significant_;
  degree_less<topology::ptr > degree_;
  order_less<topology::ptr > order_;
public:
  network_sorter(bool a): node_most_significant_(a) {}
  bool operator() (topology::ptr const &a,
      topology::ptr const &b) {
    if (node_most_significant_) return order_(a,b);
    return degree_(a,b);
  }
};

typedef multiset<topology::ptr, network_sorter> n_multiset;

//typedef multiset<topology*, degree_less<topology*> > n_multiset;

void
insert_gammas(int maxn, n_multiset &netset)
{
  cout << "forming gammas" << endl;
  for (int d=2; d<10; d++) {
      for (int k=1; k<=d; k++) {
          topology::ptr n(sstmac::hw::gamma::construct(d,k));
          if (n->n() > maxn) break;
          netset.insert(n);
      }
  }
}

void
insert_fattrees(int maxn, n_multiset &netset)
{
  cout << "forming fat trees" << std::endl;
  for (int l=2; l<12; l++) {
      for (int k=2; k<13; k+=2) {
          if (fattree::size(l,k) > maxn) break;
          topology::ptr n(fattree::construct(l,k));
          netset.insert(n);
      }
  }
}

void
insert_clos(int maxn, n_multiset &netset)
{
  cout << "forming clos networks" << std::endl;
  for (int n=2; n<12; n++) {
      if (clos::size(n) > maxn) break;
      topology::ptr net(clos::construct(n));
      netset.insert(net);
  }
}

void
insert_hypercubes(int maxn, n_multiset &netset)
{
  cout << "forming hypercubes" << endl;
  for (int d=2; d<11; d++) {
      topology::ptr n(hypercube::construct(d));
      if (n->n() > maxn) break;
      netset.insert(n);
  }
}

void
insert_rings(int maxn, n_multiset &netset)
{
  cout << "forming rings" << endl;
  for (int d=2; d<=maxn; d++) {
      topology::ptr n(ring::construct(d));
      if (n->n() > maxn) break;
      netset.insert(n);
  }
}


void
insert_2Dtori(int maxn, n_multiset &netset)
{
  cout << "forming 2D tori" << endl;
  for (int x=2; x<=maxn/2; x++) {
      //      for (int y=2; y<=x; y++) {
      topology::ptr n(torus::construct_2D(x,x,0));
      if (n->n() > maxn) break;
      netset.insert(n);
      n = torus::construct_2D(x,x,1);
      if (n->n() <= maxn) netset.insert(n);
      n = torus::construct_2D(x,x,2);
      if (n->n() <= maxn) netset.insert(n);
      //        }
  }
}


void
insert_3Dtori(int maxn, n_multiset &netset)
{
  cout << "forming 3D tori" << endl;
  for (int x=2; x<=maxn/4; x++) {
      //      for (int y=2; y<=x; y++) {
      //          for (int z=2; z<=y; z++) {
      topology::ptr n(torus::construct_3D(x,x,x,0));
      if (n->n() > maxn) break;
      netset.insert(n);
      n = torus::construct_3D(x,x,x,1);
      if (n->n() <= maxn) netset.insert(n);
      n = torus::construct_3D(x,x,x,2);
      if (n->n() <= maxn) netset.insert(n);
      //            }
  //        }
  }
}

void
insert_6Dtori(int maxn, n_multiset &netset)
{
  cout << "forming 6D tori" << endl;
  for (int x=2; x<=maxn/32; x++) {
      //      for (int y=2; y<=x; y++) {
      //          for (int z=2; z<=y; z++) {
      vector<int> dim(6);
      dim[0] = x;
      dim[1] = x;
      dim[2] = x;
      dim[3] = x;
      dim[4] = x;
      dim[5] = x;
      topology::ptr n(hdtorus::construct(dim,0, 1));
      if (n->n() > maxn) break;
      netset.insert(n);
      n = hdtorus::construct(dim,1, 1);
      if (n->n() <= maxn) netset.insert(n);
      n = hdtorus::construct(dim,2, 1);
      if (n->n() <= maxn) netset.insert(n);
      //            }
  //        }
  }
}


void
insert_tbird(int maxn, n_multiset &netset)
{
  cout << "forming tbird" << endl;
  if (maxn >= 4608) {
      topology::ptr n(tbird::construct());
      netset.insert(n);
  }
}

void
analyze_networks(n_multiset &netset, int maxb, int maxp, bool tight_packing)
{
  write_analysis_header();
  for (n_multiset::iterator i = netset.begin();
      i != netset.end();
      i++) {
      analyze_one_line(**i, maxb, maxp, tight_packing);
  }
}


void
follow_hops(int step, const topology::ptr& net,
    vertex_index loc, vertex_index dst)
{
  if (loc == dst) return;
  std::vector<lwedge> nxt;
  try {
      net->next_hop(loc,dst,nxt);
  }
  catch (std::logic_error &e) {
      std::cout << "  next hop failed: " << e.what() << std::endl;
      return;
  }
  std::cout << "  step " << std::setw(4) << step
      << ": next hop from " << std::setw(4) << loc.value()
      << " to " << std::setw(4) << dst.value()
      << ":";
  for (int i=0; i<nxt.size(); i++) {
      std::cout << " " << std::setw(4) << nxt[i].to();
      if (nxt[i].from() != loc) {
          std::cout << "(bad loc in edge)";
      }
      try {
          net->g().edge_to_index(nxt[i]);
      }
      catch (...) {
          std::cout << "(bad edge)";
      }
  }
  std::cout << std::endl;
  for (int i=0; i<nxt.size(); i++) {
      follow_hops(step+1, net, nxt[i].to(), dst);
  }
}

int
main(int argc, char **argv)
{
  std::string netname = "invalid";
  if (argc > 1) {
      netname = argv[1];
  }

  bool verbose = false;
  bool print_dot = false;
  bool show_help = false;
  int rc = 0;

  while(1) {
      int c = getopt(argc, argv, "vhn:d");
      if (c < 0) break;
      switch (c) {
      case 'h':
        show_help = true;
        break;
      case 'v':
        verbose = true;
        break;
      case 'd':
        print_dot = true;
        break;
      case 'n':
        netname = optarg;
        break;
      default:
        show_help = true;
        rc = 1;
      }
  }

  if (show_help) {
      help(argv[0]);
      exit(rc);
  }

  if(netname != "invalid"){ //if a netname was given, do a topology walk on it

      std::string remaining_tokens = netname;
      std::string netbase = get_token(remaining_tokens);
      int narg = 0;
      std::string token;
      std::vector<int> args;
      while ((token = get_token(remaining_tokens, &narg)).size() > 0) {
          args.push_back(atoi(token.c_str()));
      }

      topology::ptr net;
      net = construct_topology(args,netbase);

      net->g().print();
      net->g().print_all(std::cout, verbose);

      ofstream dot("net.dot");
      net->g().print_dot(dot);

      walk w;
      const set<vertex> &v = net->nodes();
      for (set<vertex>::const_iterator j = v.begin();
          j != v.end();
          j++) {
          const vertex &v0 = *j;
          for (set<vertex>::const_iterator i = v.begin();
              i != v.end();
              i++) {
              net->route(v0,*i,w);
              cout << "route from " << v0 << " to " << *i << ": ";
              if (w.ev().size()) {
                  cout << w.ev()[0].from();
                  for (int k=0; k<w.ev().size(); k++) {
                      const lwedge &e = w.ev()[k];
                      cout << " (" << e.id();
                      try {
                          edge_index ei = net->g().edge_to_index(e);
                          if (net->edge_width(ei) != 1) {
                              cout << "," << net->edge_width(ei);
                          }
                      }
                      catch (...) {
                          cout << ",BAD";
                      }
                      cout << ")-> " << e.to();
                  }
              }
              cout << endl;
              net->g().check_walk(v0,*i,w);
          }
      }

      RNG::UniformInteger::ptr
      rng = RNG::SimpleCombo::construct();
      RNG::UniformInteger_functor rngf(rng);
      int64_t order = net->g().order();
      if (order > 0) {
          for (int i=0; i<10; i++) {
              vertex_index loc(rngf(static_cast<RNG::rngint_t>(order))),
                  dst(rngf(static_cast<RNG::rngint_t>(order)));
              std::cout << "hopping from " << loc << " to " << dst << ":"
                  << std::endl;
              follow_hops(1, net,loc,dst);
          }
      }

      if (verbose) {
          bool tight_packing = false;
          int maxb = 50;
          int maxp = 0;
          std::vector<int> sc(maxb);
          net->stochastic_max_link_load(sc, maxp, tight_packing);
          int maxsc = *std::max_element(sc.begin(), sc.end());
          double avesc = average(sc);
          double devsc = stddev(sc);
          int64_t maxp_p = (maxp==0?net->n():maxp);
          if (maxp_p > net->n()) maxp_p = net->n();
          std::cout.setf(std::ios_base::fixed, std::ios_base::floatfield);
          std::cout << std::setw(20) << net->name()
                      << std::setw(6) << maxp_p
                      << std::setw(1) << (tight_packing?'p':' ')
                      << std::setw(6) << net->n()
                      << std::setw(8) << net->nswitch()
                      << std::setw(6) << net->nlink()
                      << std::setw(4) << net->d()
                      << std::setw(4) << net->diameter()
                      << std::setw(8) << std::setprecision(4) << net->a()
                      << std::setw(4) << maxsc
                      << std::setw(8) << std::setprecision(4) << avesc
                      << std::setw(8) << std::setprecision(4) << devsc
                      << std::endl;
      }
  }else{   //if a netname was not given, we do a static analysis of all networks
      bool degree_order = false;

      bool tight_packing = false;

      int maxn = 256;
      int maxb = 20;
      int maxp = 0;


      network_sorter sorter(!degree_order);
      n_multiset netset(sorter);

      insert_gammas(maxn, netset);
      insert_fattrees(maxn, netset);
      insert_clos(maxn, netset);
      insert_hypercubes(maxn, netset);
      insert_rings(maxn, netset);
      insert_2Dtori(maxn, netset);
      insert_3Dtori(maxn, netset);
     // insert_6Dtori(maxn, netset);
      insert_tbird(maxn, netset);

      cout << endl;
      analyze_networks(netset,maxb,maxp,tight_packing);

  }

  return 0;
}

